Package: santoku
Type: Package
Title: A Versatile Cutting Tool
Version: 1.2.0
Authors@R: 
    c(
    person(given = "David",
           family = "Hugh-Jones",
           role = c("aut", "cre"),
           email = "davidhughjones@gmail.com"),
    person(given = "Daniel",
           family = "Possenriede",
           role = c("ctb"),
           email = "possenriede@gmail.com")
    )
Maintainer: David Hugh-Jones <davidhughjones@gmail.com>
Description: A tool for cutting data into intervals. Allows singleton intervals.
  Always includes the whole range of data by default. Flexible labelling. 
  Convenience functions for cutting by quantiles etc. Handles dates, times, units
  and other vectors.
License: MIT + file LICENSE
Encoding: UTF-8
RoxygenNote: 7.3.3
Suggests: bench, bit64, covr, haven, Hmisc, hms, knitr, lubridate,
        purrr, rmarkdown, scales, stringi, testthat (>= 3.2.0), units,
        withr, xts, zoo
Config/testthat/edition: 3
LinkingTo: Rcpp
Depends: R (>= 3.5.0)
Imports: Rcpp, assertthat, glue, lifecycle, rlang, vctrs
URL: https://github.com/hughjonesd/santoku,
        https://hughjonesd.github.io/santoku/
BugReports: https://github.com/hughjonesd/santoku/issues
VignetteBuilder: knitr
RdMacros: lifecycle
NeedsCompilation: yes
Packaged: 2026-04-28 14:15:23 UTC; davidhugh-jones
Author: David Hugh-Jones [aut, cre],
  Daniel Possenriede [ctb]
Repository: CRAN
Date/Publication: 2026-04-28 14:40:02 UTC
Built: R 4.6.0; x86_64-apple-darwin20; 2026-05-01 20:57:12 UTC; unix
Archs: santoku.so.dSYM

# Generated by roxygen2: do not edit by hand

S3method(apply_format,"function")
S3method(apply_format,character)
S3method(apply_format,list)
S3method(apply_format.character,character)
S3method(apply_format.character,default)
S3method(apply_format.character,numeric)
S3method(apply_format.list,default)
S3method(as.double,breaks)
S3method(brk_width,Duration)
S3method(brk_width,default)
S3method(class_bounds,Date)
S3method(class_bounds,POSIXct)
S3method(class_bounds,default)
S3method(class_bounds,difftime)
S3method(class_bounds,integer64)
S3method(class_bounds,numeric)
S3method(class_bounds,units)
S3method(class_bounds,zoo)
S3method(endpoint_labels,Date)
S3method(endpoint_labels,POSIXt)
S3method(endpoint_labels,default)
S3method(endpoint_labels,double)
S3method(endpoint_labels,integer)
S3method(endpoint_labels,numeric)
S3method(endpoint_labels,quantileBreaks)
S3method(endpoint_labels,sdBreaks)
S3method(format,breaks)
S3method(print,breaks)
S3method(santoku_cast_common,Date)
S3method(santoku_cast_common,Date.Date)
S3method(santoku_cast_common,Date.POSIXct)
S3method(santoku_cast_common,POSIXct)
S3method(santoku_cast_common,POSIXct.Date)
S3method(santoku_cast_common,POSIXct.POSIXct)
S3method(santoku_cast_common,default)
S3method(santoku_cast_common,default.default)
S3method(santoku_cast_common,default.hexmode)
S3method(santoku_cast_common,default.integer64)
S3method(santoku_cast_common,default.octmode)
S3method(santoku_cast_common,default.ts)
S3method(santoku_cast_common,default.zoo)
S3method(santoku_cast_common,double)
S3method(santoku_cast_common,double.default)
S3method(santoku_cast_common,double.integer64)
S3method(santoku_cast_common,hexmode)
S3method(santoku_cast_common,hexmode.default)
S3method(santoku_cast_common,hexmode.hexmode)
S3method(santoku_cast_common,integer64)
S3method(santoku_cast_common,integer64.default)
S3method(santoku_cast_common,integer64.double)
S3method(santoku_cast_common,integer64.integer64)
S3method(santoku_cast_common,octmode)
S3method(santoku_cast_common,octmode.default)
S3method(santoku_cast_common,octmode.octmode)
S3method(santoku_cast_common,ts)
S3method(santoku_cast_common,ts.default)
S3method(santoku_cast_common,zoo)
S3method(santoku_cast_common,zoo.default)
S3method(scaled_endpoints,breaks)
S3method(scaled_endpoints,default)
S3method(sequence_width,Period)
S3method(sequence_width,default)
export(brk_default)
export(brk_equally)
export(brk_evenly)
export(brk_fn)
export(brk_manual)
export(brk_mean_sd)
export(brk_n)
export(brk_pretty)
export(brk_proportions)
export(brk_quantiles)
export(brk_spikes)
export(brk_width)
export(chop)
export(chop_deciles)
export(chop_equally)
export(chop_evenly)
export(chop_fn)
export(chop_mean_sd)
export(chop_n)
export(chop_pretty)
export(chop_proportions)
export(chop_quantiles)
export(chop_spikes)
export(chop_width)
export(dissect)
export(exactly)
export(fillet)
export(is.breaks)
export(kiru)
export(lbl_dash)
export(lbl_date)
export(lbl_datetime)
export(lbl_discrete)
export(lbl_endpoint)
export(lbl_endpoints)
export(lbl_glue)
export(lbl_intervals)
export(lbl_manual)
export(lbl_midpoints)
export(lbl_seq)
export(percent)
export(santoku_cast_common.Date)
export(santoku_cast_common.POSIXct)
export(santoku_cast_common.default)
export(santoku_cast_common.double)
export(santoku_cast_common.hexmode)
export(santoku_cast_common.integer64)
export(santoku_cast_common.octmode)
export(santoku_cast_common.ts)
export(santoku_cast_common.zoo)
export(tab)
export(tab_deciles)
export(tab_dissect)
export(tab_equally)
export(tab_evenly)
export(tab_fn)
export(tab_mean_sd)
export(tab_n)
export(tab_pretty)
export(tab_proportions)
export(tab_quantiles)
export(tab_spikes)
export(tab_width)
import(assertthat)
importFrom(Rcpp,sourceCpp)
importFrom(lifecycle,deprecated)
useDynLib(santoku, .registration = TRUE)

[![PyPI version](https://badge.fury.io/py/santoku.svg)](https://badge.fury.io/py/santoku)

[![Install Python Dependencies, Run Tests on Pushes to Main Branch, Bump Version and Tag and Publish to PyPI](https://github.com/wiris/santoku/workflows/Install%20Python%20Dependencies,%20Run%20Tests%20on%20Pushes%20to%20Main%20Branch,%20Bump%20Version%20and%20Tag%20and%20Publish%20to%20PyPI/badge.svg)](https://github.com/wiris/santoku/actions)

[![Code style: black](https://img.shields.io/badge/code%20style-black-000000.svg)](https://github.com/psf/black)

# What is Santoku?

Santoku is a toolkit written in Python for interacting with AWS, Google Cloud platform, Salesforce and Slack.

The purpose of Santoku is to have the interactions with all the external services collected in a single package. The package contains wrappers around the respective APIs and high level methods for the most common patterns in order to simplify the interaction with those services, whether by being shorter to type, more descriptive, more specific to our needs or simply easier to read for developers.

## Quickstart

### Installation

If you have the wheel, you can install it with:

```bash
pip install --upgrade --force-reinstall dist/santoku-*.whl
```

Run the following command to install it from PyPI:

```bash
pip install santoku
```

Or use the following to install it via Poetry

```bash
poetry add santoku
```

### How To Use It

You can use the package as follows:

```python
from santoku.slack.slack_bot_handler import SlackBotHandler

slack_bot = SlackBotHandler.from_aws_secrets_manager("your_secret")
```

## Content

The package `santoku` contains several subpackages: `aws`, `google`, `salesforce`, `slack`, `sql`. Each subpackage provides connection to different external services and are formed by a collection of modules, where each module consists of handlers for more specific services. Each handler class has unit tests to ensure the correct behaviour of the methods of these classes.

### AWS

AWS (Amazon Web Services) is a cloud computing platform that provides a set of primitive abstract technical infrastructure and distributed computing building blocks and tools.

The connection to AWS has been done through the AWS SDK, which in Python is called [boto3](https://github.com/boto/boto3). We provide wrappers of the `boto3` SDK to make easy the operations to interact with different services.

The use of this subpackage requires having AWS credentials somewhere. We provide flexibility to either keep credentials in AWS credentials/configuration file, set environment variables, or to pass them directly as arguments in the initializer of each handler class. More info on AWS configurations and credentials [here](https://docs.aws.amazon.com/cli/latest/userguide/cli-chap-configure.html).

The unit tests in this subpackage implement mocks to the AWS services and do not pretended to access or modify the environment of your real account. In order to have safer unit tests for the AWS subpackage, we use moto, a mocking library for most AWS services, which allows our methods to interact with a fully mocked version of the AWS environment via decorators while not needing an actual connection to the internet or a test AWS account.

#### Amazon S3

Amazon Simple Storage Service (Amazon S3) is an object storage service that offers scalability, data availability, security, and performance.

We provide methods to easily list and delete objects inside buckets; read and write content within S3 objects; upload a dataframe into csv or parket format to a specific location; generate and upload an Amazon Quicksight manifest in S3 in order to create analysis in Amazon Quicksight, and so on.

An object can be uploaded to S3 with the following:

```python
from santoku.aws.s3_handler import S3Handler

s3_handler = S3Handler()
s3_handler.put_object(bucket="your_bucket_name", object_key="your_object_key", content="Your object content.")
```

#### AWS Secrets Manager

AWS Secrets Manager protects secrets needed to access applications, services, and IT resources. The service allows rotating, managing, and retrieving credentials, keys, and other secrets.

We provide methods to get the content of a previously created secret.

```python
from santoku.aws.secrets_manager_handler import SecretsManagerHandler

secrets_manager = SecretsManagerHandler()
secret_content = secrets_manager.get_secret_value(secret_name="your_secret_name")
```

We use this service as our default credential manager. Most classes that require some form of authentication in santoku are provided with alternative class methods that retrieve the credentials directly from Secrets Manager. For example, instead of directly providing credentials to the BigQuery handling class, we simply provide it with the name of the secret where they are stored:

```python
from santoku.google.bigquery import BigQueryHandler

bigquery_handler = BigQueryHandler(
    type="your_type",
    project_id="your_project_id"
    private_key_id="your_private_key_id"
    private_key="your_private_key"
    client_email="your_client_email"
    client_id="your_client_id"
    auth_uri="your_auth_uri"
    token_uri="your_token_uri"
    auth_provider_x509_cert_url="your_auth_provider_x509_cert_url"
    client_x509_cert_url="your_client_x509_cert_url"
)
```

or

```python
bigquery_handler = BigQueryHandler.from_aws_secrets_manager(
    secret_name="your_secret_name"
)
```

#### Amazon Simple Queue Service

Amazon Simple Queue Service (SQS) is a fully managed message queuing service that supports programmatic sending of messages via web service applications as a way to communicate over the Internet.

We provide methods to receive, delete, and send single or a batch of messages.

```python
from santoku.aws.sqs_handler import SQSHandler

sqs_handler = SQSHandler()
entries = [
    {
        "Id": "Id1",
        "MessageBody": "Your message 1",
    },
    {
        "Id": "Id2",
        "MessageBody": "Your message 1",
    }
]
sqs_handler.send_message_batch(queue_name="your_queue_name", entries=entries)
```

### Google Cloud Platform

Google Cloud Platform a suite of cloud computing services provided by Google that runs on the same Cloud infrastructure that Google uses internally for its end-user products.

The connection to Google Cloud Platform has been done using the [google-cloud-core](https://googleapis.dev/python/google-api-core/latest/index.html) package.

The use of this subpackage requires having [Google Cloud Platform credentials](https://cloud.google.com/docs/authentication/production#obtaining_and_providing_service_account_credentials_manually) (in this case, a service account for programmatic access), these can be passed as arguments in the initializer of the handler class directly, or you can store them in AWS Secrets Manager and retrieve them during the initialization using the class method instead.

We provide a handler that allows doing queries on BigQuery services:

```python
query_results = bigquery_handler.get_query_results(query="SELECT * FROM `your_table`")
```

### Salesforce

Salesforce is a Customer Relationship Management (CRM) platform that gives to the marketing, sales, commerce, and service depertments a single, shared view of every customer.

The connection to Salesforce has been done using the [Salesforce REST API](https://developer.salesforce.com/docs/atlas.en-us.api_rest.meta/api_rest/quickstart.htm).

The use of this subpackage requires having Salesforce credentials, these can be passed as arguments in the initializer of the handler class directly, or you can store them in AWS Secrets Manager and retrieve them during the initialization using the class method instead.

This subpackage provide methods to insert/modify/delete salesforce object records. You can perform operations by doing HTTP requests directly or using methods with higher level of abstraction, which are easier to handle. The lasts ones are just wrappers of the HTTP request method. To obtain records you can perform queries using SOQL.

The unit tests require valid Salesforce credentials to be executed. The tests are implemented in the way that no new data will remain in the account and no existent data will be modified. However, having Salesforce credentials for sandbox use is recommended.

You can use the package to perform a request as follows.

```python
from santoku.salesforce.objects_handler import ObjectsHandler

objects_handler = ObjectsHandler(
    auth_url="your_auth_url",
    username="your_username",
    password="your_password",
    client_id="your_client_id",
    client_secret="your_client_secret",
)
contact_payload = {"FirstName": "Alice", "LastName": "Ackerman", "Email": "alice@example.com"}

objects_handler.do_request(method="POST", path="sobjects/Contact", payload=contact_payload)
```

or insert a record with

```python
objects_handler.insert_record(sobject="Contact", payload=contact_payload)
```

Finally, you can do a SOQL query with:

```python
records = objects_handler.do_query_with_SOQL("SELECT Id, Name from Contact")
```

### Slack

Slack is a proprietary business communication platform. A Slack Bot is a nifty way to run code and automate tasks. In Slack, a bot is controlled programmatically via a bot user token that can access one or more of Slack’s APIs.

The connection to Slack has been done using the [Slack Web API](https://slack.dev/python-slackclient/basic_usage.html).

The use of this subpackage requires having Slack API Token of a Slack Bot, which can be passed as argument in the initializer of the handler class directly, or you can store it in AWS Secrets Manager and retrieve it during the initialization using the class method instead.

This subpackage provide methods to send messages to a channel. A message can be sent with:

```python
from santoku.slack.slack_bot_handler import SlackBotHandler

slack_bot_handler = SlackBotHandler(api_token="your_api_token")
slack_bot_handler.send_message(channel="your_chanel_name", message="Your message.")
```

### SQL

SQL (Structured Query Language) is a domain-specific language designed for managing data held in a relational database management system (RDBMS). The purpose of this subpackage is to provide connection to different RDBMSs.

#### MySQL

MySQL is an open-source RDBMS. The connection to MySQL has been done using the [MySQL Connector for python](https://dev.mysql.com/doc/connector-python/en/).

The use of this subpackage requires having MySQL authentication parameters, which can be passed as argument in the initializer of the handler class directly, or you can store it in AWS Secrets Manager and retrieve it during the initialization using the class method instead.

This subpackage provides methods to do queries and retrieve the results in different forms.

```python
from santoku.sql.mysql_handler import MySQLHandler

mysql_handler = MySQLHandler(user="your_user", password="your_password", host="your_host", database="your_database")
mysql_handler.get_query_results(query="SELECT * FROM your_table")
```

## Development

### Project

We use Poetry to handle this project. Poetry is a tool for dependency management and packaging in Python. It allows you to declare the libraries your project depends on and it will manage (install/update) them for you. More detais in [their documentation](https://python-poetry.org/docs/basic-usage/).

Poetry is already included in the development environment.

#### Dependencies

If you want to add dependencies to your project, you can specify them in the `tool.poetry.dependencies` section of the `pyproject.toml` file. Also, instead of modifying the `pyproject.toml` file by hand, you can use the add command:

```bash
poetry add <package_name>
```

Poetry will automatically find a suitable version constraint and install the package and subdependencies.

**Note:** Remember to commit changes of `poetry.lock` and `pyproject.toml` files after adding a new dependency.

**Note:** You can find more details on how to handle versions of packages [here](https://python-poetry.org/docs/dependency-specification/).

### Environment

We provide a development environment that uses the Visual Studio Code Remote - Containers extension. This extension lets you use a Docker container in order to have a consistent and easily reproducible development environment.

The files needed to build the container are located in the `.devcontainer` directory:

* `devcontainer.json` contains a set of configurations, tells VSCode how to access the container and which extensions it should install.
* `Dockerfile` defines instructions for the building of the container image.

More info [here](https://code.visualstudio.com/docs/remote/containers-tutorial).

#### Environment Variables

The containerized environment will automatically set as environment variables the your variables stored in a `.env` file at the top level of the repository. For example, this is required for certain tests that require credentials, which are (of course) not versioned. Be aware that:

* The Docker image building process will **fail** if you do not include a `.env` file at the top level of the repository.
* If you change the contents of the `.env` file you will need to rebuild the container for the changes to take effect within the environment.

#### Sharing Git credentials with your container

The containerized environment will automatically forward your local SSH agent if one is running. More info [here](https://code.visualstudio.com/docs/remote/containers#_using-ssh-keys). It works for Windows and Linux.

### Creating a PR

Create a pull request (PR) to propose and collaborate on changes to the project. These changes MUST BE proposed in a branch, which ensures that the default branch only contains finished and approved work.

Be sure to run tests locally before commiting your changes.

#### Running tests

The tests are implemented with pytest and there are unit tests for each of the handler modules. Tests in the `aws` subpackage implement mocks to S3 and do not require real credentials, however, the remaining tests in other subpackages do.

To run the tests just execute `pytest` if you are already inside Poetry virtual environment or `poetry run pytest`.

Moreover, when a PR is created a GitHub Actions CI pipeline (see [`.github/workflows/ci.yml`](./.github/workflows/ci.yml)) is executed. This pipeline is in charge of running tests.

### Release

Wheel is automatically created and uploaded to PyPI by the GitHub Actions CD pipeline (see [`.github/workflows/cd.yml`](./.github/workflows/cd.yml)) when the PR is merged in main branch.


## Why Santoku?

From Wikipedia:

```text
The Santoku bōchō (Japanese: 三徳包丁; "three virtues" or "three uses") or Bunka bōchō (文化包丁) is a general-purpose kitchen knife originating in Japan. Its blade is typically between 13 and 20 cm (5 and 8 in) long, and has a flat edge and a sheepsfoot blade that curves down an angle approaching 60 degrees at the point. The term Santoku may refer to the wide variety of ingredients that the knife can handle: meat, fish and vegetables, or to the tasks it can perform: slicing, chopping and dicing, either interpretation indicating a multi-use, general-purpose kitchen knife.
```

# Utils
This subpackage implements supporting methods and abstract classes that should be useful in a variety of projects.

# Configuration
The `configuration` module provides classes to store and manage configuration files at runtime. For example, the `ConfigurationManager` class handles serving complex custom configuration files to an application, switching between different configurations and (optionally) strongly validating the schema of all configurations in a standard way, all at runtime.

Configurations can be loaded from Python dictionaries or serialized JSON files, like so:

```python
from santoku.utils import ConfigurationManager

# from Python data structures
configuration_manager = ConfigurationManager.(
    configurations=list_of_configurations,
    schema=schema_dict,
    initial_configuration="my_configuration"
)

# from JSON files
configuration_manager = ConfigurationManager.from_json(
    configurations_file_path="path/to/configurations.json",
    schema_file_path="path/to/schema.json",
    initial_configuration="my_configuration"
)
```

The value of each setting can be easily accessed:

```python
setting_value = configuration_manager.get_setting("my_setting")

# nested settings can be accessed by passing the list of keys, in order
nested_boolean_setting = configuration_manager.get_setting("object_setting", "nested_setting")
```


## Schema validation
For schema validation of configuration files, we use [JSON Schema](https://json-schema.org/) and the [jsonschema](https://pypi.org/project/jsonschema/) package. At the time of writing, we support drafts 4, 6 and 7 of the JSON Schema specification. `jsonschema` does not support the latest draft (2019-09, formerly knwon as draft 8), although it is [one of the most requested issues](https://github.com/Julian/jsonschema/issues/613) and the maintainer said that it will eventually be supported.

### Best Practices
When managing configurations with schemas, we recommend the following best practices:
- Add the `required` field to each object in the schema, such that if a setting is missing from the configuration a validation error will ensure that you do not execute with configurations. Irrelevant values (e.g. for hierarchical settings) can be made nullable and thus ignored (see below).
    ```json
    {
        "type": "object",
        "properties": {
            "first": {...},
            "second": {...},
        },
        "required": ["first", "second"]
    }
    ```
- Set `additionalProperties` to `false` on each object to indicate that nothing can be defined in addition to what the schema specifies. This will keep configuration files clean and to the point. This field defaults to `true`, but we recommend making that assertion explicit in your schemas if you ever need it.
    ```json
    {
        "type": "object",
        "properties": {...},
        "required": [...],
        "additionalProperties": false
    }
    ```
- To make a setting nullable, add `null` to the list of types for that particular field. In Python, a `null` value is parsed to `None`, which makes handling of unset settings easy.
    ```json
    {
    "setting": {
        "type": ["string", "null"]
        }
    }
    ```

## Example
Say you have the following schema for the settings your app needs:

```JSON
# schema.json
{
    "$schema": "http://json-schema.org/draft-07/schema#",
    "description": "Settings for my app",
    "type": "object",
    "properties": {
        "boolean_setting": {"type": "boolean"},
        "int_setting": {"type": "integer"},
        "object_setting": {
            "type": "object",
            "properties": {
                "nested_float_setting": {"type": "numeric"},
                "nested_string_setting": {"type": "string"},
            },
            "required": ["nested_float_setting", "nested_string_setting"],
            "additionalProperties": false
        }
    },
    "required": ["boolean_setting", "int_setting", "object_setting"],
    "additionalProperties": false
}
```

Your different configurations must then be provided in array form, with `name` being a unique identifier and the `settings` object conforming to the `schema.json` above:

```JSON
# configs.json
[
    {
        "name": "configuration_1",
        "settings": {
            "boolean_setting": true,
            "int_setting": 42,
            "object_setting": {
                "nested_float_setting": 3.14,
                "nested_string_setting": "Hello, Sailor!",
            }
        }
    },
    {
        "name": "configuration_2",
        "settings": ...
    },
    ...
]
```

## Note on recursive type hints
This module would greatly benefit from recursive type hints. We tried but discovered that currently mypy does not support this. To summarize, it is possible since Python 3.5 (when type hints where first implemented) to define custom types using type aliasing, like so:

```python
from typing import List, Union

# a custom type for a list of str and/or int
strings_and_ints = List[Union[str, int]]
```

Recursive types can be created with the restriction that, to use a type inside its own definition, one must use a string literal instead of simply the type. This is called a [forward reference](https://www.python.org/dev/peps/pep-0484/#forward-references).

```python
from typing import Dict, List, Union

# use "JSON" instead of just JSON
JSON = Union[bool, int, float, str, Dict[str, "JSON"], List["JSON"]]
```

Since this issue is currently [the most demanded from the mypy project](https://github.com/python/mypy/issues/731), it is expected that we will eventually have support for recursive type checking. However, even if mypy raises an error, the code above does compile. For now, we make do with a wrong but simple version that works for our purposes:

```python
from typing import Union
JSON = Union[list, dict]
```
    

#' Create a `breaks` object manually
#'
#' @param breaks A vector, which must be sorted.
#' @param left_vec A logical vector, the same length as `breaks`.
#'   Specifies whether each break is left-closed or right-closed.
#'
#' @inherit breaks-doc return
#'
#' @details
#'
#' All breaks must be closed on exactly one side, like `..., x) [x, ...`
#' (left-closed) or `..., x) [x, ...` (right-closed).
#'
#' For example, if `breaks = 1:3` and `left = c(TRUE, FALSE, TRUE)`, then the
#' resulting intervals are \preformatted{
#' T        F       T
#' [ 1,  2 ] ( 2, 3 )
#' }
#'
#' Singleton breaks are created by repeating a number in `breaks`. Singletons
#' must be closed on both sides, so if there is a repeated number
#' at indices `i`, `i+1`, `left[i]` *must* be `TRUE` and `left[i+1]` must be
#' `FALSE`.
#'
#' `brk_manual()` ignores `left` and `close_end` arguments passed in
#' from [chop()], since `left_vec` sets these manually.
#' `extend` and `drop` arguments are respected as usual.
#'
#' @export
#'
#' @examples
#' lbrks <- brk_manual(1:3, rep(TRUE, 3))
#' chop(1:3, lbrks, extend = FALSE)
#'
#' rbrks <- brk_manual(1:3, rep(FALSE, 3))
#' chop(1:3, rbrks, extend = FALSE)
#'
#' brks_singleton <- brk_manual(
#'       c(1,    2,    2,     3),
#'       c(TRUE, TRUE, FALSE, TRUE))
#'
#' chop(1:3, brks_singleton, extend = FALSE)
#'
brk_manual <- function (breaks, left_vec) {
  assert_that(
          is.numeric(breaks),
          noNA(breaks),
          is.logical(left_vec),
          noNA(left_vec),
          length(left_vec) == length(breaks)
        )

  function (x, extend, left, close_end) {
    if (! left) warning("Ignoring `left` with `brk_manual()`")
    if (! close_end) warning("Ignoring `close_end` with `brk_manual()`")
    breaks <- create_breaks(breaks, left_vec)
    breaks <- extend_and_close(breaks, x, extend, left = TRUE, close_end = FALSE)

    breaks
  }
}


#' @rdname chop_fn
#' @export
#' @order 2
brk_fn <- function (fn, ...) {
  assert_that(is.function(fn))

  function (x, extend, left, close_end) {
    breaks <- fn(x, ...)
    # some functions (e.g. quantile()) return a named vector
    # which might create surprise labels:
    breaks <- unname(breaks)
    assert_that(is.numeric(breaks))
    if (length(breaks) == 0) {
      return(empty_breaks())
    }

    breaks <- create_extended_breaks(breaks, x, extend, left, close_end)

    breaks
  }
}


#' @rdname chop_pretty
#'
#' @export
#' @order 2
brk_pretty <- function (n = 5, ...) {
  assert_that(is.count(n))

  function (x, extend, left, close_end) {
    breaks <- base::pretty(x, n = n, ...)
    if (length(breaks) == 0 || is.null(breaks)) {
      return(empty_breaks())
    }

    breaks <- create_extended_breaks(breaks, x, extend, left, close_end)

    breaks
  }
}


#' @rdname chop_mean_sd
#' @export
#' @order 2
#' @importFrom lifecycle deprecated
brk_mean_sd <- function (sds = 1:3, sd = deprecated()) {
  if (lifecycle::is_present(sd)) {
    lifecycle::deprecate_warn(
            when = "0.7.0",
            what = "brk_mean_sd(sd)",
            with = "brk_mean_sd(sds = 'vector of sds')"
          )
    assert_that(is.number(sd), sd > 0)
    # we start from 0 but remove the 0
    # this works for e.g. sd = 0.5, whereas seq(1L, sd, 1L) would not:
    sds <- seq(0L, sd, 1L)[-1]
    if (! sd %in% sds) sds <- c(sds, sd)
  }

  assert_that(is.numeric(sds), all(sds > 0))

  function (x, extend, left, close_end) {
    x_mean <- mean(x, na.rm = TRUE)
    x_sd <- stats::sd(x, na.rm = TRUE)

    if (is.na(x_mean) || is.na(x_sd) || x_sd == 0) {
      return(empty_breaks())
    }

    # add negative sds, then scale them by mean and sd
    sds <- sort(sds)
    sds <- c(-rev(sds), 0, sds)
    breaks <- sds * x_sd + x_mean
    breaks <- create_lr_breaks(breaks, left)

    needs <- needs_extend(breaks, x, extend, left, close_end)
    if ((needs & LEFT) > 0) sds <- c(-Inf, sds)
    if ((needs & RIGHT) > 0) sds <- c(sds, Inf)
    breaks <- extend_and_close(breaks, x, extend, left, close_end)

    class(breaks) <- c("sdBreaks", class(breaks))
    attr(breaks, "scaled_endpoints") <- sds

    breaks
  }
}

test_that("systematic tests", {
  x_vals <- list(
    ordinary = 4:1,
    real     = 4:1 + 0.5,
    NAs      = c(NA, 1:3),
    all_NAs  = c(NA_real_, NA_real_, NA_real_),
    inf      = c(-Inf, Inf, 1:3),
    inf_lo   = c(-Inf, 1:3),
    inf_hi   = c(Inf, 1:3),
    "NaN"    = c(NaN, 1:3),
    same     = rep(1, 3),
    one      = 3,
    none     = numeric(0),
    char     = letters[1:3],
    complex  = 1:3 + 1i,
    Date     = as.Date("1950-01-01") + 0:20,
    POSIXct  = as.POSIXct("2000-01-01") + 0:30
  )
  brk_funs <- list(
    brk_evenly      = expression(brk_evenly(2)),
    brk_proportions = expression(brk_proportions(c(0.25, 0.6))),
    brk_manual      = expression(brk_manual(1:3, c(FALSE, TRUE, FALSE))),
    brk_mean_sd     = expression(brk_mean_sd()),
    brk_pretty      = expression(brk_pretty()),
    brk_n           = expression(brk_n(5)),
    brk_n_merge     = expression(brk_n(5, tail = "merge")),
    brk_quantiles   = expression(brk_quantiles(1:3/4)),
    brk_default     = expression(brk_default(1:3)),
    brk_default2    = expression(brk_default(c(1, 2, 2, 3))),
    brk_default_lo  = expression(brk_default(1)),
    brk_default_hi  = expression(brk_default(5)),
    brk_width       = expression(brk_width(1)),
    brk_width2      = expression(brk_width(1, 0)),
    brk_spikes  = expression(brk_spikes(1:3, n = 2)),
    brk_w_difft_day = expression(brk_width(as.difftime(5, units = "days"))),
    brk_w_difft_sec = expression(brk_width(as.difftime(5, units = "secs"))),
    brk_def_Date    = expression(brk_default(as.Date("1950-01-05") + c(0, 5))),
    brk_def_POSIXct = expression(brk_default(as.POSIXct("2000-01-01") + c(10, 20)))
  )
  lbl_funs <- list(
    lbl_dash          = expression(lbl_dash()),
    lbl_intervals     = expression(lbl_intervals()),
    lbl_seq           = expression(lbl_seq("a")),
    lbl_endpoints     = expression(lbl_endpoints()),
    lbl_midpoints     = expression(lbl_midpoints())
  )

  test_df <- expand.grid(
    x         = x_vals,
    brk_fun   = names(brk_funs),
    lbl_fun   = names(lbl_funs),
    # we translate NA to NULL in chop(); doing this means we don't need a list():
    extend    = c(TRUE, FALSE, NA),
    left      = c(TRUE, FALSE),
    close_end = c(TRUE, FALSE),
    # ditto:
    raw       = c(TRUE, FALSE, NA),
    drop      = c(TRUE, FALSE),
    stringsAsFactors = FALSE
  )

  # remove some pointless conditions:

  skip_test <- function (cond) {
    cond <- substitute(cond)
    test_df <<- test_df[with(test_df, ! eval(cond)), ]
  }

  skip_test(! left & brk_fun == "brk_manual")
  skip_test(! close_end & brk_fun == "brk_manual")

  POSIXct_breaks <- c("brk_def_POSIXct", "brk_w_difft_sec")
  Date_breaks <- c("brk_def_Date", "brk_w_difft_day")
  skip_test(names(x) %in% c("Date", "POSIXct")  &
              ! brk_fun %in% c(Date_breaks, POSIXct_breaks))
  skip_test(! names(x) %in% c("Date", "POSIXct") &
              brk_fun %in% c(Date_breaks, POSIXct_breaks))
  # don't try to break dates by 1 second width (very slow!)
  skip_test(names(x) != "POSIXct" & brk_fun == "brk_w_difft_sec")

  test_df$expect <- "succeed"
  test_df$row <- seq_len(nrow(test_df))

  # some things should fail
  should_fail <-   function (cond) test_df$expect[cond] <<- "error"
  should_warn <-   function (cond) test_df$expect[cond] <<- "warn"
  should_either <- function (cond) test_df$expect[cond] <<- "either"
  dont_care <-     function (cond) test_df <<- test_df[! cond, ]

  should_fail(names(test_df$x) == "char")

  # but if we break by quantities, OK...
  char_by_quantities <- names(test_df$x) == "char" &
          test_df$brk_fun %in% c("brk_equally", "brk_quantiles", "brk_n",
                                 "brk_n_merge")
  # so long as we aren't trying raw midpoints
  raw <- ! is.na(test_df$raw) & test_df$raw
  should_warn(char_by_quantities & !
                (test_df$lbl_fun == "lbl_midpoints" & raw)
  )
  # ... or midpoints with brk_n()
  should_fail(char_by_quantities & test_df$lbl_fun == "lbl_midpoints"
              & test_df$brk_fun %in% c("brk_n", "brk_n_merge"))

  # brk_default_hi and _lo have a single break, so if you can't
  # extend it, there are no possible intervals:
  should_fail(with(test_df,
          brk_fun %in% c("brk_default_hi", "brk_default_lo") &
          extend == FALSE
        ))

  # ditto when extend is NULL and there's no non-NA data
  # here we have to fail even though with some data we'd be OK
  should_fail(with(test_df,
                    brk_fun %in% c("brk_default_hi", "brk_default_lo") &
                    names(x) %in% c("all_NAs", "none") &
                    is.na(extend)
  ))

  # raw endpoints get duplicated if multiple quantiles are infinite:
  dont_care(with(test_df,
                   names(x) %in% c("inf_lo", "inf_hi") &
                   brk_fun == "brk_quantiles" &
                   lbl_fun == "lbl_midpoints" &
                   raw == TRUE &
                   extend == TRUE &
                   close_end == FALSE
                 ))
  dont_care(with(test_df,
                   names(x) == "inf_lo" &
                   brk_fun == "brk_quantiles" &
                   lbl_fun == "lbl_endpoints" &
                   raw == TRUE &
                   extend == TRUE &
                   left == FALSE &
                   close_end == FALSE
                 ))

  # lbl_endpoints() can create duplicates
  # when you extend an open interval to add a singleton
  # e.g. {1}, (1, 2]
  dont_care(with(test_df,
                   lbl_fun == "lbl_endpoints" &
                   left == FALSE & is.na(extend)
                 ))
  dont_care(with(test_df,
                   lbl_fun == "lbl_endpoints" &
                   brk_fun %in% c("brk_default_lo", "brk_manual") &
                   left == TRUE & is.na(extend)
                 ))

  # quantiles here likely to create duplicate endpoints
  dont_care(with(test_df,
                 names(x) %in% c("one", "same", "char") &
                 lbl_fun == "lbl_endpoints" &
                 brk_fun == "brk_quantiles" &
                 extend == TRUE & raw == TRUE
               ))

  # brk_quantiles() should warn on duplicate quantiles
  should_warn(with(test_df,
                 names(x) %in% c("one", "same") &
                 brk_fun == "brk_quantiles"
               ))

  # brk_default has breaks 1,2,2,3
  # with lbl_endpoints, this may create duplicate left endpoints
  # ie the user asked for something we can't do
  dont_care(with(test_df,
          names(x) %in%
            c("ordinary", "inf", "inf_lo", "inf_hi", "NaN", "NAs") &
          brk_fun == "brk_default2" &
          lbl_fun == "lbl_endpoints"
        ))
  dont_care(with(test_df,
          brk_fun == "brk_default2" &
          lbl_fun == "lbl_endpoints" &
          drop == FALSE
        ))
  dont_care(with(test_df,
          brk_fun == "brk_spikes" &
          lbl_fun == "lbl_endpoints" &
          drop == FALSE
        ))
  dont_care(with(test_df,
          brk_fun %in% c("brk_n", "brk_n_merge") &
          lbl_fun == "lbl_endpoints"
        ))

  # lbl_midpoints struggles with Inf for obvious reasons
  dont_care(with(test_df,
          names(x) %in% c("inf", "inf_lo", "inf_hi") &
          brk_fun %in% c("brk_n", "brk_n_merge") &
          lbl_fun == "lbl_midpoints"
        ))

  should_fail(names(test_df$x) == "complex")

  # we sample the same 10000 rows every day
  seed <- as.numeric(Sys.Date())
  set.seed(seed)
  test_everything <- isTRUE(as.logical(Sys.getenv("CI"))) ||
                              getOption("santoku.test_everything", FALSE)

  sample_rows <- if (test_everything) {
                   seq_len(nrow(test_df))
                 } else {
                   sort(sample(nrow(test_df), 10000, replace = FALSE))
                 }

  for (r in sample_rows) {
    tdata <- test_df[r, ]
    if (is.na(tdata$expect)) next

    # v basic debugging interactively. Replace r by the row that gives a test failure
    # cat(r, "\n")
    # if (r==63194) browser()
    if (is.na(tdata$extend)) tdata$extend <- NULL
    if (is.na(tdata$raw)) tdata$raw <- NULL

    x <- tdata$x[[1]]
    format_null <- function (x) if (is.null(x)) "NULL" else x
    info <- sprintf(
          "seed: %s row: %s
          command: chop(%s, %s, labels = %s, extend = %s, left = %s,
                     close_end = %s, raw = %s, drop = %s)",
          seed, tdata$row, tdata$x, as.character(brk_funs[[tdata$brk_fun]]),
          as.character(lbl_funs[[tdata$lbl_fun]]), format_null(tdata$extend),
          tdata$left, tdata$close_end, format_null(tdata$raw), tdata$drop)

    # NA means "no error":
    regexp <- switch(tdata$expect, "succeed" = NA, NULL)
    err_class <- switch(tdata$expect, "warn" = "warning", "either" = NULL, "error")
    exp_fn <- if (tdata$expect == "error") expect_error else expect_condition
    # suppressWarnings or we drown in them:

    suppressWarnings(exp_fn(
            chop(!!x,
              breaks    = eval(brk_funs[[!!tdata$brk_fun]]),
              labels    = eval(lbl_funs[[!!tdata$lbl_fun]]),
              extend    = !!tdata$extend,
              left      = !!tdata$left,
              close_end = !!tdata$close_end,
              raw       = !!tdata$raw,
              drop      = !!tdata$drop
            ),
            regexp = regexp,
            class  = err_class,
            info   = info
          ))
  }
})

test_that("brk_manual", {
  for (l in c(TRUE, FALSE)) for (r in c(TRUE, FALSE)) {
    expect_silent(x <- brk_res(brk_manual(1:2, c(l, r))))
    expect_s3_class(x, "breaks")
  }

  expect_error(brk_res(brk_manual(c(2, 2), c(TRUE, TRUE))))
  expect_error(brk_res(brk_manual(c(2, 2), c(FALSE, TRUE))))
  expect_error(brk_res(brk_manual(c(2, 2), c(FALSE, FALSE))))
  expect_silent(brk_res(brk_manual(c(2, 2), c(TRUE, FALSE))))

  expect_error(brk_res(brk_manual(1, c(TRUE, FALSE))))
  expect_error(brk_res(brk_manual(1:2, c(TRUE))))
  expect_error(brk_res(brk_manual("a", TRUE)))
  expect_error(brk_res(brk_manual(1, "c")))

  expect_error(brk_res(brk_manual(c(1, NA), c(TRUE, TRUE))))
  expect_error(brk_res(brk_manual(2:1, c(TRUE, TRUE))))

  expect_error(brk_res(brk_manual(c(1, 2, 2, 2, 3), rep(TRUE, 5))),
        regexp = "equal")

  expect_warning(brk_res(brk_manual(1:3, rep(TRUE, 3)), left = FALSE))
  expect_warning(brk_res(brk_manual(1:3, rep(TRUE, 3)), close_end = FALSE))
})


test_that("brk_n", {
  for (i in 1:10) {
    x <- rnorm(sample(10:20, 1L))
    b <- sample(5L, 1L)
    expect_true(all(tab(!!x, brk_n(!!b), drop = TRUE) <= !!b),
          info = sprintf("length(x) %s b %s", length(x), b))
    # right-closed breaks
    expect_true(all(tab(!!x, brk_n(!!b), drop = TRUE, left = FALSE) <= !!b),
                info = sprintf("length(x) %s b %s left = FALSE", length(x), b))
  }

  # test with duplicates in x
  for (i in 1:10) {
    x <- rnorm(10)
    x <- sample(x, replace = TRUE)
    b <- sample(5L, 1L)
    tbl <- tab(x, brk_n(b), drop = TRUE)
    # all but the last category should have size >= b
    expect_true(all(tbl[-length(tbl)] >= b),
          info = sprintf("length(x) %s b %s", length(x), b))
    # right-closed breaks
    tbl <- tab(x, brk_n(b), drop = TRUE, left = FALSE)
    expect_true(all(tbl[-1] >= b),
          info = sprintf("length(x) %s b %s", length(x), b))
  }
})


test_that("brk_n, tail = 'merge'", {
  x <- 1:5
  res <- brk_res(brk_n(3, tail = "merge"), x = x)
  expect_equal(as.vector(tab(x, res)), 5)

  x <- 1:6
  res <- brk_res(brk_n(3, tail = "merge"), x = x)
  expect_equal(as.vector(tab(x, res)), c(3, 3))

  x <- 1:7
  res <- brk_res(brk_n(3, tail = "merge"), x = x)
  expect_equal(as.vector(tab(x, res)), c(3, 4))

  x <- c(1, 1, 1, 2, 2)
  res <- brk_res(brk_n(3, tail = "merge"), x = x)
  expect_equal(as.vector(tab(x, res)), 5)

  x <- c(1, 1, 1, 2, 2, 2)
  res <- brk_res(brk_n(3, tail = "merge"), x = x)
  expect_equal(as.vector(tab(x, res)), c(3, 3))

  x <- c(1, 1, 1, 2, 2, 2, 2)
  res <- brk_res(brk_n(3, tail = "merge"), x = x)
  expect_equal(as.vector(tab(x, res)), c(3, 4))
})


test_that("bugfix: brk_n shouldn't error with too many non-unique values", {
  expect_error(
    brk_res(brk_n(2), x = c(1, 1, 1, 1, 5, 5, 5, 5)),
    regexp = NA
  )
})


test_that("bugfix: brk_n shouldn't take too few elems after non-unique values", {
  x <- c(1, 1, 1, 1, 2, 3, 4)
  res <- brk_res(brk_n(3), x = x)
  expect_equal(as.vector(tab(x, res)), c(4, 3))

  x <- c(1, 2, 3, 3, 4, 5, 6)
  res <- brk_res(brk_n(3), x = x)
  expect_equal(as.vector(tab(x, res)), c(4, 3))

  x <- c(1, 2, 3, 3, 4)
  res <- brk_res(brk_n(2), x = x)
  expect_equal(as.vector(tab(x, res)), c(2, 2, 1))
})


test_that("brk_width", {
  b <- brk_res(brk_width(1), 0.5:1.5)
  expect_equal(diff(as.vector(b)), 1)

  width <- runif(1)
  b <- brk_res(brk_width(width), 0.5:1.5)
  bvec <- as.vector(b)
  expect_equal(diff(bvec)[1], width)
  expect_equal(bvec[1], 0.5)

  b <- brk_res(brk_width(1), rep(NA, 2))
  expect_identical(as.vector(b), c(-Inf, Inf))

  b <- brk_res(brk_width(1), c(Inf, -Inf, NA))
  expect_identical(as.vector(b), c(-Inf, Inf))

  b <- brk_res(brk_width(1), c(NA, 2, 4, NA))
  expect_equal(diff(as.vector(b))[1], 1)
})


test_that("brk_width, negative width", {
  b <- brk_res(brk_width(-1), 0.5:1.5)
  expect_equal(diff(as.vector(b)), 1)

  width <- runif(1, min = -1, max = 0)
  b <- brk_res(brk_width(width), 0.5:1.5)
  bvec <- as.vector(b)
  expect_equal(diff(bvec)[1], -width)
  expect_equal(bvec[length(bvec)], 1.5)

  b <- brk_res(brk_width(-2, start = 2.5), 0:4)
  expect_identical(as.vector(b), c(-1.5, 0.5, 2.5))
})


test_that("brk_evenly", {
  b <- brk_res(brk_evenly(5), 0:10)
  expect_identical(as.vector(b), c(0, 2, 4, 6, 8, 10))
})


test_that("brk_proportions", {
  b <- brk_res(brk_proportions(c(0.2, 0.8)), 0:10)
  expect_identical(as.vector(b), c(2, 8))

  expect_error(brk_proportions(c(0, 1, 2)))
  expect_error(brk_proportions(c(-1, 0.5)))
  expect_error(brk_proportions(c(0.5, NA)))
})


test_that("brk_mean_sd", {
  x <- rnorm(100)
  expect_silent(b <- brk_res(brk_mean_sd(1:3), x = x))
  m <- mean(x)
  sd <- sd(x)
  sd_ints <- seq(m - 3 * sd, m + 3 * sd, sd)
  expect_equal(as.numeric(b), sd_ints)

  expect_silent(brk_res(brk_mean_sd(1:3), x = rep(NA, 2)))
  expect_silent(brk_res(brk_mean_sd(1:3), x = rep(1, 3)))
  expect_silent(brk_res(brk_mean_sd(1:3), x = 1))

  lifecycle::expect_deprecated(res <- brk_res(brk_mean_sd(sd = 3)))
  expect_equivalent(
    res, brk_res(brk_mean_sd(1:3))
  )
})


test_that("brk_quantiles", {
  expect_silent(brk_res(brk_quantiles(1:3/4)))

  x <- 1:10
  brks <- brk_quantiles(1:3/4)(x, FALSE, TRUE, FALSE)
  expect_equivalent(c(brks), quantile(x, 1:3/4))

  expect_silent(brks <- brk_quantiles(numeric(0))(x, TRUE, TRUE, FALSE))
  expect_equivalent(c(brks), c(-Inf, Inf))

  x <- 1:10
  brks <- brk_quantiles(1:3/4, weights = 1:10)(x, FALSE, TRUE, FALSE)
  expect_equivalent(c(brks), Hmisc::wtd.quantile(x, weights = 1:10, probs = 1:3/4))
})


test_that("brk_quantiles with duplicate quantiles", {
  x <- rep(1, 5)
  expect_warning(brks <- brk_quantiles(1:3/4)(x, FALSE, TRUE, FALSE))
  expect_equivalent(c(brks), c(1, 1))

  x <- c(1, 1, 2, 3, 4)
  expect_warning(brks <- brk_quantiles(0:5/5)(x, FALSE, TRUE, FALSE))
  expect_equivalent(c(brks), c(1.0, 1.0, 1.6, 2.4, 3.2, 4.0))

  x <- c(1, 1, 1, 2, 3)
  expect_warning(brks <- brk_quantiles(0:5/5)(x, FALSE, TRUE, FALSE))
  expect_equivalent(c(brks), c(1.0, 1.0, 1.4, 2.2, 3.0))

  expect_silent(
    brks <- brk_quantiles(0:5/5, recalc_probs = TRUE)(x, FALSE, TRUE, FALSE)
  )
  expect_equivalent(c(brks), c(1.0, 1.0, 1.4, 2.2, 3.0))
})


test_that("brk_equally", {
  expect_silent(brk_res(brk_equally(5)))
  expect_error(brk_equally(4.5))

  brks <- brk_res(brk_equally(3))
  expect_equivalent(brks, brk_res(brk_quantiles(0:3/3)))
})


test_that("brk_equally warns when too few breaks created", {
  dupes <- rep(1, 4)
  expect_warning(brk_res(brk_equally(4), x = dupes))
})


test_that("brk_pretty", {
  expect_silent(brks <- brk_res(brk_pretty(5), x = 1:10))
  expect_equivalent(brks, brk_res(brk_default(pretty(1:10)), x = 1:10))

  expect_silent(brks2 <- brk_res(brk_pretty(5, high.u.bias = 0), x = 1:10))
  expect_equivalent(
    brks2,
    brk_res(brk_default(pretty(1:10, high.u.bias = 0)), x = 1:10)
  )
})


test_that("brk_fn", {
  x <- 1:10
  expect_silent(
    brks <- brk_res(brk_fn(scales::breaks_extended(5)), x = x)
  )
  expect_equivalent(
    brks,
    brk_res(brk_default(scales::breaks_extended(5)(x)))
  )

  expect_silent(
    brks2 <- brk_res(brk_fn(pretty, n = 10), x = x)
  )
  expect_equivalent(
    brks2,
    brk_res(brk_default(pretty(x, n = 10)), x = x)
  )
})


test_that("brk_spikes", {
  x <- c(rep(5, 5), 1:10)

  expect_silent(
    brks <- brk_res(brk_spikes(c(2, 8), n = 5), x = x)
  )
  expect_equivalent(
    c(brks),
    c(2, 5, 5, 8)
  )

  expect_silent(
    brks2 <- brk_res(brk_spikes(c(2, 5, 8), n = 5), x = x)
  )
  expect_equivalent(
    c(brks2),
    c(2, 5, 5, 8)
  )

  expect_silent(
    brks3 <- brk_res(brk_spikes(c(2, 5, 5, 8), n = 5), x = x)
  )
  expect_equivalent(
    c(brks3),
    c(2, 5, 5, 8)
  )

  x2 <- c(1, 1, 1, 2, 3, 3, 4, 4, 4, 4, 5, 6, 7, 8, 8, 8, 9)
  expect_silent(
    brks3.1 <- brk_res(brk_spikes(c(2, 8), n = 3), x = x2)
  )
  expect_equivalent(
    c(brks3.1),
    c(1, 1, 2, 4, 4, 8, 8)
  )

  expect_silent(
    brks4 <- brk_res(brk_spikes(brk_width(5), n = 5), x = x)
  )
  expect_equivalent(
    c(brks4),
    c(1, 5, 5, 6, 11)
  )

  expect_silent(
    brks5 <- brk_res(brk_spikes(c(2, 8), prop = 0.4), x = x)
  )
  expect_equivalent(
    c(brks5),
    c(2, 5, 5, 8)
  )

  expect_error(
    brk_spikes(c(2, 8), n = 5, prop = 0.5)
  )

  expect_error(
    brk_spikes(c(2, 8))
  )
})


test_that("printing", {
  b <- brk_res(brk_default(1:3))
  expect_output(print(b))
  expect_silent(format(b))
  b_empty <- brk_res(brk_default(1))
  expect_output(print(b_empty))
})

#' Replaces labels with names from the breaks vector
#'
#' Only non-zero-char names are used.
#'
#' @param labels Passed in from chop, possibly via a `lbl_*` function
#' @param breaks Breaks object created via a `brk_*` function. Some of
#'   these preserve names of a given argument (`brk_default()`,
#'   `brk_proportions()`, `brk_quantiles()`)
#'
#' @return The altered labels
#' @noRd
#'
add_break_names <- function(labels, breaks) {
  if (is.null(names(breaks))) return(labels)

  is_named <- nzchar(names(breaks))
  # These are possibly-extended breaks; last break is the rightmost endpoint
  # and any name is ignored:
  is_named[length(is_named)] <- FALSE
  break_names_for_labels <- names(breaks)[is_named]

  # length(labels) == length(breaks) - 1
  is_named <- is_named[-length(is_named)]
  labels[is_named] <- break_names_for_labels

  return(labels)
}


#' Return formatted strings for endpoints
#'
#' Methods will pick up a `scaled_endpoints`
#' attribute if one exists. This provides the numbers
#' for when `raw = FALSE`.
#'
#' Different breaks subclasses may have different default formats.
#'
#' A `.numeric` method exists so that formatted labels can be created
#' from e.g. midpoints or other things that aren't breaks
#' themselves.
#'
#' @param breaks Either a breaks object, or a numeric vector
#' @param raw Report raw numbers instead of e.g. quantiles?
#' @param fmt Format string or function
#' @param ... Not used
#'
#' @return A character vector of break endpoints.
#' @noRd
#'
endpoint_labels <- function (breaks, raw, fmt = NULL, ...) {
  UseMethod("endpoint_labels")
}


#' @export
endpoint_labels.numeric <- function (breaks, raw, fmt = NULL, ...) {
  endpoints <- scaled_endpoints(breaks, raw = raw)

  elabels <- if (! is.null(fmt)) {
    apply_format(fmt, endpoints)
  } else {
    unique_truncation(endpoints)
  }

  elabels[is.infinite(endpoints)] <- sub("Inf ?", symbol_infinity(),
                                         elabels[is.infinite(endpoints)])

  return(elabels)
}


#' @export
endpoint_labels.integer <- endpoint_labels.numeric

#' @export
endpoint_labels.double <- endpoint_labels.numeric


#' @export
endpoint_labels.default <- function (breaks, raw, fmt = NULL, ...) {
  endpoints <- scaled_endpoints(breaks, raw = raw)

  elabels <- if (! is.null(fmt)) {
    apply_format(fmt, endpoints)
  } else {
    base::format(endpoints)
  }

  return(elabels)
}


#' @export
endpoint_labels.Date <- function (breaks, raw, fmt = NULL, ...) {
  elabels <- scaled_endpoints(breaks, raw = raw)
  # this could be a number. If so, a `fmt` for `sprintf`
  # will work fine:
  if (! inherits(elabels, "Date")) return(NextMethod())

  # set default format
  fmt <- fmt %||% "%F"

  elabels_chr <- apply_format(fmt, elabels)
  minus_inf <- is.infinite(elabels) & elabels < as.Date("1970-01-01")
  plus_inf  <- is.infinite(elabels) & elabels > as.Date("1970-01-01")
  elabels_chr[minus_inf] <- symbol_infinity(minus = TRUE)
  elabels_chr[plus_inf]  <- symbol_infinity()

  elabels_chr
}


#' @export
endpoint_labels.POSIXt <- function (breaks, raw, fmt = NULL, ...) {
  elabels <- scaled_endpoints(breaks, raw = raw)
  # same comment as endpoint_labels.Date above:
  if (! inherits(elabels, "POSIXt")) return(NextMethod())

  # set default format
  fmt <- fmt %||% "%F %H:%M:%S"

  elabels_chr <- apply_format(fmt, elabels)
  minus_inf <- is.infinite(elabels) & elabels < as.POSIXct("1970-01-01")
  plus_inf  <- is.infinite(elabels) & elabels > as.POSIXct("1970-01-01")
  elabels_chr[minus_inf] <- symbol_infinity(minus = TRUE)
  elabels_chr[plus_inf]  <- symbol_infinity()

  elabels_chr
}


#' @export
endpoint_labels.quantileBreaks <- function (breaks, raw, fmt = NULL, ...) {
  if (raw) return(NextMethod())

  # set default format
  fmt <- fmt %||% percent

  elabels <- scaled_endpoints(breaks, raw = FALSE)
  elabels <- apply_format(fmt, elabels)

  return(elabels)
}


#' @export
endpoint_labels.sdBreaks <- function (breaks, raw, fmt = NULL, ...) {
  if (raw) return(NextMethod())

  # set default format
  fmt <- fmt %||% "%.3g sd"

  elabels <- scaled_endpoints(breaks, raw = FALSE)
  elabels <- apply_format(fmt, elabels)

  return(elabels)
}


#' Return numeric (or whatever) endpoints of breaks, possibly scaled
#'
#' @param breaks Breaks or numeric object
#' @param raw Logical. If `FALSE`, return endpoints scaled as e.g. sds or
#'   quantiles
#'
#' @return Numbers, dates, etc. with no `breaks` class.
#' @noRd
scaled_endpoints <- function (breaks, raw) {
  UseMethod("scaled_endpoints")
}


#' @export
scaled_endpoints.breaks <- function (breaks, raw) {
  if (raw) {
    unclass_breaks(breaks)
  } else {
    attr(breaks, "scaled_endpoints") %||% unclass_breaks(breaks)
  }
}


#' @export
scaled_endpoints.default <- function (breaks, raw) {
  if (raw) {
    breaks
  } else {
    attr(breaks, "scaled_endpoints") %||% breaks
  }
}


#' Apply `fmt` to an object
#'
#' @param fmt A one-argument function, or a character string.
#' @param endpoint Endpoints of a break. Various classes.
#'
#' @return A character vector.
#' @noRd
apply_format <- function (fmt, endpoint, ...) {
  UseMethod("apply_format")
}


#' @export
apply_format.function <- function (fmt, endpoint, ...) {
  fmt(endpoint, ...)
}


#' @export
#' @method apply_format character
apply_format.character <- function (fmt, endpoint, ...) {
  UseMethod("apply_format.character", endpoint)
}


#' @export
#' @method apply_format.character default
apply_format.character.default <- function (fmt, endpoint, ...) {
  base::format(endpoint, fmt, ...)
}


#' @export
#' @method apply_format.character numeric
apply_format.character.numeric <- function (fmt, endpoint, ...) {
  sprintf(fmt, endpoint, ...)
}


#' @export
#' @method apply_format.character character
apply_format.character.character <- function (fmt, endpoint, ...) {
  sprintf(fmt, endpoint, ...)
}


#' @export
#' @method apply_format list
apply_format.list <- function (fmt, endpoint, ...) {
  UseMethod("apply_format.list", endpoint)
}


#' @export
#' @method apply_format.list default
apply_format.list.default <- function (fmt, endpoint, ...) {
  do.call(base::format, c(list(x = endpoint), fmt))
}


is_format <- function (fmt) is.string(fmt) || is.function(fmt) || is.list(fmt)

on_failure(is_format) <- function(call, env) {
  paste0(deparse(call$fmt), " is not a valid format (a string, list or function)")
}


#' Truncates `num` to look nice, while preserving uniqueness
#'
#' @param num A numeric vector.
#'
#' @return A character vector
#' @noRd
unique_truncation <- function (num) {
  want_unique <- ! duplicated(num) # "real" duplicates are allowed!
                                   # we keep the first of each duplicate set.

  for (digits in seq(4L, 22L)) {
    res <- formatC(num, digits = digits, width = -1L)
    if (anyDuplicated(res[want_unique]) == 0L) return(res)
  }

  stop("Could not format breaks to avoid duplicates")
}


em_dash <- function () {
  if (l10n_info()[["UTF-8"]]) "\u2014" else "-"
}


symbol_infinity <- function (minus = FALSE) {
  infty <- if (l10n_info()[["UTF-8"]]) "\u221e" else "Inf"
  infty <- getOption("santoku.infinity", infty)
  if (minus) paste0("-", infty) else infty
}

#' Parse a `strftime` format string
#'
#' Splits a format string into literal and directive tokens.
#'
#' @param fmt A `strftime` format string.
#'
#' @return A data frame with columns `type` (`"literal"` or `"directive"`)
#'   and `token`.
#' @noRd
parse_strftime <- function(fmt) {
  assert_that(is.string(fmt))

  chars <- strsplit(fmt, "", fixed = TRUE)[[1]]
  n <- length(chars)
  i <- 1L
  types <- character(0)
  tokens <- character(0)

  while (i <= n) {
    if (chars[i] != "%") {
      start <- i
      while (i <= n && chars[i] != "%") i <- i + 1L
      types <- c(types, "literal")
      tokens <- c(tokens, paste0(chars[start:(i - 1L)], collapse = ""))
      next
    }

    if (i < n && chars[i + 1L] == "%") {
      types <- c(types, "literal")
      tokens <- c(tokens, "%")
      i <- i + 2L
      next
    }

    start <- i
    i <- i + 1L

    if (i <= n && chars[i] %in% c("E", "O")) i <- i + 1L
    while (i <= n && grepl("[0-9]", chars[i])) i <- i + 1L
    if (i <= n) i <- i + 1L

    types <- c(types, "directive")
    tokens <- c(tokens, paste0(chars[start:(i - 1L)], collapse = ""))
  }

  data.frame(type = types, token = tokens, stringsAsFactors = FALSE)
}

#' Return relative component rank for a strftime directive
#'
#' Lower values are "greater" components, e.g. year before month before day.
#'
#' @param token A single strftime directive token (e.g. "%Y").
#'
#' @return Numeric rank, or `NA_real_` if unknown.
#' @noRd
strftime_rank <- function(token) {
  code <- gsub("^%[EO]?([0-9]*)", "", token)

  if (code %in% c("Y", "y", "C", "G", "g")) return(1)
  if (code %in% c("m", "b", "B", "h")) return(2)
  if (code %in% c("d", "e", "j", "a", "A", "u", "w", "p", "P")) return(3)
  if (code %in% c("H", "I", "k", "l", "M", "S", "OS")) return(4)

  NA_real_
}


#' Format date/time endpoints into token matrix
#'
#' @param x Date/time-like vector.
#' @param fmt A `strftime` format string.
#' @param spec Parsed strftime spec.
#'
#' @return A character matrix: rows are endpoints, columns are format tokens.
#' @noRd
format_strftime_tokens <- function(x, fmt = NULL, spec = parse_strftime(fmt)) {
  n <- length(x)

  tokens <- lapply(seq_len(nrow(spec)), function(i) {
    if (identical(spec$type[[i]], "literal")) {
      rep(spec$token[[i]], n)
    } else {
      format(x, spec$token[[i]])
    }
  })

  do.call(cbind, tokens)
}


#' Collapse two formatted date/time labels
#'
#' @param l_tokens Left endpoint tokens.
#' @param r_tokens Right endpoint tokens.
#' @param symbol Separator for full ranges.
#'
#' @return A single collapsed range label.
#' @noRd
collapse_datetime_label <- function(
    l_tokens,
    r_tokens,
    spec,
    symbol = "-"
) {
  spaced_symbol <- paste0(" ", symbol, " ")
  full <- paste0(paste0(l_tokens, collapse = ""), spaced_symbol,
                 paste0(r_tokens, collapse = ""))

  ranks <- vapply(spec$token, strftime_rank, FUN.VALUE = numeric(1))
  is_directive <- spec$type == "directive"
  comparable <- is_directive & !is.na(ranks)
  differs <- comparable & (l_tokens != r_tokens)

  if (!any(differs)) return(full)

  diff_rank <- min(ranks[differs])
  collapse_rank <- diff_rank

  # If both day and time differ, keep month with both endpoints so that
  # day labels remain anchored to a month (issue #58).
  if (diff_rank == 3 && any(differs & ranks > diff_rank)) {
    collapse_rank <- 2
  }

  higher_differs <- comparable & (ranks < diff_rank) & (l_tokens != r_tokens)
  active_components <- which(comparable & (ranks >= collapse_rank))

  if (any(higher_differs) || length(active_components) == 0 || diff_rank == 1) {
    return(full)
  }

  left_end <- max(active_components)
  right_start <- min(active_components)
  left_part <- paste0(l_tokens[seq_len(left_end)], collapse = "")
  right_part <- paste0(r_tokens[right_start:length(r_tokens)], collapse = "")

  if (!nzchar(left_part) || !nzchar(right_part)) return(full)

  # joiner <- if (diff_rank %in% 3:4) symbol else spaced_symbol
  joiner <- if (ranks[right_start] == ranks[left_end]) symbol else spaced_symbol
  paste0(left_part, joiner, right_part)
}


#' @rdname lbl_datetime
#' @export
lbl_date <- function(
    fmt = "%e %b %Y",
    symbol = "-",
    unit = as.difftime(1, units = "days"),
    single = "{l}",
    first = NULL,
    last = NULL
) {
  lbl_datetime(
    fmt = fmt,
    symbol = symbol,
    unit = unit,
    single = single,
    first = first,
    last = last
  )
}


#' Label dates and datetimes
#'
#' @description
#' `r lifecycle::badge("experimental")`
#'
#' `lbl_date()` and `lbl_datetime()` produce nice labels for dates
#' and datetimes. Where possible ranges are simplified, like
#' like "13-14 Jul 2026" or "11:15-12:15 1 Dec 2025".
#'
#' @inherit label-doc
#' @inherit first-last-doc
#' @param symbol String: separator to use for full ranges.
#' @param unit Optional interval unit for non-overlapping labels. If not `NULL`,
#'.  endpoints are adjusted in the style of [lbl_discrete()].
#'
#' @family labelling functions
#'
#' @export
#'
#' @examples
#' winter <- as.Date("2025-12-01") + 0:89
#' tab(winter, as.Date(c("2025-12-25", "2026-01-06")),
#'     labels = lbl_date())
#' new_year <- as.POSIXct("2025-12-31 23:00") + 0:120 * 60
#' round_midnight <- as.POSIXct(c("2025-12-31 23:59", "2026-01-01 00:05"))
#' tab(new_year, round_midnight,
#'     labels = lbl_datetime())
#' tab(new_year, round_midnight,
#'     labels = lbl_datetime(unit = as.difftime(1, units = "mins")))
lbl_datetime <- function(
    fmt = "%H:%M:%S %b %e %Y",
    symbol = "-",
    unit = NULL,
    single = "{l}",
    first = NULL,
    last = NULL
) {
  assert_that(
    is.string(fmt),
    is.string(symbol),
    length(unit) <= 1L,
    is.string(single) || is.null(single),
    is.string(first) || is.null(first),
    is.string(last) || is.null(last)
  )

  function(breaks, raw = FALSE) {
    assert_that(is.breaks(breaks))

    len_breaks <- length(breaks)
    endpoints <- scaled_endpoints(breaks, raw = raw)
    pieces <- discrete_interval_endpoints(
      breaks = breaks,
      unit = unit,
      endpoints = endpoints
    )
    l <- pieces$l
    r <- pieces$r
    is_singleton <- pieces$singletons
    too_small <- pieces$too_small
    l_closed <- pieces$l_closed
    r_closed <- pieces$r_closed

    if (any(too_small)) {
      warning("Intervals smaller than `unit` are labelled as \"--\"")
    }

    spec <- parse_strftime(fmt)
    l_tokens <- format_strftime_tokens(l, spec = spec)
    r_tokens <- format_strftime_tokens(r, spec = spec)

    labels <- vapply(seq_len(len_breaks - 1L), function(i) {
      collapse_datetime_label(
        l_tokens = l_tokens[i, ],
        r_tokens = r_tokens[i, ],
        spec = spec,
        symbol = symbol
      )
    }, FUN.VALUE = character(1))

    l <- apply(l_tokens, 1, paste0, collapse = "")
    r <- apply(r_tokens, 1, paste0, collapse = "")

    labels[too_small] <- "--"

    if (!is.null(single)) {
      labels[is_singleton] <- glue::glue(single,
        l = l[is_singleton],
        r = r[is_singleton],
        l_closed = l_closed[is_singleton],
        r_closed = r_closed[is_singleton]
      )
    }

    if (!is.null(first)) {
      labels[1] <- glue::glue(first, l = l[1], r = r[1],
        l_closed = l_closed[1], r_closed = r_closed[1]
      )
    }

    if (!is.null(last)) {
      ll <- len_breaks - 1L
      labels[ll] <- glue::glue(last, l = l[ll], r = r[ll],
        l_closed = l_closed[ll], r_closed = r_closed[ll]
      )
    }

    labels
  }
}

#' Chop into fixed-width intervals
#'
#' `chop_width()` chops `x` into intervals of fixed `width`.
#'
#' @param width Width of intervals.
#' @param start Starting point for intervals. By default the smallest
#'   finite `x` (largest if `width` is negative).
#' @inheritParams chop
#' @inherit chop-doc params return
#'
#' @details
#' If `width` is negative, `chop_width()` sets `left = FALSE` and intervals will
#' go downwards from `start`.
#'
#' @family chopping functions
#' @seealso [brk_width-for-datetime]
#'
#' @export
#' @order 1
#'
#' @examples
#' chop_width(1:10, 2)
#'
#' chop_width(1:10, 2, start = 0)
#'
#' chop_width(1:9, -2)
#'
#' chop(1:10, brk_width(2, 0))
#'
chop_width <- function (
                x,
                width,
                start,
                ...,
                left = sign(width) > 0
              ) {
  chop(x, brk_width(width, start), ..., left = left)
}


#' Chop into equal-width intervals
#'
#' `chop_evenly()` chops `x` into `intervals` intervals of equal width.
#'
#' @param intervals Integer: number of intervals to create.
#' @inheritParams chop
#' @inherit chop-doc params return
#'
#' @family chopping functions
#'
#' @export
#' @order 1
#' @examples
#' chop_evenly(0:10, 5)
#'
chop_evenly <- function (
                 x,
                 intervals,
                 ...
               ) {
  chop(x, brk_evenly(intervals), ...)
}


#' Chop into proportions of the range of x
#'
#' `chop_proportions()` chops `x` into `proportions` of its range, excluding
#' infinite values.
#'
#' By default, labels show the raw numeric endpoints. To label intervals by
#' the proportions, use `raw = FALSE`.
#'
#' @param proportions Numeric vector between 0 and 1: proportions of x's range.
#'   If `proportions` has names, these will be used for labels.
#' @inheritParams chop
#' @inherit chop-doc params return
#'
#' @export
#' @order 1
#' @family chopping functions
#' @examples
#' chop_proportions(0:10, c(0.2, 0.8))
#' chop_proportions(0:10, c(Low = 0, Mid = 0.2, High = 0.8))
#'
chop_proportions <- function (
                      x,
                      proportions,
                      ...,
                      raw    = TRUE
                    ) {
  chop(x, brk_proportions(proportions), ..., raw = raw)
}

test_that("character", {
  x <- LETTERS
  br <- c("F", "M")
  expect_warning(
    chop(x, br)
  )

  withr::local_options(santoku.warn_character = FALSE)

  expect_silent(
    chop(x, br)
  )

  # here, we think there should *always* be a warning
  expect_warning(
    chop(x, br, extend = TRUE)
  )

  expect_silent(
    chop_equally(x, 13)
  )

  expect_silent(
    chop_n(x, 13)
  )
})


test_that("ordered", {
  x <- ordered(1:10)
  br <- ordered(c(5, 8), levels = levels(x))

  expect_silent(
    chop(x, br)
  )

  # here, we think there should *always* be a warning
  expect_warning(
    chop(x, br, extend = TRUE)
  )

  expect_silent(
    chop_n(x, 5)
  )

  expect_silent(
    chop_equally(x, groups = 2)
  )
})


test_that("hexmode", {
  x <- as.hexmode(1:10 + 10)
  br <- as.hexmode(c(13, 15, 15, 18))

  expect_silent(
    chop(x, br, extend = FALSE)
  )

  expect_silent(
    chop(x, br)
  )

  # here, there happens to be a warning as of 0.7.0.9000,
  # but we'd be happy if we could represent +/- Inf as hexmode
  suppressWarnings(expect_error(
    chop(x, br, extend = TRUE),
    regexp = NA
  ))
})


test_that("octmode", {
  x <- as.octmode(1:10 + 10)
  br <- as.octmode(c(13, 15, 15, 18))

  expect_silent(
    chop(x, br, extend = FALSE)
  )

  expect_silent(
    chop(x, br)
  )

  expect_silent(
    chop(x, c(12, 15, 15, 18))
  )

  expect_silent(
    chop(1:10 + 10, br)
  )

  suppressWarnings(expect_error(
    chop(x, br, extend = TRUE),
    regexp = NA
  ))
})


test_that("stat::ts", {
  x <- ts(1:10)
  # note: we need to specify integer breaks
  # vec_cast can't cope with ts(<integer>) and ts(<double>)
  br <- c(5L, 8L)

  expect_silent(
    chop(x, br)
  )

  x <- ts(c(1.0, 3.0, 5.0))
  br <- c(2.0, 4.0)
  expect_silent(
    chop(x, br)
  )

  expect_silent(
    chop(x, br, extend = TRUE)
  )

  expect_silent(
    chop_equally(x, groups = 3)
  )

  expect_silent(
    chop_width(x, width = 2)
  )

  x <- ts(1:10)
  br <- ts(c(5.0, 8.0))
  expect_silent(
    chop(x, br)
  )
})


test_that("zoo::zoo", {
  skip_if_not_installed("zoo")

  x <- zoo::zoo(1:10, 1:10)

  expect_silent(
    chop(x, c(3, 5, 5, 7))
  )

  suppressWarnings(expect_error(
    # gives a warning but no error as of 0.7.0.9000
    chop(x, c(3, 5, 5, 7), extend = TRUE),
    regexp = NA
  ))

  expect_silent(
    chop_width(x, 2)
  )

  expect_silent(
    chop_equally(x, 2)
  )
})


test_that("xts::xts", {
  skip_if_not_installed("xts")

  x <- xts::xts(1:10, Sys.Date() + 1:10)

  expect_silent(
    chop(x, c(3, 5, 5, 7))
  )

  suppressWarnings(expect_error(
    # gives a warning but no error as of 0.7.0.9000
    chop(x, c(3, 5, 5, 7), extend = TRUE),
    regexp = NA
  ))

  expect_silent(
    chop_width(x, 2)
  )

  expect_silent(
    chop_equally(x, 2)
  )
})


test_that("units::units", {
  skip_if_not_installed("units")

  x <- units::set_units(1:10, cm)
  br <- units::set_units(c(3, 5, 5, 8), cm)
  br_mm <- units::set_units(c(30, 50, 50, 80), mm)
  expect_silent(
    chop(x, br)
  )

  expect_silent(
    chop(x, br, extend = TRUE)
  )

  expect_equal(
    as.numeric(chop(x, br_mm)),
    c(1, 1, 2, 2, 3, 4, 4, 5, 5, 5)
  )

  br_m2 <- units::set_units(c(3,5,5,8), m^2)
  expect_error(
    chop(x, br_m2)
  )

  expect_silent(
    chopped <- chop_width(x, units::set_units(0.05, m))
  )

  expect_equal(
    as.numeric(chopped), c(rep(1, 5), rep(2, 5))
  )

  start <- units::set_units(20, mm)
  expect_silent(
    chopped <- chop_width(x, units::set_units(0.05, m), start)
  )

  expect_equal(
    as.numeric(chopped), c(1, rep(2, 5), rep(3, 4))
  )

  expect_silent(
    chopped <- chop_evenly(x, intervals = 2)
  )

  expect_equal(
    as.numeric(chopped), c(rep(1, 5), rep(2, 5))
  )

  expect_silent(
    chop_equally(x, 5)
  )

  expect_silent(
    chop_n(x, 3)
  )

  expect_silent(
    chop(x, br, labels = lbl_discrete(unit = units::set_units(1, cm)))
  )

  # we don't support mixed units, since units doesn't support
  # comparison operators on those
})


test_that("package_version", {
  x <- as.package_version(c("0.5", "0.5.1", "1.0", "1.1.1", "1.2.0"))
  br <- as.package_version(c("0.7", "1.1", "1.2.0"))

  expect_silent(
    chop(x, br)
  )

  expect_warning(
    chop(x, br, extend = TRUE)
  )
})


test_that("difftime", {
  days <- as.Date("1970-01-01") + 0:30
  difftimes_d <- days[10:15] - days[12:7]
  difftimes_h <- difftimes_d
  units(difftimes_h) <- "hours"

  expect_silent(
    chop(difftimes_d, difftimes_d[c(3,5)])
  )

  expect_silent(
    chop(difftimes_d, difftimes_h)
  )

  expect_silent(
    chop(difftimes_d, difftimes_d[c(3,5)], extend = TRUE)
  )
})


test_that("bit64", {
  skip_if_not_installed("bit64")

  x64 <- bit64::as.integer64(1:10)
  b64 <- bit64::as.integer64(c(3, 5, 5, 7))

  expect_silent(
    chop(x64, b64)
  )

  expect_silent(
    chopped <- chop(x64, b64, extend = TRUE)
  )
  expect_equivalent(
    as.numeric(chopped),
    c(1, 1, 2, 2, 3, 4, 5, 5, 5, 5)
  )

  expect_silent(
    chop(1:10, b64)
  )

  expect_silent(
    chop(x64, as.integer(c(3, 5, 5, 7)))
  )

  expect_silent(
    chop(x64, c(3, 5, 5, 7))
  )

  expect_silent(
    chop(c(1, 3, 5, 7), b64)
  )

  expect_equivalent(
    chop(x64, c(2.5, 7.5), labels = letters[1:3]),
    factor(c(1, 1, 2, 2, 2, 2, 2, 3, 3, 3), labels = letters[1:3])
  )

  x64_big <- bit64::as.integer64("1000000000000000000") + 1:10
  b64_big <- bit64::as.integer64("1000000000000000000") + c(3, 5, 5, 7)

  expect_silent(
    chop(x64_big, b64_big)
  )

  expect_warning(
    chop(c(bit64::as.integer64(1), x64_big), 2.5)
  )

})


test_that("hms::hms", {
  skip_if_not_installed("hms")

  x <- hms::hms(minutes = 1:180)
  br <- hms::hms(hours = 1:2)

  expect_silent(
    chopped <- chop(x, br)
  )
  expect_equal(
    as.numeric(chopped),
    rep(1:3, c(59, 60, 61)),
    ignore_attr = TRUE
  )

  expect_silent(
    chop(x, br, extend = TRUE)
  )
})


test_that("haven::labelled", {
  skip_if_not_installed("haven")

  x <- haven::labelled(as.double(1:10), c("Lo" = 1, "Hi" = 10))
  br <- haven::labelled(c(3, 5), c("Mid" = 3, "Mid2" = 5))

  expect_silent(
    chop(x, c(2, 5, 5, 8))
  )

  expect_silent(
    chop(x, br)
  )

  expect_silent(
    chop(x, br, extend = TRUE)
  )
})

#' @name label-doc
#' @param fmt String, list or function. A format for break endpoints.
#' @param raw `r lifecycle::badge("deprecated")`. Use the `raw` argument to [chop()]
#'   instead.
#' @param symbol String: symbol to use for the dash.
#' @param ... Arguments passed to format methods.
#'
#' @section Formatting endpoints:
#'
#' If `fmt` is not `NULL` then it is used to format the endpoints.
#'
#' * If `fmt` is a string, then numeric endpoints will be formatted by
#'   `sprintf(fmt, breaks)`; other endpoints, e.g. [Date] objects, will be
#'   formatted by `format(breaks, fmt)`.
#'
#' * If `fmt` is a list, then it will be used as arguments to [format].
#'
#' * If `fmt` is a function, it should take a vector of numbers (or other objects
#'   that can be used as breaks) and return a character vector. It may be helpful
#'   to use functions from the `{scales}` package, e.g. [scales::label_comma()].
#'
#' @return A function that creates a vector of labels.
NULL


#' @name first-last-doc
#' @param single Glue string: label for singleton intervals. See [lbl_glue()]
#'   for details. If `NULL`, singleton intervals will be labelled the same way
#'   as other intervals.
#' @param first Glue string: override label for the first category. Write e.g.
#'   `first = "<{r}"` to create a label like `"<18"`. See [lbl_glue()]
#'   for details.
#' @param last String: override label for the last category. Write e.g.
#'   `last = ">{l}"` to create a label like `">65"`. See [lbl_glue()]
#'   for details.
NULL


#' Label chopped intervals using set notation
#'
#' These labels are the most exact, since they show you whether
#' intervals are "closed" or "open", i.e. whether they include their endpoints.
#'
#' Mathematical set notation looks like this:
#'
#' * \code{[a, b]}: all numbers `x` where `a <= x <= b`;
#' * \code{(a, b)}: all numbers where `a < x < b`;
#' * \code{[a, b)}: all numbers where `a <= x < b`;
#' * \code{(a, b]}: all numbers where `a < x <= b`;
#' * \code{{a}}: just the number `a` exactly.
#'
#' @inherit label-doc
#' @inherit first-last-doc
#'
#' @family labelling functions
#'
#' @export
#'
#' @examples
#'
#' tab(-10:10, c(-3, 0, 0, 3),
#'       labels = lbl_intervals())
#'
#' tab(-10:10, c(-3, 0, 0, 3),
#'       labels = lbl_intervals(fmt = list(nsmall = 1)))
#'
#' tab_evenly(runif(20), 10,
#'       labels = lbl_intervals(fmt = percent))
#'
lbl_intervals <- function (
                   fmt    = NULL,
                   single = "{{{l}}}",
                   first  = NULL,
                   last   = NULL,
                   raw    = FALSE
                 ) {
  if (! isFALSE(raw)) {
    lifecycle::deprecate_warn("0.9.0", "lbl_intervals(raw)", "chop(raw)")
  }
  interval_glue <- "{ifelse(l_closed, '[', '(')}{l}, {r}{ifelse(r_closed, ']', ')')}"
  lbl_glue(label = interval_glue, single = single, fmt = fmt, first = first,
             last = last, raw = raw)
}


#' Label discrete data
#'
#' `lbl_discrete()` creates labels for discrete data, such as integers.
#' For example, breaks
#' `c(1, 3, 4, 6, 7)` are labelled: `"1-2", "3", "4-5", "6-7"`.
#'
#' @inherit label-doc
#' @param unit Minimum difference between distinct values of data.
#'   For integers, 1.
#' @inherit first-last-doc
#'
#' @details
#' No check is done that the data are discrete-valued. If they are not, then
#' these labels may be misleading. Here, discrete-valued means that if
#' `x < y`, then `x <= y - unit`.
#'
#' Be aware that Date objects may have non-integer values. See [Date].
#'
#' @family labelling functions
#'
#' @export
#'
#' @examples
#' tab(1:7, c(1, 3, 5), lbl_discrete())
#'
#' tab(1:7, c(3, 5), lbl_discrete(first = "<= {r}"))
#'
#' tab(1:7 * 1000, c(1, 3, 5) * 1000, lbl_discrete(unit = 1000))
#'
#' # Misleading labels for non-integer data
#' chop(2.5, c(1, 3, 5), lbl_discrete())
#'
lbl_discrete <- function (
                  symbol = em_dash(),
                  unit = 1L,
                  fmt = NULL,
                  single = NULL,
                  first = NULL,
                  last = NULL
                ) {
  assert_that(
          is.string(symbol),
          is.scalar(unit),
          is.null(fmt) || is_format(fmt),
          is.string(single) || is.null(single),
          is.string(first) || is.null(first),
          is.string(last) || is.null(last)
        )

  function (breaks, raw = NULL) {
    assert_that(all(ceiling(as.numeric(breaks)) == floor(as.numeric(breaks))),
          msg = "Non-integer breaks")

    len_breaks <- length(breaks)
    singletons <- singletons(breaks)
    left <- attr(breaks, "left")
    breaks <- unclass_breaks(breaks)

    l <- breaks[-len_breaks]
    r <- breaks[-1]
    left_l <- left[-len_breaks]
    left_r <- left[-1]

    # if you're right-closed we add `unit` to your left endpoint:
    l[! left_l] <- l[! left_l] + unit
    # if you're left-closed we deduct `unit` from your right endpoint:
    r[left_r] <- r[left_r] - unit
    # sometimes this makes the two endpoints the same:
    singletons <- singletons | r == l

    too_small <- r < l
    if (any(too_small)) {
      warning("Intervals smaller than `unit` are labelled as \"--\"")
    }

    labels_l <- endpoint_labels(l, raw = FALSE, fmt = fmt)
    labels_r <- endpoint_labels(r, raw = FALSE, fmt = fmt)

    labels <- paste0(labels_l, symbol, labels_r)
    labels[singletons] <- labels_l[singletons]
    labels[too_small] <- "--"

    l_closed <- left_l
    # r_closed is used for "]" in labels so need to switch it here:
    r_closed <- ! left_r

    if (! is.null(single)) {
      labels[singletons] <- glue::glue(single, l = labels_l[singletons],
                                         r = labels_r[singletons],
                                         l_closed = l_closed[singletons],
                                         r_closed = r_closed[singletons])
    }

    if (! is.null(first)) {
      labels[1] <- glue::glue(first, l = labels_l[1], r = labels_r[1],
                              l_closed = l_closed[1], r_closed = r_closed[1])
    }

    if (! is.null(last)) {
      ll <- len_breaks - 1
      labels[ll] <- glue::glue(last, l = labels_l[ll], r = labels_r[ll],
                               l_closed = l_closed[ll], r_closed = r_closed[ll])
    }

    return(labels)
  }
}


#' Label chopped intervals like 1-4, 4-5, ...
#'
#' This label style is user-friendly, but doesn't distinguish between
#' left- and right-closed intervals. It's good for continuous data
#' where you don't expect points to be exactly on the breaks.
#'
#' If you don't want unicode output, use `lbl_dash("-")`.
#'
#' @inherit label-doc
#' @inherit first-last-doc
#'
#' @family labelling functions
#'
#' @export
#'
#' @examples
#' chop(1:10, c(2, 5, 8), lbl_dash())
#'
#' chop(1:10, c(2, 5, 8), lbl_dash(" to ", fmt = "%.1f"))
#'
#' chop(1:10, c(2, 5, 8), lbl_dash(first = "<{r}"))
#'
#' pretty <- function (x) prettyNum(x, big.mark = ",", digits = 1)
#' chop(runif(10) * 10000, c(3000, 7000), lbl_dash(" to ", fmt = pretty))
lbl_dash <- function (
              symbol = em_dash(),
              fmt    = NULL,
              single = "{l}",
              first  = NULL,
              last   = NULL,
              raw    = FALSE
            ) {
  if (! isFALSE(raw)) {
    lifecycle::deprecate_warn("0.9.0", "lbl_dash(raw)", "chop(raw)")
  }

  label_glue <- paste0("{l}", symbol, "{r}")
  lbl_glue(label = label_glue, fmt = fmt, single = single, first = first,
           last = last, raw = raw)
}

#' Chop common values into singleton intervals
#'
#' `chop_spikes()` lets you chop common values of `x` into their own
#' singleton intervals. This can help make unusual values visible.
#'
#' This function is `r lifecycle::badge("experimental")`.
#'
#' @param breaks A numeric vector of cut-points or a call to a `brk_*` function.
#'   The resulting [`breaks`][breaks-class] object will be modified to add
#'   singleton breaks.
#' @param n,prop Scalar. Provide either `n`, a number of values, or `prop`,
#'   a proportion of `length(x)`. Values of `x` which occur at least this
#'   often will get their own singleton break.
#' @inheritParams chop
#' @inherit chop-doc params return
#'
#' @export
#' @order 1
#' @family chopping functions
#' @seealso [dissect()] for a different approach.
#' @examples
#' x <- c(1:4, rep(5, 5), 6:10)
#' chop_spikes(x, c(2, 7), n = 5)
#' chop_spikes(x, c(2, 7), prop = 0.25)
#' chop_spikes(x, brk_width(5), n = 5)
#'
#' set.seed(42)
#' x <- runif(40, 0, 10)
#' x <- sample(x, 200, replace = TRUE)
#' tab_spikes(x, brk_width(2, 0), prop = 0.05)
chop_spikes <- function (
    x,
    breaks,
    ...,
    n = NULL,
    prop = NULL
) {
  chop(x, brk_spikes(breaks, n = n, prop = prop), ...)
}


#' Cut data into intervals, separating out common values
#'
#' Sometimes it's useful to separate out common elements of `x`.
#' `dissect()` chops `x`, but puts common elements of `x` ("spikes")
#' into separate categories.
#'
#' Unlike [chop_spikes()], `dissect()` doesn't break up
#' intervals which contain a spike. As a result, unlike `chop_*` functions,
#' `dissect()` does not chop `x` into disjoint intervals. See the examples.
#'
#' If breaks are data-dependent, their labels may be misleading after common
#' elements have been removed. See the example below. To get round this,
#' set `exclude_spikes` to `TRUE`. Then breaks will be calculated after
#' removing spikes from the data.
#'
#' Levels of the result are ordered by the minimum element in each level. As
#' a result, if `drop = FALSE`, empty levels will be placed last.
#'
#' This function is `r lifecycle::badge("experimental")`.
#'
#' @param x,breaks,... Passed to [chop()].
#' @inheritParams chop_spikes
#' @param spike_labels [Glue][glue::glue()] string for spike labels. Use `"{l}"`
#' for the spike value.
#' @param exclude_spikes Logical. Exclude spikes before chopping `x`? This
#'   can affect the location of data-dependent breaks.
#'
#' @return
#' `dissect()` returns the result of [chop()], but with common values put into
#' separate factor levels.
#'
#' `tab_dissect()` returns a contingency [table()].
#'
#' @seealso [chop_spikes()] for a different approach.
#' @export
#' @order 1
#'
#' @examples
#' x <- c(2, 3, 3, 3, 4)
#' dissect(x, c(2, 4), n = 3)
#' dissect(x, brk_width(2), prop = 0.5)
#'
#' set.seed(42)
#' x <- runif(40, 0, 10)
#' x <- sample(x, 200, replace = TRUE)
#' # Compare:
#' table(dissect(x, brk_width(2, 0), prop = 0.05))
#' # Versus:
#' tab_spikes(x, brk_width(2, 0), prop = 0.05)
#'
#' # Potentially confusing data-dependent breaks:
#' set.seed(42)
#' x <- rnorm(99)
#' x[1:9] <- x[1]
#' tab_quantiles(x, 1:2/3)
#' tab_dissect(x, brk_quantiles(1:2/3), n = 9)
#' # Calculate quantiles excluding spikes:
#' tab_dissect(x, brk_quantiles(1:2/3), n = 9, exclude_spikes = TRUE)
dissect <- function (x,
                     breaks,
                     ...,
                     n = NULL,
                     prop = NULL,
                     spike_labels = "{{{l}}}",
                     exclude_spikes = FALSE) {
  assert_that(
    is.number(n) || is.number(prop),
    is.null(n) || is.null(prop),
    is.string(spike_labels),
    is.flag(exclude_spikes),
    msg = "exactly one of `n` and `prop` must be a scalar numeric"
  )
  assert_that(
    # it's ok for one of these to be null
    n >= 0 || prop >= 0
  )

  spikes <- find_spikes(x, n, prop)
  x_spikes <- match(x, spikes)
  is_spike <- ! is.na(x_spikes)
  x_spikes <- x_spikes[is_spike]

  if (exclude_spikes) {
    x_not_spikes <- x[! is_spike]
    chopped_not_spikes <- chop(x_not_spikes, breaks, ...)
    chopped <- factor(rep(NA_integer_, length(x)),
                      levels = levels(chopped_not_spikes))
    chopped[! is_spike] <- chopped_not_spikes
  } else {
    chopped <- chop(x, breaks, ...)
  }

  elabels <- endpoint_labels(spikes, raw = TRUE)
  glue_env <- new.env()
  assign("l", elabels, envir = glue_env)
  spike_labels <- glue::glue(spike_labels, .envir = glue_env)

  new_levels <- c(levels(chopped), spike_labels)
  levels(chopped) <- new_levels

  chopped[is_spike] <- spike_labels[x_spikes]

  # We reorder the levels of chopped in order of their smallest elements.
  # Note that if `drop = FALSE`, empty intervals will be at the end.
  # The alternative would be to call `breaks` again and get the left endpoints
  # but this is complex.
  chopped <- stats::reorder(chopped, x, FUN = quiet_min)
  attr(chopped, "scores") <- NULL # remove leftover from reorder()

  chopped
}


#' Find common elements in `x`
#'
#' @param x A vector
#' @param n Number of elements that counts as common. Specify exactly one of `n`
#'   and `prop`.
#' @param prop Proportion of `length(x)` that counts as common
#'
#' @return The common elements, not necessarily in order. NA values are never
#' considered as common.
#' @noRd
find_spikes <- function (x, n, prop) {
  n <- n %||% (length(x) * prop)
  unique_x <- unique(x)
  x_counts <- tabulate(match(x, unique_x))
  spikes <- unique_x[x_counts >= n]
  spikes <- spikes[! is.na(spikes)]

  spikes
}

#' Equal-width intervals for dates or datetimes
#'
#' `brk_width()` can be used with time interval classes from base R or the
#' `lubridate` package.
#'
#' @param width A scalar [difftime], [Period][lubridate::Period-class] or
#'   [Duration][lubridate::Duration-class] object.
#'
#' @param start A scalar of class [Date] or [POSIXct][DateTimeClasses].
#'   Can be omitted.
#'
#' @details
#' If `width` is a Period, [`lubridate::add_with_rollback()`][`lubridate::m+`]
#' is used to calculate the widths. This can be useful for e.g. calendar months.
#'
#' @examples
#'
#' if (requireNamespace("lubridate")) {
#'   year2001 <- as.Date("2001-01-01") + 0:364
#'   tab_width(year2001, months(1),
#'         labels = lbl_discrete(" to ", fmt = "%e %b %y"))
#' }
#'
#' @name brk_width-for-datetime
NULL


#' @rdname chop_width
#' @export
#' @order 2
brk_width <- function (width, start) UseMethod("brk_width")


#' @rdname brk_width-for-datetime
#' @export
brk_width.Duration <- function (width, start) {
  loadNamespace("lubridate")
  width <- lubridate::make_difftime(as.numeric(width))
  NextMethod()
}


#' @rdname chop_width
#' @export
#' @order 2
brk_width.default <- function (width, start) {
  assert_that(is.scalar(width))

  sm <- missing(start)
  if (! sm) assert_that(is.scalar(start))

  function (x, extend, left, close_end) {
    # finite if x has any non-NA finite elements:
    min_x <- quiet_min(x[is.finite(x)])
    max_x    <- quiet_max(x[is.finite(x)])

    if (sm) {
      start <- if (sign(width) > 0) min_x else max_x
    }
    until <- if (sign(width) > 0) max_x else min_x

    if (is.finite(start) && is.finite(until)) {
      breaks <- sequence_width(width, start, until)
    } else {
      return(empty_breaks())
    }

    if (sign(width) <= 0) breaks <- rev(breaks)

    breaks <- create_extended_breaks(breaks, x, extend, left, close_end)

    breaks
  }
}


#' @rdname chop_evenly
#' @export
#' @order 2
brk_evenly <- function(intervals) {
  assert_that(is.count(intervals))

  function (x, extend, left, close_end) {
    min_x <- quiet_min(x[is.finite(x)])
    max_x <- quiet_max(x[is.finite(x)])
    if (sign(max_x - min_x) <= 0) return(empty_breaks())

    breaks <- seq(min_x, max_x, length.out = intervals + 1L)
    breaks <- create_extended_breaks(breaks, x, extend, left, close_end)

    breaks
  }
}


#' @rdname chop_proportions
#' @export
#' @order 2
brk_proportions <- function(proportions) {
  assert_that(is.numeric(proportions), noNA(proportions),
                all(proportions >= 0), all(proportions <= 1))
  proportions <- sort(proportions)

  function (x, extend, left, close_end) {
    min_x <- quiet_min(x[is.finite(x)])
    max_x <- quiet_max(x[is.finite(x)])
    range_x <- max_x - min_x
    if (sign(range_x) <= 0) return(empty_breaks())

    breaks <- min_x + range_x * proportions
    breaks <- create_lr_breaks(breaks, left)

    scaled_endpoints <- proportions
    needs <- needs_extend(breaks, x, extend, left, close_end)
    if ((needs & LEFT) > 0) scaled_endpoints <- c(0, scaled_endpoints)
    if ((needs & RIGHT) > 0) scaled_endpoints <- c(scaled_endpoints, 1)
    breaks <- extend_and_close(breaks, x, extend, left, close_end)

    attr(breaks, "scaled_endpoints") <- scaled_endpoints
    names(breaks) <- names(scaled_endpoints)

    breaks
  }
}

#' Define singleton intervals explicitly
#'
#' `exactly()` duplicates its input.
#' It lets you define singleton intervals like this: `chop(x, c(1, exactly(2), 3))`.
#' This is the same as `chop(x, c(1, 2, 2, 3))` but conveys your intent more
#' clearly.
#'
#' @param x A numeric vector.
#'
#' @return The same as `rep(x, each = 2)`.
#' @export
#'
#' @examples
#' chop(1:10, c(2, exactly(5), 8))
#'
#' # same:
#' chop(1:10, c(2, 5, 5, 8))
exactly <- function (x) rep(x, each = 2)



#' Simple percentage formatter
#'
#' `percent()` formats `x` as a percentage.
#' For a wider range of formatters, consider the [`scales`
#' package](https://cran.r-project.org/package=scales).
#'
#' @param x Numeric values.
#'
#' @return `x` formatted as a percent.
#' @export
#'
#' @examples
#' percent(0.5)
percent <- function (x) {
  paste0(unique_truncation(x * 100), "%")
}


#' Report singleton intervals
#'
#' @param breaks A breaks object
#'
#' @return
#' A logical vector of `length(breaks) - 1`, `TRUE`
#' if the corresponding interval is a singleton.
#' @noRd
#'
#' @examples
#' brk <- brk_res(brk_default(c(1, 1, 2, 3, 3, 4)))
#' brk
#' # Breaks object: {1} (1, 2) [2, 3) {3} (3, 4)
#'
#' singletons(brk)
#' # TRUE FALSE FALSE TRUE FALSE
singletons <- function (breaks) {
  duplicated(breaks)[-1]
}


#' Find duplicates that would be illegal breaks
#'
#' @param x A vector, which should be sorted
#'
#' @return A logical vector of length(x), TRUE if the corresponding element
#'   is the second duplicate in a row.
#' @noRd
#'
#' @examples
#' find_illegal_duplicates(c(1, 2, 2, 3, 3, 3, 4))
find_illegal_duplicates <- function (x) {
  if (length(x) == 0) return(logical(0))
  dupes <- duplicated(x)
  # If element n is duplicated, and element n+1 is duplicated, then n + 1
  # is illegal.
  # The first element is never a duplicated middle.
  c(FALSE, dupes[-length(dupes)] & dupes[-1])
}


`%||%` <- function (x, y) if (is.null(x)) y else x


quiet_min <- function (x) suppressWarnings(min(x, na.rm = TRUE))


quiet_max <- function (x) suppressWarnings(max(x, na.rm = TRUE))



#' Stricter `as.numeric`
#'
#' This converts warnings to errors, and errors if any NAs are introduced,
#' but is less strict than `vctrs::vec_cast()`
#'
#' @param x A vector
#'
#' @return `as.numeric(x)`, with no new NAs
#' @noRd
#'
strict_as_numeric <- function (x) {
  nas <- is.na(x)

  x <- tryCatch(as.numeric(x),
                  warning = function (w) stop("Warning from as.numeric(x)")
                )
  if (any(is.na(x) & ! nas)) stop("Could not convert some elements")

  x
}


#' Test a break
#'
#' @param brk_fun A call to a `brk_` function
#' @param x,extend,left,close_end Passed in to `brk_fun`
#'
#' @return A `breaks` object.
#' @noRd
#'
#' @examples
#' brk_res(brk_default(1:3))
#'
brk_res <- function (
  brk_fun,
  x         = 1:2,
  extend    = FALSE,
  left      = TRUE,
  close_end = TRUE
) {
  brk_fun(x, extend = extend, left = left, close_end = close_end)
}

#' Chop by quantiles
#'
#' `chop_quantiles()` chops data by quantiles.
#' `chop_deciles()` is a convenience function which chops into deciles.
#'
#' @param probs A vector of probabilities for the quantiles. If `probs` has
#'   names, these will be used for labels.
#' @param ... For `chop_quantiles`, passed to [chop()]. For `brk_quantiles()`,
#'   passed to [stats::quantile()] or [Hmisc::wtd.quantile()].
#' @param weights `NULL` or numeric vector of same length as `x`. If not
#'   `NULL`, [Hmisc::wtd.quantile()] is used to calculate weighted quantiles.
#' @param recalc_probs Logical. Recalculate probabilities of quantiles using
#'   [`ecdf(x)`][stats::ecdf()]? See below.
#'
#' @inheritParams chop
#' @inherit chop-doc params return
#'
#' @details
#' For non-numeric `x`, `left` is set to `FALSE` by default. This works better
#' for calculating "type 1" quantiles, since they round down. See
#' [stats::quantile()].
#'
#' By default, `chop_quantiles()` shows the requested probabilities in the
#' labels. To show the numeric quantiles themselves, set `raw = TRUE`.
#'
#' When `x` contains duplicates, consecutive quantiles may be the same number. If
#' so, interval labels may be misleading, and if `recalc_probs = FALSE` a warning is
#' emitted. Set `recalc_probs = TRUE` to recalculate the probabilities of the quantiles
#' using the [empirical cumulative distribution function][stats::ecdf()] of `x`.
#' Doing so may give you different labels from what you expect, and will
#' remove any names from `probs`, but it never changes the actual
#' quantiles used for breaks. At present, `recalc_probs = TRUE` is incompatible 
#' with non-null `weights`. See the example below.
#'
#' @family chopping functions
#'
#' @export
#' @order 1
#'
#' @examples
#' chop_quantiles(1:10, 1:3/4)
#'
#' chop_quantiles(1:10, c(Q1 = 0, Q2 = 0.25, Q3 = 0.5, Q4 = 0.75))
#'
#' chop(1:10, brk_quantiles(1:3/4))
#'
#' chop_deciles(1:10)
#'
#' # to label by the quantiles themselves:
#' chop_quantiles(1:10, 1:3/4, raw = TRUE)
#'
#' # duplicate quantiles:
#' x <- c(1, 1, 1, 2, 3)
#' quantile(x, 1:5/5)
#' tab_quantiles(x, 1:5/5)
#' tab_quantiles(x, 1:5/5, recalc_probs = TRUE)
chop_quantiles <- function(
                    x,
                    probs,
                    ...,
                    labels    = if (raw) lbl_intervals() else
                                         lbl_intervals(single = NULL),
                    left      = is.numeric(x),
                    raw       = FALSE,
                    weights   = NULL,
                    recalc_probs  = FALSE
                  ) {
  chop(x, brk_quantiles(probs, weights = weights, recalc_probs = recalc_probs),
       labels = labels, ..., left = left, raw = raw)
}


#' @rdname chop_quantiles
#' @export
#' @order 1
chop_deciles <- function(x, ...) {
  chop_quantiles(x, 0:10/10, ...)
}


#' Chop equal-sized groups
#'
#' `chop_equally()` chops `x` into groups with an equal number of elements.
#'
#' @param groups Number of groups.
#' @inheritParams chop
#' @inherit chop-doc params return
#'
#' @details
#' `chop_equally()` uses [brk_quantiles()] under the hood. If `x` has duplicate
#' elements, you may get fewer `groups` than requested. If so, a warning will
#' be emitted. See the examples.
#'
#' @family chopping functions
#'
#' @export
#' @order 1
#' @examples
#' chop_equally(1:10, 5)
#'
#' # You can't always guarantee equal-sized groups:
#' dupes <- c(1, 1, 1, 2, 3, 4, 4, 4)
#' quantile(dupes, 0:4/4)
#' chop_equally(dupes, 4)
#' # Or as many groups as you ask for:
#' chop_equally(c(1, 1, 2, 2), 3)
chop_equally <- function (
                  x,
                  groups,
                  ...,
                  labels    = lbl_intervals(),
                  left      = is.numeric(x),
                  raw       = TRUE
                ) {
  chop(x, brk_equally(groups), ..., labels = labels, left = left, raw = raw)
}


#' Chop into fixed-sized groups
#'
#' `chop_n()` creates intervals containing a fixed number of elements.
#'
#' @param n Integer. Number of elements in each interval.
#' @inheritParams chop
#' @param tail String. What to do if the final interval has fewer than `n` elements?
#'   `"split"` to keep it separate. `"merge"` to merge it with the neighbouring
#'   interval.
#' @inherit chop-doc params return
#'
#'
#' @details
#'
#' The algorithm guarantees that intervals contain no more than `n` elements, so
#' long as there are no duplicates in `x` and `tail = "split"`. It also
#' guarantees that intervals contain no fewer than `n` elements, except possibly
#' the last interval (or first interval if `left` is `FALSE`).
#'
#' To ensure that all intervals contain at least `n` elements (so long as there
#' are at least `n` elements in `x`!) set `tail = "merge"`.
#'
#' If `tail = "split"` and there are intervals containing duplicates with more
#' than `n` elements, a warning is given.
#'
#' @export
#' @order 1
#' @family chopping functions
#' @examples
#' chop_n(1:10, 5)
#'
#' chop_n(1:5, 2)
#' chop_n(1:5, 2, tail = "merge")
#'
#' # too many duplicates
#' x <- rep(1:2, each = 3)
#' chop_n(x, 2)
#'
chop_n <- function (
            x,
            n,
            ...,
            tail = "split"
          ) {
  res <- chop(x, brk_n(n, tail = tail), ...)
  if (tail == "split" && max(tabulate(res)) > n) {
    warning("Some intervals contain more than ", n, " elements")
  }

  res
}

d1 <- seq(as.Date("1975-10-27"), as.Date("1975-11-27"), by = "day")

db1 <- as.Date(c("1975-11-01", "1975-11-15"))

dt1 <- seq(as.POSIXct("2000-01-01 15:00"), length = 20, by = "1 min")

dtb1 <- dt1[c(5, 15)]

table_vals <- function (x) unclass(table(x))


test_that("Basic chop", {
  expect_silent(chop(d1, db1))
  expect_silent(chop(dt1, dtb1))
})


test_that("Chop with conversion", {
  lb <- lbl_seq()
  withr::with_timezone("UTC", {
    expect_equivalent(chop(d1, db1, lb), chop(as.POSIXct(d1), db1, lb))
    expect_equivalent(chop(d1, db1, lb), chop(d1, as.POSIXct(db1), lb))
  })
})


test_that("Basic breaks", {
  expect_silent(brk_res(brk_default(db1), d1, extend = FALSE))
  expect_silent(brk_res(brk_default(db1), d1, extend = NULL))
  expect_silent(brk_res(brk_default(db1), d1, extend = TRUE))

  expect_silent(brk_res(brk_default(db1), d1, left = FALSE))
  expect_silent(brk_res(brk_default(db1), d1, close_end = TRUE))
  expect_silent(brk_res(brk_default(db1), d1, left = FALSE, close_end = TRUE))
})


test_that("chop_equally", {
  expect_silent(res <- chop_equally(d1, groups = 4))
  expect_equivalent(table_vals(res), rep(8, 4))

  expect_silent(res2 <- chop_equally(dt1, groups = 4))
  expect_equivalent(table_vals(res2), rep(5, 4))
})


test_that("chop_n", {
  expect_silent(res <- chop_n(d1, 4))
  expect_equivalent(table_vals(res), rep(4, 8))

  expect_silent(res2 <- chop_n(dt1, 5))
  expect_equivalent(table_vals(res2), rep(5, 4))
})


test_that("chop_quantiles", {
  # `left = FALSE` works better with type 1 quantiles, which round down.
  expect_silent(res1 <- chop_quantiles(d1, 0:4/4, left = FALSE))
  expect_equivalent(table_vals(res1), rep(8, 4))

  expect_silent(res2 <- chop_quantiles(dt1, 0:5/5, left = FALSE))
  expect_equivalent(table_vals(res2), rep(4, 5))
})


test_that("chop_mean_sd", {
  expect_silent(res <- chop_mean_sd(d1))
  cmp <- cut(d1, mean(d1) + (-2:2) * sd(d1), right = FALSE)
  expect_equivalent(table_vals(res), table_vals(cmp))

  expect_silent(res2 <- chop_mean_sd(dt1))
  cmp2 <- cut(dt1, mean(dt1) + (-2:2) * sd(dt1), right = FALSE)
  expect_equivalent(table_vals(res2), table_vals(cmp2))

  expect_silent(res3 <- chop_mean_sd(d1, c(1, 1.4)))
  # the -10 and 10 capture values outside 1.4 sds:
  cmp3 <- cut(d1,  mean(d1) + c(-10, -1.4, -1, 0, 1, 1.4, 10) * sd(d1), right = FALSE)
  expect_equivalent(table_vals(res3), table_vals(cmp3))
})


test_that("chop_pretty", {
  expect_silent(res <- chop_pretty(d1))
  cmp <- chop(d1, base::pretty(d1))
  expect_equivalent(
    table_vals(res),
    table_vals(cmp)
  )
})


test_that("chop_width: difftime", {
  difftime_w1 <- as.difftime(4, units = "days")

  expect_silent(res1 <- chop_width(d1, width = difftime_w1))
  expect_equivalent(table_vals(res1), rep(4, 8))

  expect_silent(
    res2 <- chop_width(d1, width = difftime_w1, start = as.Date("1975-11-01"))
  )
  tv <- table_vals(res2)
  expect_true(all(tv[c(-1, -length(tv))] == 4))

  difftime_w2 <- as.difftime(5, units = "mins")

  expect_silent(res3 <- chop_width(dt1, width = difftime_w2))
  expect_equivalent(table_vals(res3), rep(5, 4))

  expect_silent(
    res4 <- chop_width(dt1, width = difftime_w2,
          start = as.POSIXct("2000-01-01 15:10"))
  )
  expect_equivalent(table_vals(res4), c(10, 5, 5))

  expect_silent(
    res5 <- chop_width(d1, width = as.difftime(-4, units = "days"))
  )
  expect_equivalent(table_vals(res5), rep(4, 8))

  expect_silent(
    res6 <- chop_width(d1, width = as.difftime(-4, units = "days"),
          start = as.Date("1975-11-25"))
  )
  tv <- table_vals(res6)
  expect_true(all(tv[c(-1, -length(tv))] == 4))

  expect_silent(
    res7 <- chop_width(dt1, width = as.difftime(-5, units = "mins"))
  )
  expect_equivalent(table_vals(res7), rep(5, 4))

  expect_silent(chop_width(d1, as.difftime(7, units = "days")))
  expect_silent(chop_width(dt1, as.difftime(7, units = "mins")))
})


test_that("chop_width: Duration", {
  skip_if_not_installed("lubridate")
  library(lubridate)

  duration_w1 <- ddays(4)

  expect_silent(res1 <- chop_width(d1, width = duration_w1))
  expect_equivalent(table_vals(res1), rep(4, 8))

  expect_silent(
    res2 <- chop_width(d1, width = duration_w1, start = as.Date("1975-11-16"))
  )
  expect_equivalent(table_vals(res2), c(20, 4, 4, 4))

  duration_w2 <- dminutes(5)

  expect_silent(res3 <- chop_width(dt1, width = duration_w2))
  expect_equivalent(table_vals(res3), rep(5, 4))

  expect_silent(
    res4 <- chop_width(dt1, duration_w2, start = as.POSIXct("2000-01-01 15:10"))
  )
  expect_equivalent(table_vals(res4), c(10, 5, 5))

  expect_silent(chop_width(d1, ddays(7)))
  expect_silent(chop_width(dt1, dminutes(7)))
})


test_that("chop_width: Period", {
  skip_if_not_installed("lubridate")
  library(lubridate)

  period_w1 <- days(8)

  expect_silent(res1 <- chop_width(d1, width = period_w1))
  expect_equivalent(table_vals(res1), rep(8, 4))

  expect_silent(
    res2 <- chop_width(d1, period_w1, start = as.Date("1975-11-12"))
  )
  expect_equivalent(table_vals(res2), c(16, 8, 8))

  period_w2 <- minutes(5)

  expect_silent(res3 <- chop_width(dt1, width = period_w2))
  expect_equivalent(table_vals(res3), rep(5, 4))

  expect_silent(
    res4 <- chop_width(dt1, period_w2, start = as.POSIXct("2000-01-01 15:07"))
  )
  expect_equivalent(table_vals(res4), c(7, 5, 5, 3))

  expect_silent(chop_width(d1, days(7)))
  expect_silent(chop_width(dt1, minutes(7)))

  # TODO: include tests that Period deals with quirks
})


test_that("chop_width: Period quirks", {
  skip_if_not_installed("lubridate")
  library(lubridate)

  noughties <- seq(as.Date("2000-01-01"), as.Date("2009-12-31"), by = "day")

  res1 <- chop_width(noughties, years(1))
  expect_equivalent(
    table_vals(res1),
    c(366, 365, 365, 365, 366, 365, 365, 365, 366, 365)
  )

  y2k <- seq(as.Date("2000-01-01"), as.Date("2000-12-31"), by = "day")
  res2 <- chop_width(y2k, months(1))
  expect_equivalent(
    table_vals(res2),
    c(31, 29, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31)
  )
})


test_that("chop_evenly", {
  expect_silent(res1 <- chop_evenly(d1, 8))
  expect_equivalent(table_vals(res1), rep(4, 8))

  expect_silent(res2 <- chop_evenly(dt1, 4))
  expect_equivalent(table_vals(res2), rep(5, 4))

  expect_silent(chop_evenly(d1, 7))
  expect_silent(chop_evenly(dt1, 7))
})


test_that("brk_spikes", {
  d_reps <- c(rep(d1[3], 5), d1[-3])

  expect_silent(
    res <- chop(d_reps, brk_spikes(db1, n = 5))
  )

  expect_in("{1975-10-29}", levels(res))
  expect_equal(sum(res == "{1975-10-29}"), 5)
})


test_that("chop timezones", {
  dt_z1 <- seq(as.POSIXct("2000-01-01 09:00:00", tz = "GMT"),
        by = "hour", length.out = 24)
  # 8 hours behind. Hi Tom and Dan!
  dtb_z2 <- as.POSIXct("2000-01-01 12:30:00", tz = "America/Los_Angeles")

  res1 <- chop(dt_z1, dtb_z2)
  expect_equivalent(table_vals(res1), c(12, 12))
  # we convert breaks to the timezone of x
  expect_match(levels(res1), "20:30", fixed = TRUE)
})


test_that("Date labels", {
  li <- lbl_intervals()
  b <- brk_res(brk_default(db1), close_end = FALSE)
  expect_equivalent(
    li(b), "[1975-11-01, 1975-11-15)"
  )

  withr::local_options(santoku.infinity = "Inf")
  b2 <- brk_res(brk_default(db1), x = as.Date("1975-01-01"), extend = TRUE,
                close_end = FALSE)
  expect_equivalent(
    li(b2), c("[-Inf, 1975-11-01)", "[1975-11-01, 1975-11-15)", "[1975-11-15, Inf]")
  )

  expect_equivalent(
    lbl_intervals(fmt = "%y %m %d")(b),
    "[75 11 01, 75 11 15)"
  )

  expect_equivalent(
    lbl_dash(" to ")(b),
    "1975-11-01 to 1975-11-15"
  )

  expect_equivalent(
    lbl_dash(" to ", fmt = "%d/%m")(b),
    "01/11 to 15/11"
  )
})


test_that("POSIXct labels", {
  li <- lbl_intervals()
  b <- brk_res(brk_default(dtb1), close_end = FALSE)
  expect_equivalent(
    li(b), "[2000-01-01 15:04:00, 2000-01-01 15:14:00)"
  )

  withr::local_options(santoku.infinity = "Inf")
  b2 <- brk_res(brk_default(dtb1), x = as.POSIXct("2000-01-01 15:00:00"), extend = TRUE)
  expect_equivalent(
    li(b2), c("[-Inf, 2000-01-01 15:04:00)",
              "[2000-01-01 15:04:00, 2000-01-01 15:14:00)",
              "[2000-01-01 15:14:00, Inf]")
  )

  expect_equivalent(
    lbl_intervals(fmt = "%H:%M")(b),
    "[15:04, 15:14)"
  )

  expect_equivalent(
    lbl_dash(" to ")(b),
    "2000-01-01 15:04:00 to 2000-01-01 15:14:00"
  )

  expect_equivalent(
    lbl_dash(" to ", fmt = "%H.%M")(b),
    "15.04 to 15.14"
  )
})

test_that("basic functionality", {
  x <- 1:3
  lbrks <- brk_manual(1:3, rep(TRUE, 3))
  rbrks <- brk_manual(1:3, rep(FALSE, 3))
  rc_brks <- brk_manual(1:3, c(TRUE, TRUE, FALSE))

  expect_equivalent(
    chop(x, lbrks, lbl_seq("1"), extend = FALSE),
    factor(c(1, 2, NA))
  )
  expect_equivalent(
    chop(x, rbrks, lbl_seq("1"), extend = FALSE),
    factor(c(NA, 1, 2))
  )
  expect_equivalent(
    chop(x, rc_brks, lbl_seq("1"), extend = FALSE),
    factor(c(1, 2, 2))
  )
})


test_that("NA, NaN and Inf", {
  y <- c(1:3, NA, NaN)
  expect_equivalent(
    chop(y, 1:3, lbl_seq("1"), extend = FALSE, close_end = FALSE),
    factor(c(1, 2, NA, NA, NA))
  )

  x <- c(-Inf, 1, Inf)
  r <- chop(x, 1:2, labels = letters[1:3])
  expect_equivalent(r, factor(c("a", "b", "c"), levels = letters[1:3]))

  x <- c(-Inf, 1, Inf)
  # if extend is NULL, we should ensure even Inf is included
  r <- chop(x, -Inf, left = FALSE, labels = c("-Inf", "a"), close_end = FALSE)
  expect_equivalent(r, factor(c("-Inf", "a", "a")))
  r <- chop(x, Inf, labels = c("a", "Inf"), close_end = FALSE)
  expect_equivalent(r, factor(c("a", "a", "Inf")))

  # otherwise, we respect close_end = FALSE
  r <- chop(x, brk_default(c(-Inf, Inf)), labels = "a",
        extend = FALSE, left = FALSE, close_end = FALSE)
  expect_equivalent(r, factor(c(NA, "a", "a")))
  r <- chop(x, c(-Inf, Inf), labels = "a", extend = FALSE, close_end = FALSE)
  expect_equivalent(r, factor(c("a", "a", NA)))

  all_na <- rep(NA_real_, 5)
  expect_silent(chop(all_na, 1:2))
  # not sure if this should be OK or not...
  # expect_silent(chop_quantiles(all_na, c(.25, .75)))
  all_na[1] <- NaN
  expect_silent(chop(all_na, 1:2))
})


test_that("singleton breaks", {
  expect_silent(chop(1:4, 2))
  expect_silent(chop(1:4, 1))
  expect_silent(chop(1:4, 4))
  expect_silent(chop(1:4, 0))
  expect_silent(chop(1:4, 5))
  expect_silent(chop(1, 1))
})


test_that("labels", {
  x <- seq(0.5, 2.5, 0.5)
  expect_equivalent(
          chop(x, 1:2, labels = letters[1:3]),
          factor(c("a", "b", "b", "c", "c"), levels = letters[1:3])
        )
  expect_error(chop(1:10, 3:4, labels = c("a", "a", "a")))
  expect_error(chop(1:10, 3:4, labels = c("a", "b")))
  expect_error(chop(1:10, 3:4, labels = c("a", "b", "c", "d")))

  expect_equivalent(
          chop(x, 1:2, labels = NULL),
          c(1, 2, 2, 3, 3)
        )
})


test_that("break names as labels", {
  expect_equivalent(
    chop(1:4, c(Low = 1, High = 3, 4)),
    factor(c("Low", "Low", "High", "High"))
  )
  expect_equivalent(
    chop(1:5, c(Low = 1, Mid = 3, High = 4)),
    factor(c("Low", "Low", "Mid", "High", "High"))
  )
  expect_equivalent(
    chop(0:4, c(Low = 1, High = 3)),
    factor(c("[0, 1)", "Low", "Low", "High", "High"))
  )
  expect_equivalent(
    chop(1:4, c(Low = 1, Mid = 2, 3, 4), labels = lbl_endpoints()),
    factor(c("Low", "Mid", "3", "3"))
  )
})


test_that("extend", {
  expect_equivalent(
          chop(c(1, 4), 2:3, labels = lbl_seq("1"), extend = TRUE),
          factor(c(1, 3))
        )
  expect_equivalent(
          chop(c(1, 4), 2:3, labels = lbl_seq("1"), extend = FALSE),
          factor(c(NA, NA))
        )
})


test_that("close_end", {
  res <- chop(1:4, 2:3, close_end = TRUE, drop = FALSE)
  expect_equivalent(
    levels(res),
    c("[1, 2)", "[2, 3)", "[3, 4]")
  )

  res <- chop(1:4, 2:3, close_end = FALSE, extend = FALSE, drop = FALSE)
  expect_equivalent(
    levels(res),
    c("[2, 3)")
  )

  res <- chop(1:4, 2:3, close_end = TRUE, extend = FALSE, drop = FALSE)
  expect_equivalent(
    levels(res),
    c("[2, 3]")
  )
})


test_that("raw", {
  x <- 1:10

  expect_silent(
    res <- chop(x, brk_quantiles(c(0.25, 0.75)), raw = TRUE)
  )
  expect_equivalent(
    levels(res),
    c("[1, 3.25)", "[3.25, 7.75)", "[7.75, 10]")
  )

  expect_silent(
    res <- chop(x, brk_quantiles(c(0.25, 0.75)), raw = FALSE)
  )
  expect_equivalent(
    levels(res),
    c("[0%, 25%)", "[25%, 75%)", "[75%, 100%]")
  )

  # raw overrides raw in labels
  withr::local_options(lifecycle_verbosity = "quiet")

  expect_silent(
    res <- chop(x, brk_quantiles(c(0.25, 0.75)),
                  labels = lbl_intervals(raw = FALSE), raw = TRUE)
  )
  expect_equivalent(
    levels(res),
    c("[1, 3.25)", "[3.25, 7.75)", "[7.75, 10]")
  )

  expect_silent(
    res <- chop(x, brk_quantiles(c(0.25, 0.75)),
                  labels = lbl_intervals(raw = TRUE), raw = FALSE)
  )
  expect_equivalent(
    levels(res),
    c("[0%, 25%)", "[25%, 75%)", "[75%, 100%]")
  )
})


test_that("drop", {
  x <- c(1, 3)
  expect_equivalent(
          levels(chop(x, 1:3, labels = lbl_seq("1"), extend = TRUE,
            drop = TRUE)),
          as.character(c(2, 4))
        )
  expect_equivalent(
          levels(chop(x, 1:3, labels = lbl_seq("1"), extend = TRUE,
            drop = FALSE)),
          as.character(1:4)
        )
})


test_that("chop_width", {
  x <- 1:10
  expect_equivalent(
    chop_width(x, 2, labels = lbl_seq("1")),
    factor(rep(1:5, each = 2))
  )
  expect_equivalent(
    chop_width(x, 2, 0, labels = lbl_seq("1")),
    factor(c(1, rep(2:4, each = 2), 5, 5, 5))
  )
})


test_that("chop_evenly", {
  x <- 1:10
  expect_equivalent(
    chop_evenly(x, 2, labels = lbl_seq("1")),
    factor(rep(1:2, each = 5))
  )
  expect_error(r <- chop_evenly(x, groups = 2))
})


test_that("chop_proportions", {
  expect_equivalent(
    chop_proportions(0:10, c(0.2, 0.8), labels = lbl_seq("1")),
    factor(rep(1:3, c(2, 6, 3)))
  )

  expect_equivalent(
    chop_proportions(0:10, c(Low = 0, Mid = 0.2, High = 0.8)),
    factor(c(rep("Low", 2), rep("Mid", 6), rep("High", 3)))
  )

  withr::local_options(lifecycle_verbosity = "quiet")
  expect_equivalent(
    chop_proportions(0:10, c(0.2, 0.8), labels = lbl_intervals(), raw = FALSE),
    chop_proportions(0:10, c(0.2, 0.8), labels = lbl_intervals(raw = FALSE),
                       raw = NULL)
  )
})


test_that("chop_quantiles", {
  expect_equivalent(
    chop_quantiles(1:6, c(.25, .5, .75), labels = lbl_seq("1")),
    as.factor(c(1, 1, 2, 3, 4, 4))
  )

  expect_equivalent(
    chop_quantiles(1:6, c(Q1 = 0, Q2 = 0.25, Q3 = 0.5, Q4 = 0.75)),
    factor(c("Q1", "Q1", "Q2", "Q3", "Q4", "Q4"))
  )

  x <- c(1, 1, 1, 2, 3)
  expect_equivalent(
    chop_quantiles(x, 1:4/5, recalc_probs = TRUE),
    factor(c("[0%, 60%]", "[0%, 60%]", "[0%, 60%]", "[60%, 80%)", "[80%, 100%]"))
  )

  withr::local_options(lifecycle_verbosity = "quiet")
  expect_equivalent(
    chop_quantiles(1:6, c(.25, .5, .75), raw = TRUE),
    chop_quantiles(1:6, c(.25, .5, .75), labels = lbl_intervals(raw = TRUE),
                     raw = NULL)
  )
})


test_that("chop_equally", {
  x <- 1:6
  expect_equivalent(
    chop_equally(x, 2, labels = lbl_seq("1")),
    as.factor(rep(1:2, each = 3))
  )

  withr::local_options(lifecycle_verbosity = "quiet")
  expect_equivalent(
    chop_equally(x, 2, labels = lbl_intervals(raw = FALSE), raw = NULL),
    chop_equally(x, 2, raw = FALSE)
  )

  expect_equivalent(
    chop_equally(x, 2, labels = lbl_intervals(raw = TRUE)),
    chop_equally(x, 2, raw = TRUE)
  )

  expect_warning(
    chop_equally(c(1, 1, 1, 1), 4),
    "Fewer"
  )
})


test_that("chop_deciles", {
  x <- rnorm(100)
  expect_identical(
    chop_quantiles(x, 0:10/10),
    chop_deciles(x)
  )
})


test_that("chop_n", {
  expect_silent(res <- chop_n(rnorm(100), 10))
  expect_equivalent(as.vector(table(res)), rep(10, 10))

  # chop_n should give accurate answers even when left = FALSE
  res <- chop_n(1:4, 2, left = FALSE)
  expect_equivalent(as.vector(table(res)), rep(2, 2))

  expect_warning(chop_n(rep(1:3, each = 3), 2))
})


test_that("Bugfix: chop_n(tail = 'merge') works with n > length(x)", {
  expect_silent(
    chop_n(1:3, 4, tail = "merge", extend = FALSE)
  )
})


test_that("chop_mean_sd", {
  x <- -1:1 # mean 0, sd 1
  expect_silent(res <- chop_mean_sd(x))
  expect_equivalent(as.vector(table(res)), c(1, 1, 1))
  expect_silent(res2 <- chop_mean_sd(x, sds = 1:2))
  expect_silent(chop_mean_sd(x, sds = c(1, 1.96)))

  lifecycle::expect_deprecated(res3 <- chop_mean_sd(x, sd = 2))
  expect_equivalent(res2, res3)

  withr::local_options(lifecycle_verbosity = "quiet")
  expect_equivalent(
    chop_mean_sd(x, raw = TRUE),
    chop_mean_sd(x, labels = lbl_intervals(raw = TRUE), raw = NULL),
  )
})


test_that("chop_pretty", {
  expect_silent(res <- chop_pretty(1:10))
  expect_silent(res <- chop_pretty(1:10, 3))
  expect_silent(res <- chop_pretty(1:10, 3))
})


test_that("chop_fn", {
  expect_silent(res <- chop_fn(1:10, pretty))
  expect_silent(res <- chop_fn(1:10, quantile, c(.2, .8)))
  expect_equivalent(
    chop_fn(1:5, median),
    factor(c("[1, 3)", "[1, 3)", "[3, 5]", "[3, 5]", "[3, 5]"))
  )
  expect_equivalent(
    chop_fn(1:5, median, left = FALSE),
    factor(c("[1, 3]", "[1, 3]", "[1, 3]", "(3, 5]", "(3, 5]"))
  )
})


test_that("chop_spikes", {
  x <- c(1:2, rep(3, 3), 4:5)

  expect_silent(res <- chop_spikes(x, breaks = 2, n = 2))
  expect_equivalent(
    res,
    factor(c("[1, 2)", "[2, 3)", rep("{3}", 3), rep("(3, 5]", 2)))
  )

  expect_silent(res2 <- chop_spikes(x, breaks = 2, prop = 0.3))
  expect_equivalent(res, res2)

  expect_silent(res3 <- chop_spikes(x, breaks = brk_width(4), n = 2))
  expect_equivalent(
    res3,
    factor(c(rep("[1, 3)", 2), rep("{3}", 3), rep("(3, 5]", 2)))
  )
})


test_that("dissect", {
  x <- c(1:2, rep(3, 3), 4:6)

  expect_silent(res <- dissect(x, breaks = c(2, 5), n = 2))
  expect_equivalent(
    res,
    factor(c("[1, 2)", "[2, 5)", rep("{3}", 3), "[2, 5)", "[5, 6]", "[5, 6]"))
  )

  expect_silent(res2 <- dissect(x, breaks = c(2, 5), prop = 0.25))
  expect_equivalent(res, res2)

  x <- c(1, 2, 3, 4, 5, 5, 5, 5)
  expect_silent(res3 <- dissect(x, breaks = brk_equally(2), n = 2,
                                exclude_spikes = TRUE))
  expect_equivalent(
    res3,
    factor(c("[0%, 50%)", "[0%, 50%)", "[50%, 100%]", "[50%, 100%]",
             rep("{5}", 4)))
  )
})


test_that("fillet", {
  x <- -2:2
  expect_silent(sole <- fillet(x, -1:1))
  expect_identical(sole, chop(x, -1:1, extend = FALSE, drop = FALSE))
})

#' Hacked version of [vctrs::vec_cast_common()]
#'
#' @param x,y Vectors to cast
#'
#' @return A list of two vectors of the same class; or
#' errors, if that isn't possible.
#'
#' This almost always defers to `vctrs::vec_cast_common()`.
#'
#' Often, we are more relaxed than `vctrs` because
#' we have a more specific use case (comparing numeric
#' values). So e.g. we're fine with comparing a `ts`
#' object to a number, whereas other binary
#' operations might not make sense.
#'
#' @noRd
#'
santoku_cast_common <- function (x, y) {
  UseMethod("santoku_cast_common")
}


#' Internal functions
#'
#' @name santoku-cast
#' @param x,y Vectors to cast.
#'
#' @return A list.
#' @keywords internal
#'
#' These are internal functions. Do not use.
NULL


# The rawNamespace below means NAMESPACE gets both the S3method() and the
# export() tag.

#' @export
#' @rdname santoku-cast
#' @method santoku_cast_common default
#' @rawNamespace export(santoku_cast_common.default)
santoku_cast_common.default <- function (x, y) {
  UseMethod("santoku_cast_common.default", object = y)
}


#' @export
santoku_cast_common.default.default <- function (x, y) {
  vctrs::vec_cast_common(x, y)
}

# Specific default.x methods are in their sections below.


# ==== double ====

# We have specific double methods just to catch bit64 objects

#' @export
#' @rdname santoku-cast
#' @method santoku_cast_common double
#' @rawNamespace export(santoku_cast_common.double)
santoku_cast_common.double <- function (x, y) {
  UseMethod("santoku_cast_common.double", object = y)
}


#' @export
santoku_cast_common.double.default <- function (x, y) {
  # almost always delegate to default
  santoku_cast_common.default(x, y)
}


#' @export
santoku_cast_common.double.integer64 <- function (x, y) {
  loadNamespace("bit64")
  # we cast the integer64 to double.
  # See santoku_cast_common.integer64.double below for why.
  list(x, as.double(y))
}


# ==== Date ====

# We delegate to vctrs for Date+numeric (which gives an error)
# Not obvious how to interpret conversion of Date to numeric


#' @export
#' @rdname santoku-cast
#' @method santoku_cast_common Date
#' @rawNamespace export(santoku_cast_common.Date)
santoku_cast_common.Date <- function (x, y) {
  UseMethod("santoku_cast_common.Date", object = y)
}


#' @export
santoku_cast_common.Date.Date <- function (x, y) {
  list(x, y)
}


#' @export
santoku_cast_common.Date.POSIXct <- function (x, y) {
  list(as.POSIXct(x), y)
}


# ==== POSIXct ====

# We delegate to vctrs for POSIXct/numeric, see Date above

#' @export
#' @rdname santoku-cast
#' @method santoku_cast_common POSIXct
#' @rawNamespace export(santoku_cast_common.POSIXct)
santoku_cast_common.POSIXct <- function (x, y) {
  UseMethod("santoku_cast_common.POSIXct", object = y)
}


#' @export
santoku_cast_common.POSIXct.POSIXct <- function (x, y) {
  list(x, y)
}


#' @export
santoku_cast_common.POSIXct.Date <- function (x, y) {
  list(x, as.POSIXct(y))
}


# ==== ts ====

#' @export
#' @rdname santoku-cast
#' @method santoku_cast_common ts
#' @rawNamespace export(santoku_cast_common.ts)
santoku_cast_common.ts <- function (x, y) {
  UseMethod("santoku_cast_common.ts", object = y)
}


#' @export
santoku_cast_common.ts.default <- function (x, y) {
  # We recall so that we can pick up anything
  # unusual in y
  santoku_cast_common(unclass(x), y)
}


#' @export
santoku_cast_common.default.ts <- function (x, y) {
  santoku_cast_common(x, unclass(y))
}


# ==== zoo ====

#' @export
#' @rdname santoku-cast
#' @method santoku_cast_common zoo
#' @rawNamespace export(santoku_cast_common.zoo)
santoku_cast_common.zoo <- function (x, y) {
  UseMethod("santoku_cast_common.zoo", object = y)
}

# we don't have a zoo.zoo method because returning
# two zoo objects would cause comparisons to only
# work where the indices are the same. So, we always
# work on the underlying data.

#' @export
santoku_cast_common.zoo.default <- function (x, y) {
  loadNamespace("zoo")
  santoku_cast_common(zoo::coredata(x), y)
}

#' @export
santoku_cast_common.default.zoo <- function (x, y) {
  loadNamespace("zoo")
  santoku_cast_common(x, zoo::coredata(y))
}


# ==== integer64 ====

#' @export
#' @rdname santoku-cast
#' @method santoku_cast_common integer64
#' @rawNamespace export(santoku_cast_common.integer64)
santoku_cast_common.integer64 <- function (x, y) {
  UseMethod("santoku_cast_common.integer64", object = y)
}


#' @export
santoku_cast_common.integer64.integer64 <- function (x, y) {
  list(x, y)
}


#' @export
santoku_cast_common.integer64.double <- function (x, y) {
  loadNamespace("bit64")
  # we cast the integer64 to double.
  # This may lose precision, but if so, it gives a warning.
  # If we cast double to integer64, then
  # e.g. chop(as.integer64(1:5), 2.5)
  # silently converts 2.5 to 2 and gives a wrong answer
  list(as.double(x), y)
}


#' @export
santoku_cast_common.integer64.default <- function (x, y) {
  loadNamespace("bit64")
  santoku_cast_common(x, bit64::as.integer64(y))
}


#' @export
santoku_cast_common.default.integer64 <- function (x, y) {
  loadNamespace("bit64")
  santoku_cast_common(bit64::as.integer64(x), y)
}


# ==== hexmode ====

#' @export
#' @rdname santoku-cast
#' @method santoku_cast_common hexmode
#' @rawNamespace export(santoku_cast_common.hexmode)
santoku_cast_common.hexmode <- function (x, y) {
  UseMethod("santoku_cast_common.hexmode", object = y)
}


#' @export
santoku_cast_common.hexmode.hexmode <- function (x, y) {
  # if both items are hexmode, OK fine.
  list(x, y)
}


#' @export
santoku_cast_common.hexmode.default <- function (x, y) {
  santoku_cast_common(as.numeric(x), y)
}


#' @export
santoku_cast_common.default.hexmode <- function (x, y) {
  santoku_cast_common(x, as.numeric(y))
}


# ==== octmode ====

#' @export
#' @rdname santoku-cast
#' @method santoku_cast_common octmode
#' @rawNamespace export(santoku_cast_common.octmode)
santoku_cast_common.octmode <- function (x, y) {
  UseMethod("santoku_cast_common.octmode", object = y)
}


#' @export
santoku_cast_common.octmode.octmode <- function (x, y) {
  list(x, y)
}


#' @export
santoku_cast_common.octmode.default <- function (x, y) {
  santoku_cast_common(as.numeric(x), y)
}


#' @export
santoku_cast_common.default.octmode <- function (x, y) {
  santoku_cast_common(x, as.numeric(y))
}

#' Label chopped intervals by their midpoints
#'
#' This uses the midpoint of each interval for
#' its label.
#'
#' @inherit label-doc
#' @inherit first-last-doc
#'
#' @family labelling functions
#'
#' @export
#'
#' @examples
#' chop(1:10, c(2, 5, 8), lbl_midpoints())
lbl_midpoints <- function (
                   fmt    = NULL,
                   single = NULL,
                   first  = NULL,
                   last   = NULL,
                   raw    = FALSE
                 ) {
  if (! isFALSE(raw)) {
    lifecycle::deprecate_warn("0.9.0", "lbl_midpoints(raw)", "chop(raw)")
  }

  RAW <- raw # avoid "recursive default argument reference"
  function (breaks, raw = RAW) {
    assert_that(is.breaks(breaks))

    break_nums <- scaled_endpoints(breaks, raw = raw)
    l_nums <- break_nums[-length(break_nums)]
    r_nums <- break_nums[-1]
    # doing this, rather than (l_nums + r_nums)/2, works for e.g. Date objects:
    midpoints <- l_nums + (r_nums - l_nums)/2

    # we've applied raw already (anyway, midpoints is just a numeric)
    midpoints <- endpoint_labels(midpoints, raw = TRUE, fmt = fmt)

    gluer <- lbl_glue(label = "{m}", fmt = fmt, single = single, first = first,
                        last = last, m = midpoints)
    labels <- gluer(breaks, raw = raw)

    labels
  }
}


#' Label chopped intervals by their left or right endpoints
#'
#' This is useful when the left endpoint unambiguously indicates the
#' interval. In other cases it may give errors due to duplicate labels.
#'
#' `lbl_endpoint()` is `r lifecycle::badge("defunct")` and gives an
#' error since santoku 1.0.0.
#'
#' @inherit label-doc
#' @inherit first-last-doc
#' @param left Flag. Use left endpoint or right endpoint?
#'
#' @family labelling functions
#'
#' @export
#'
#' @examples
#' chop(1:10, c(2, 5, 8), lbl_endpoints(left = TRUE))
#' chop(1:10, c(2, 5, 8), lbl_endpoints(left = FALSE))
#' if (requireNamespace("lubridate")) {
#'   tab_width(
#'           as.Date("2000-01-01") + 0:365,
#'          months(1),
#'          labels = lbl_endpoints(fmt = "%b")
#'        )
#' }
#'
#' \dontrun{
#'   # This gives breaks `[1, 2) [2, 3) {3}` which lead to
#'   # duplicate labels `"2", "3", "3"`:
#'   chop(1:3, 1:3, lbl_endpoints(left = FALSE))
#' }
lbl_endpoints <- function (
                   left   = TRUE,
                   fmt    = NULL,
                   single = NULL,
                   first  = NULL,
                   last   = NULL,
                   raw    = FALSE
                 ) {
  assert_that(is.flag(left))

  if (! isFALSE(raw)) {
    lifecycle::deprecate_warn("0.9.0", "lbl_endpoints(raw)", "chop(raw)")
  }

  label <- if (left) "{l}" else "{r}"
  lbl_glue(label, fmt = fmt, single = single, first = first, last = last,
             raw = raw)
}


#' @rdname lbl_endpoints
#' @export
lbl_endpoint <- function (
                  fmt  = NULL,
                  raw  = FALSE,
                  left = TRUE
                ) {
  lifecycle::deprecate_stop(when = "0.8.0", what = "lbl_endpoint()",
                              with = "lbl_endpoints()")
}


#' Label chopped intervals in sequence
#'
#' `lbl_seq()` labels intervals sequentially, using numbers or letters.
#'
#' @param start String. A template for the sequence. See below.
#'
#' @details
#'`start` shows the first element of the sequence. It must contain exactly *one*
#' character out of the set "a", "A", "i", "I" or "1". For later elements:
#'
#' * "a" will be replaced by "a", "b", "c", ...
#' * "A" will be replaced by "A", "B", "C", ...
#' * "i" will be replaced by lower-case Roman numerals "i", "ii", "iii", ...
#' * "I" will be replaced by upper-case Roman numerals "I", "II", "III", ...
#' * "1" will be replaced by numbers "1", "2", "3", ...
#'
#' Other characters will be retained as-is.
#'
#' @family labelling functions
#' @inherit label-doc return
#'
#' @export
#'
#' @examples
#' chop(1:10, c(2, 5, 8), lbl_seq())
#'
#' chop(1:10, c(2, 5, 8), lbl_seq("i."))
#'
#' chop(1:10, c(2, 5, 8), lbl_seq("(A)"))
lbl_seq <- function(start = "a") {
  assert_that(is.string(start))
  # check like contains just one of a, A, i, I, 1
  match <- gregexpr("(a|A|i|I|1)", start)[[1]]
  if (length(match) > 1) stop("More than one a/A/i/I/1 found in `start`: ", start)
  if (match == -1) stop("No a/A/i/I/1 found in `start`: ", start)
  # replace that with the format-string and call lbl_manual appropriately
  key <- substr(start, match, match)
  fmt <- sub("(a|A|i|I|1)", "%s", start)

  res <- switch(key,
    "a" = function (breaks, raw = NULL) {
            if (length(breaks) > 27L) {
              stop("Can't use more than 26 intervals with lbl_seq(\"a\")")
            }
            sprintf(fmt, letters[seq_len(length(breaks) - 1L)])
          },
    "A" = function (breaks, raw = NULL) {
            if (length(breaks) > 27L) {
              stop("Can't use more than 26 intervals with lbl_seq(\"A\")")
            }
            sprintf(fmt, LETTERS[seq_len(length(breaks) - 1L)])
          },
    "i" = function (breaks, raw = NULL) {
           sprintf(fmt, tolower(utils::as.roman(seq_len(length(breaks) - 1L))))
         },
    "I" = function (breaks, raw = NULL) {
           sprintf(fmt, utils::as.roman(seq_len(length(breaks) - 1L)))
         },
    "1" = function (breaks, raw = NULL) {
            sprintf(fmt, seq_len(length(breaks) - 1L))
          }
    )

  return(res)
}


#' Defunct: label chopped intervals in a user-defined sequence
#'
#' `r lifecycle::badge("defunct")`
#'
#' `lbl_manual()` is defunct since santoku 1.0.0. It is little used and is not
#' closely related to the rest of the package. It also risks mislabelling
#' intervals, e.g. if intervals are extended. Use of `lbl_manual()` will give
#' an error.
#'
#' @param sequence A character vector of labels.
#' @inherit label-doc
#'
#' @family labelling functions
#'
#' @export
#'
#' @keywords internal
#'
#' @examples
#' \dontrun{
#' chop(1:10, c(2, 5, 8), lbl_manual(c("w", "x", "y", "z")))
#' # ->
#' chop(1:10, c(2, 5, 8), labels = c("w", "x", "y", "z"))
#' }
lbl_manual <- function (sequence, fmt = "%s") {
  lifecycle::deprecate_stop("0.9.0", "lbl_manual()",
                            details = "Just specify `labels = sequence` instead.")
}

test_that("lbl_manual and lbl_endpoint are defunct", {
  lifecycle::expect_defunct(lbl_manual(letters))
  lifecycle::expect_defunct(lbl_endpoint()(lbrk))
})


test_that("lbl_seq", {
  brk <- brk_res(brk_manual(1:3, rep(TRUE, 3)))

  expect_error(lbl_seq("b"))
  expect_error(lbl_seq("a1"))
  expect_error(lbl_seq(c("a", "b")))

  expect_equivalent(lbl_seq()(brk), c("a", "b"))
  expect_equivalent(lbl_seq("A")(brk), c("A", "B"))
  expect_equivalent(lbl_seq("i")(brk), c("i", "ii"))
  expect_equivalent(lbl_seq("I")(brk), c("I", "II"))
  expect_equivalent(lbl_seq("1")(brk), c("1", "2"))

  expect_equivalent(lbl_seq("(a)")(brk), c("(a)", "(b)"))
  expect_equivalent(lbl_seq("i.")(brk), c("i.", "ii."))
  expect_equivalent(lbl_seq("I:")(brk), c("I:", "II:"))
  expect_equivalent(lbl_seq("1)")(brk), c("1)", "2)"))

  brk_many <- brk_res(brk_manual(1:28, rep(TRUE, 28)))
  expect_error(lbl_seq("a")(brk_many))
  expect_error(lbl_seq("A)")(brk_many))
})


test_that("lbl_dash", {
  brk <- brk_res(brk_manual(1:3, rep(TRUE, 3)))
  em_dash <- em_dash()
  expect_equivalent(lbl_dash()(brk), paste0(1:2, em_dash, 2:3))
  expect_equivalent(lbl_dash("/")(brk), c("1/2", "2/3"))
})


test_that("lbl_dash arguments", {
  brk <- brk_res(brk_default(1:3), 1:2)
  expect_equivalent(lbl_dash("-", fmt = "%.2f")(brk), c("1.00-2.00", "2.00-3.00"))

  expect_equivalent(lbl_dash("-", first = "< 2")(brk), c("< 2", "2-3"))
  expect_equivalent(lbl_dash("-", last = "> 2")(brk), c("1-2", "> 2"))

  expect_equivalent(lbl_dash("-", first = "< {r}")(brk), c("< 2", "2-3"))
  expect_equivalent(lbl_dash("-", last = "> {l}")(brk), c("1-2", "> 2"))

  brackets <- function (x) paste0("(", x, ")")
  expect_equivalent(
    lbl_dash("-", fmt = brackets)(brk),
    c("(1)-(2)", "(2)-(3)")
  )

  expect_equivalent(
    lbl_dash("-", fmt = list(width = 2))(brk),
    c(" 1- 2", " 2- 3")
  )

  brk2 <- brk_res(brk_default(c(1, 2, 2, 3)), 1:2)
  expect_equivalent(
    lbl_dash("-", single = "Just {l}")(brk2),
    c("1-2", "Just 2", "2-3")
  )

  qbrk <- brk_res(brk_quantiles(c(0, .5, 1)), x = 0:10)
  expect_equivalent(lbl_dash("-")(qbrk), c("0%-50%", "50%-100%"))
  expect_equivalent(
    lbl_dash("-", fmt = "%.3f")(qbrk),
    c("0.000-0.500", "0.500-1.000")
  )


  lifecycle::expect_deprecated(lbl_dash(raw = TRUE))
  withr::local_options(lifecycle_verbosity = "quiet")

  expect_equivalent(lbl_dash("-", raw = TRUE)(qbrk), c("0-5", "5-10"))
  expect_equivalent(
    lbl_dash("-", raw = TRUE, fmt = "%.2f")(qbrk),
    c("0.00-5.00", "5.00-10.00")
  )
})


test_that("lbl_glue", {
  brk <- brk_res(brk_manual(1:3, rep(TRUE, 3)))
  expect_equivalent(
    lbl_glue("{l} to {r}")(brk),
    c("1 to 2", "2 to 3")
  )

  expect_equivalent(
    lbl_glue("{ifelse(l_closed, '[', '(')}{l},{r}{ifelse(r_closed, ']', ')')}")(brk),
    c("[1,2)", "[2,3)")
  )
})


test_that("lbl_glue arguments", {
  brk <- brk_res(brk_manual(1:3, rep(TRUE, 3)))
  expect_equivalent(
    lbl_glue("{l} to {r}", first = "Up to {r}", last = "Beyond {l}")(brk),
    c("Up to 2", "Beyond 2")
  )

  expect_equivalent(
    lbl_glue("<{l} to {r}>", fmt = "%.1f")(brk),
    c("<1.0 to 2.0>", "<2.0 to 3.0>")
  )

  expect_equivalent(
    lbl_glue("{l} to {r}", fmt = percent)(brk),
    c("100% to 200%", "200% to 300%")
  )

  expect_equivalent(
    lbl_glue("{l}/{r}", fmt = list(width = 2))(brk),
    c(" 1/ 2", " 2/ 3")
  )

  brk2 <- brk_res(brk_manual(c(1,2,2,3), c(TRUE, TRUE, FALSE, TRUE)))
  expect_equivalent(
    lbl_glue("{l} to {r}", single = "{{{l}}}")(brk2),
    c("1 to 2", "{2}", "2 to 3")
  )

  expect_equivalent(
    lbl_glue("<l> to <r>", single = "{<l>}", .open = "<", .close = ">")(brk2),
    c("1 to 2", "{2}", "2 to 3")
  )

  expect_equivalent(
    lbl_glue("{l} to {r}")(brk2),
    c("1 to 2", "2 to 2", "2 to 3")
  )

  expect_equivalent(
    lbl_glue("<{l} to {r}>", fmt = '%.1f', single = "|{sprintf('%.3f', as.numeric(l))}|")(brk2),
    c("<1.0 to 2.0>", "|2.000|", "<2.0 to 3.0>")
  )

  qbrk <- brk_res(brk_quantiles(c(0, .5, 1)), x = 0:10)

  lifecycle::expect_deprecated(lbl_glue("{l} / {r}", raw = TRUE))

  withr::local_options(lifecycle_verbosity = "quiet")
  expect_equivalent(
    lbl_glue("{l} / {r}", raw = TRUE)(qbrk),
    c("0 / 5", "5 / 10")
  )
})



test_that("lbl_endpoints", {
  lbrk <- brk_res(brk_default(c(1, 3, 5)), extend = FALSE)
  expect_equivalent(
    lbl_endpoints()(lbrk),
    c("1", "3")
  )
  expect_equivalent(
    lbl_endpoints(left = FALSE)(lbrk),
    c("3", "5")
  )

  dates <- as.Date("2000-01-01") + c(3, 5)
  dbrk <- brk_res(brk_default(dates),
                  x = as.Date("2000-01-01") + 1:10)
  expect_equivalent(
    lbl_endpoints()(dbrk),
    as.character(dates[1])
  )
})


test_that("lbl_endpoints arguments", {
  lbrk <- brk_res(brk_default(c(1, 3, 5)), extend = FALSE)
  expect_equivalent(
    lbl_endpoints(fmt = "%.2f")(lbrk),
    c("1.00", "3.00")
  )
  expect_equivalent(
    lbl_endpoints(fmt = percent)(lbrk),
    c("100%", "300%")
  )
  expect_equivalent(
    lbl_endpoints(fmt = list(nsmall = 2, decimal.mark = ","))(lbrk),
    c("1,00", "3,00")
  )
})


test_that("lbl_midpoints", {
  lbrk <- brk_res(brk_manual(1:3, rep(TRUE, 3)))
  expect_equivalent(lbl_midpoints()(lbrk), c("1.5", "2.5"))


  dates <- as.Date("2000-01-01") + c(3, 5)
  dbrk <- brk_res(brk_default(dates),
                  x = as.Date("2000-01-01") + 1:10)
  expect_equivalent(
    lbl_endpoints()(dbrk),
    c("2000-01-04")
  )
})


test_that("lbl_midpoints arguments", {
  lbrk <- brk_res(brk_manual(1:3, rep(TRUE, 3)))
  expect_equivalent(lbl_midpoints(first = "{r}")(lbrk), c("2", "2.5"))
  expect_equivalent(lbl_midpoints(last = "{l}")(lbrk), c("1.5", "2"))

  sbrk <- brk_res(brk_manual(c(1, 2, 2, 3), c(TRUE, TRUE, FALSE, TRUE)))
  expect_equivalent(lbl_midpoints(single = "[{l}]")(sbrk), c("1.5", "[2]", "2.5"))

  qbrk <- brk_res(brk_quantiles(c(0, 0.5, 1)), x = 0:10)
  expect_equivalent(lbl_midpoints(fmt = percent)(qbrk), c("25%", "75%"))
  expect_equivalent(
    lbl_midpoints(fmt = list(decimal.mark = ","))(qbrk),
    c("0,25", "0,75")
  )

  lifecycle::expect_deprecated(lbl_midpoints(raw = TRUE))

  withr::local_options(lifecycle_verbosity = "quiet")
  expect_equivalent(lbl_midpoints(raw = TRUE)(qbrk), c("2.5", "7.5"))
})


test_that("lbl_intervals", {
  lbrk <- brk_res(brk_manual(1:3, rep(TRUE, 3)))
  rbrk <- brk_res(brk_manual(1:3, rep(FALSE, 3)))
  expect_equivalent(lbl_intervals()(lbrk), c("[1, 2)", "[2, 3)"))
  expect_equivalent(lbl_intervals()(rbrk), c("(1, 2]", "(2, 3]"))

  lbrk <- brk_res(brk_default(1:3), close_end = TRUE)
  expect_equivalent(lbl_intervals()(lbrk), c("[1, 2)", "[2, 3]"))
  rbrk <- brk_res(brk_default(1:3), close_end = TRUE, left = FALSE)
  expect_equivalent(lbl_intervals()(rbrk), c("[1, 2]", "(2, 3]"))

  sbrk <- brk_res(brk_default(c(1, 2, 2, 3)))
  expect_equivalent(lbl_intervals()(sbrk), c("[1, 2)", "{2}", "(2, 3]"))

  mbrk <- brk_res(brk_manual(1:4, c(FALSE, TRUE, FALSE, TRUE)))
  expect_equivalent(lbl_intervals()(mbrk), c("(1, 2)", "[2, 3]", "(3, 4)"))
})


test_that("lbl_intervals arguments", {
  lbrk <- brk_res(brk_default(c(1, 2, 2, 3) + 0.5))
  expect_equivalent(
    lbl_intervals(fmt = "%.2f")(lbrk),
    c("[1.50, 2.50)", "{2.50}",  "(2.50, 3.50]")
  )

  expect_equivalent(
    lbl_intervals(fmt = list(digits = 2))(lbrk),
    c("[1.5, 2.5)", "{2.5}",  "(2.5, 3.5]")
  )

  lbrk <- brk_res(brk_default(1:3 * 10000))
  expect_equivalent(
    lbl_intervals(fmt = "%2g")(lbrk),
    c("[10000, 20000)", "[20000, 30000]")
  )

  qbrk <- brk_res(brk_quantiles(c(0, 0.5, 1)), x = 0:10)
  expect_equivalent(
    lbl_intervals()(qbrk),
    c("[0%, 50%)", "[50%, 100%]")
  )

  expect_equivalent(
    lbl_intervals(fmt = "%.2f")(qbrk),
    c("[0.00, 0.50)", "[0.50, 1.00]")
  )
  expect_equivalent(
    lbl_intervals(fmt = percent)(qbrk),
    c("[0%, 50%)", "[50%, 100%]")
  )
  expect_equivalent(
    lbl_intervals(fmt = list(digits = 2))(qbrk),
    c("[0.0, 0.5)", "[0.5, 1.0]")
  )

  lbrk <- brk_res(brk_default(c(1, 2, 2, 3)))
  expect_equivalent(
    lbl_intervals(first = "< {r}")(lbrk),
    c("< 2", "{2}", "(2, 3]")
  )
  expect_equivalent(
    lbl_intervals(last = "> {l}")(lbrk),
    c("[1, 2)", "{2}", "> 2")
  )
  expect_equivalent(
    lbl_intervals(single = "[{l}]")(lbrk),
    c("[1, 2)", "[2]", "(2, 3]")
  )

  lifecycle::expect_deprecated(lbl_intervals(raw = TRUE))

  withr::local_options(lifecycle_verbosity = "quiet")

  expect_equivalent(
    lbl_intervals(raw = TRUE)(qbrk),
    c("[0, 5)", "[5, 10]")
  )
  expect_equivalent(
    lbl_intervals(raw = TRUE, fmt = "%.2f")(qbrk),
    c("[0.00, 5.00)", "[5.00, 10.00]")
  )
})


test_that("lbl_discrete", {
  lbrk <- brk_res(brk_manual(1:3, rep(TRUE, 3)))
  rbrk <- brk_res(brk_manual(1:3, rep(FALSE, 3)))

  expect_equivalent(lbl_discrete()(lbrk), c("1", "2"))
  expect_equivalent(lbl_discrete()(rbrk), c("2", "3"))

  lbrk2 <- brk_res(brk_manual(c(1, 3, 5), rep(TRUE, 3)))
  em_dash <- em_dash()
  expect_equivalent(lbl_discrete()(lbrk2), paste0(c(1, 3), em_dash, c(2, 4)))
  expect_equivalent(lbl_discrete(" to ")(lbrk2), c("1 to 2", "3 to 4"))

  lbrk3 <- brk_res(brk_default(c(1, 3, 3, 5)), close_end = TRUE)
  expect_equivalent(lbl_discrete("-")(lbrk3), c("1-2", "3", "4-5"))

  # break containing (1,2) which has no integer in it:
  open_brk <- brk_res(brk_manual(1:3, c(FALSE, TRUE, FALSE)))
  expect_warning(l <- lbl_discrete()(open_brk))
  expect_equivalent(l[1], "--")
})


test_that("lbl_discrete arguments", {
  lbrk <- brk_res(brk_default(c(1, 3, 5)))
  expect_equivalent(
    lbl_discrete("-", fmt = "(%s)")(lbrk),
    c("(1)-(2)", "(3)-(5)")
  )

  brackets <- function (x) paste0("(", x, ")")
  expect_equivalent(
    lbl_discrete("-", fmt = brackets)(lbrk),
    c("(1)-(2)", "(3)-(5)")
  )

  expect_equivalent(
    lbl_discrete("-", fmt = list(nsmall = 1))(lbrk),
    c("1.0-2.0", "3.0-5.0")
  )

  expect_equivalent(
    lbl_discrete("-", first = "<= {r}")(lbrk),
    c("<= 2", "3-5")
  )

  expect_equivalent(
    lbl_discrete("-", last = ">= {l}")(lbrk),
    c("1-2", ">= 3")
  )

  sbrk <- brk_res(brk_default(c(1, 3, 3, 6)))
  expect_equivalent(
    lbl_discrete("-", single = "[{l}]")(sbrk),
    c("1-2", "[3]", "4-6")
  )

  brk1000 <- brk_res(brk_default(c(1, 3, 5) * 1000))
  expect_equivalent(
    lbl_discrete("-", unit = 1000)(brk1000),
    c("1000-2000", "3000-5000")
  )
})


test_that("santoku.infinity", {
  withr::local_options(santoku.infinity = NULL)
  brk12 <- brk_res(brk_default(1:2), extend = TRUE)

  if (l10n_info()[["UTF-8"]]) {
    expect_match(format(brk12), "\u221e")
  } else {
    expect_match(format(brk12), "Inf")
  }

  withr::with_options(list(santoku.infinity = "oo"), {
    expect_match(format(brk12), "oo")
  })
})

test_that("bug: breaks labels don't produce duplicates", {
  brk <- brk_res(brk_default(c(1.333333335, 1.333333336, 1.333333337, 5)))
  lbls <- lbl_intervals()(brk)
  expect_equivalent(anyDuplicated(lbls), 0)
  lbls <- lbl_dash()(brk)
  expect_equivalent(anyDuplicated(lbls), 0)

  brk <- brk_res(brk_quantiles(seq(0, 1, 0.0001)), x = rnorm(10000))
  lbls <- lbl_intervals()(brk)
  expect_equivalent(anyDuplicated(lbls), 0)
  lbls <- lbl_dash()(brk)
  expect_equivalent(anyDuplicated(lbls), 0)
})


test_that("bug: lbl_endpoints() works with no format and non-standard breaks", {
  expect_error(
    chop_quantiles(0:10, 0.5, labels = lbl_endpoints())
    , NA)
  expect_error(
    chop_mean_sd(0:10, labels = lbl_endpoints())
    , NA)
})

#' @rdname chop_quantiles
#'
#' @export
#' @order 2
brk_quantiles <- function (probs, ..., weights = NULL, recalc_probs = FALSE) {
  assert_that(
          is.numeric(probs),
          noNA(probs),
          all(probs >= 0),
          all(probs <= 1),
          is.null(weights) || is.numeric(weights),
          is.flag(recalc_probs)
        )
  probs <- sort(probs)

  function (x, extend, left, close_end) {
    dots <- list(...)
    dots$x <- x
    if (! is.numeric(x) && ! "type" %in% names(dots)) dots$type <- 1
    dots$probs <- probs
    dots$na.rm <- TRUE

    qs <- if (is.null(weights)) {
      do.call(stats::quantile, dots)
    } else {
      rlang::check_installed("Hmisc",
                             reason = "to use `weights` in brk_quantiles()")
      dots$weights <- weights
      do.call(Hmisc::wtd.quantile, dots)
    }

    if (anyNA(qs)) return(empty_breaks()) # data was all NA

    if (anyDuplicated(qs) > 0L) {
      if (! recalc_probs) {
        warning("`x` has duplicate quantiles: break labels may be misleading")
      }
      # We use the left-most probabilities, so e.g. if 0%, 20% and 40% quantiles
      # are all the same number, we'll use the category [0%, 20%).
      # This means we always return intervals that the user asked for, though
      # they may be more misleading than e.g. [0%, 40%).
      illegal_dupes <- find_illegal_duplicates(qs)
      qs <- qs[! illegal_dupes]
      probs <- probs[! illegal_dupes]
    }

    breaks <- create_lr_breaks(qs, left)

    needs <- needs_extend(breaks, x, extend, left, close_end)
    if ((needs & LEFT) > 0)  probs <- c(0, probs)
    if ((needs & RIGHT) > 0) probs <- c(probs, 1)
    breaks <- extend_and_close(breaks, x, extend, left, close_end)

    class(breaks) <- c("quantileBreaks", class(breaks))

    if (recalc_probs) {
      probs <- calculate_ecdf_probs(x, breaks, weights)
    }

    attr(breaks, "scaled_endpoints") <- probs
    names(breaks) <- names(probs)

    breaks
  }
}


#' Calculate the proportions of `x` that is strictly/weakly less than
#' each break
#'
#' @param x A numeric vector
#' @param breaks A breaks object
#' @param weights A vector of weights. Non-NULL weights are unimplemented
#'
#' @return A vector of proportions of `x` that are strictly less than
#' left-closed breaks, and weakly less than right-closed breaks.
#'
#' @noRd
calculate_ecdf_probs <- function (x, breaks, weights) {
  if (! is.numeric(x)) {
    stop("`recalc_probs = TRUE` can only be used with numeric `x`")
  }
  if (! is.null(weights)) {
    stop("`recalc_probs = TRUE` cannot be used with non-null `weights`")
  }

  brk_vec <- unclass_breaks(breaks)
  left_vec <- attr(breaks, "left")

  # proportion of x that is weakly less than x
  prop_lte_brk <- stats::ecdf(x)(brk_vec)
  # proportion of x that is strictly less than x
  prop_lt_brk <- 1 - stats::ecdf(-x)(-brk_vec)
  probs <- ifelse(left_vec, prop_lt_brk, prop_lte_brk)

  # Suppose your breaks are [a, b].
  # You want to expand this?
  probs
}


#' @rdname chop_equally
#'
#' @export
#' @order 2
brk_equally <- function (groups) {
  assert_that(is.count(groups))

  brq <- brk_quantiles(seq(0L, groups)/groups)

  function (x, extend, left, close_end) {
    breaks <- brq(x = x, extend = extend, left = left, close_end = close_end)

    if (length(breaks) < groups + 1) {
      warning("Fewer than ", groups, " intervals created")
    }

    breaks
  }
}


#' @rdname chop_n
#' @export
#' @order 2
brk_n <- function (n, tail = "split") {
  assert_that(is.count(n), tail == "split" || tail == "merge")

  function (x, extend, left, close_end) {
    xs <- sort(x, decreasing = ! left, na.last = NA) # remove NAs
    if (length(xs) < 1L) return(empty_breaks())

    dupes <- duplicated(xs)
    breaks <- xs[0] # ensures breaks has type of xs
    last_x <- xs[length(xs)]

    maybe_merge_tail <- function (breaks, tail) {
      if (tail == "merge" && length(breaks) > 1) {
        breaks <- breaks[-length(breaks)]
      }
      breaks
    }

    # Idea of the algorithm:
    # Loop:
    # if there are no dupes, just take a sequence of each nth element
    #   starting at 1, and exit
    # if there are remaining dupes, then take the first element
    # set m to the (n+1)th element which would normally be next
    # if element m is a dupe:
    #   - we need to go up, otherwise elements to the left will be in the next
    #     interval, and this interval will be too small
    #   - so set m to the next non-dupe (i.e. strictly larger) element
    # now delete the first m-1 elements
    # And repeat
    while (TRUE) {
      if (! any(dupes)) {
        breaks <- c(breaks, xs[seq(1L, length(xs), n)])
        if (length(xs) %% n > 0) {
          breaks <- maybe_merge_tail(breaks, tail)
        }
        break
      }
      breaks <- c(breaks, xs[1])
      m <- n + 1
      if (length(xs) <= n || all(dupes[-seq_len(m - 1)])) {
        if (length(xs) < n) {
          breaks <- maybe_merge_tail(breaks, tail)
        }
        break
      }
      if (dupes[m]) {
        # the first non-dupe will be the next element that is different
        # we know there is one, because we checked above
        m <- m + match(FALSE, dupes[-(1:m)])
      }
      discard <- seq_len(m - 1)
      xs <- xs[-discard]
      dupes <- dupes[-discard]
    }

    breaks <- c(breaks, last_x)
    if (! left) breaks <- rev(breaks)
    breaks <- create_extended_breaks(breaks, x, extend, left, close_end)

    breaks
  }
}

#' Create a breaks object
#'
#' @param obj A sorted vector or a `breaks` object.
#' @param left A logical vector, same length as `obj`.
#'
#' @return A breaks object
#'
#' @noRd
#'
create_breaks <- function (obj, left) {
  if (anyNA(obj)) stop("breaks contained NAs")
  stopifnot(all(obj == sort(obj)))
  stopifnot(is.logical(left))
  stopifnot(length(left) == length(obj))

  if (any(find_illegal_duplicates(obj))) {
    stop("breaks contained more than two consecutive equal values")
  }

  singletons <- singletons(obj)
  l_singletons <- c(singletons, FALSE)
  r_singletons <- c(FALSE, singletons)
  stopifnot(all(left[l_singletons]))
  stopifnot(all(! left[r_singletons]))

  break_classes <- class(obj)
  if (! inherits(obj, "breaks")) break_classes <- c("breaks", break_classes)

  structure(obj, left = left, class = break_classes)
}


create_extended_breaks <- function (obj, x, extend, left, close_end) {
  brks <- create_lr_breaks(obj = obj, left = left)
  extend_and_close(breaks = brks, x = x, extend = extend, left = left,
                     close_end = close_end)
}


create_lr_breaks <- function (obj, left) {
  assert_that(is.flag(left))
  left_vec <- rep(left, length(obj))

  st <- singletons(obj)
  left_vec[which(st)]     <- TRUE
  left_vec[which(st) + 1] <- FALSE

  create_breaks(obj, left_vec)
}


empty_breaks <- function () {
  create_breaks(c(-Inf, Inf), c(TRUE, FALSE))
}


#' Extend `breaks` to the left or right according to `extend` parameter,
#' and close end according to `close_end` parameter
#'
#' @param breaks,x,extend,left,close_end All passed in from `chop()` via
#' a `brk_` inner function
#'
#' @return A `breaks` object.
#' @noRd
extend_and_close <- function (breaks, x, extend, left, close_end) {
  extend_flags <- needs_extend(breaks, x, extend, left, close_end)

  if ((extend_flags & LEFT) > 0) {
    breaks <- extend_endpoint_left(breaks, x, extend)
  }
  if ((extend_flags & RIGHT) > 0) {
    breaks <- extend_endpoint_right(breaks, x, extend)
  }

  breaks <- maybe_close_end(breaks, left = left, close_end = close_end)

  return(breaks)
}



NEITHER <- as.raw(0)
LEFT    <- as.raw(1)
RIGHT   <- as.raw(2)
BOTH <- LEFT | RIGHT


#' Reports if `breaks` will/should be extended.
#'
#' @param breaks A breaks object
#' @param x Data
#' @param extend,left,close_end Parameters passed into `chop`
#'
#' @return Returns LEFT or RIGHT or BOTH only if `breaks` *will*/*must* be
#' extended i.e. gain an extra break, on the respective sides.
#'
#' @details
#' If `extend` is `FALSE`, always returns `NEITHER`. If `breaks` is length
#' zero, always returns `BOTH`.
#'
#' If extend is `NULL` then `left` and `close_end` are taken into account.
#'
#' To test whether `breaks` will be extended on either side, use
#' `(needs & LEFT) > 0` or `(needs & RIGHT) > 0`.
#'
#' @noRd
needs_extend <- function (breaks, x, extend, left, close_end) {
  if (! is.null(extend) && ! extend) return(NEITHER)
  if (length(breaks) < 1L) return(BOTH)

  needs <- NEITHER

  # temporarily close the breaks, to see if unextended closed breaks need
  # extension
  breaks <- maybe_close_end(breaks, left = left, close_end = close_end)
  left_vec <- attr(breaks, "left")

  res <- santoku_cast_common(x, unclass_breaks(breaks))
  x <- res[[1]]
  breaks <- res[[2]]

  min_x <- quiet_min(x)
  max_x <- quiet_max(x)

  if (
          isTRUE(extend)      ||
          min_x < min(breaks) ||
          (! left_vec[1] && min_x == min(breaks))
        ) {
    # "... and if ..." the first break is finite, or will be left-open
    if (is_gt_minus_inf(breaks[1]) || ! left_vec[1]) {
      needs <- needs | LEFT
    }
  }

  if (
          isTRUE(extend)      ||
          max_x > max(breaks) ||
          (left_vec[length(left_vec)] && max_x == max(breaks))
        ) {
    # "... and if ..." the last break is finite, or will be left-closed (right-open)
    if (is_lt_inf(breaks[length(breaks)]) || left_vec[length(left_vec)]) {
      needs <- needs | RIGHT
    }
  }

  return(needs)
}


#' Close end of breaks if close_end is TRUE
#'
#' This never adds a break, it just changes the breaks' `left` attribute.
#' It leaves everything unchanged if `close_end` is `FALSE`.
#'
#' @param breaks,left,close_end Passed in from a `brk_` function
#'
#' @return New breaks object, with the end perhaps closed
#' @noRd
maybe_close_end <- function (breaks, left, close_end) {
  if (! close_end) return(breaks)

  left_vec <- attr(breaks, "left")
  if (left) {
    left_vec[length(left_vec)] <- FALSE
  } else {
    left_vec[1] <- TRUE
  }
  attr(breaks, "left") <- left_vec

  return(breaks)
}


#' Extend the left endpoint of a breaks object according to user parameters
#'
#' This always adds a new break, which is `-Inf` if `extend` is `TRUE`
#' and equal to the minimum of `x` if `extend` is `NULL`.
#'
#' It fixes the `left` attribute if a new singleton break is going to be
#' created.
#'
#' @param breaks,x,extend Passed in from a `brk_` inner function
#'
#' @return A new breaks object
#' @noRd
extend_endpoint_left <- function (breaks, x, extend) {
  left <- attr(breaks, "left")
  q <- quiet_min(x)
  # non-finite q could be Inf if x is empty. Not appropriate for a left endpoint!
  extra_break <- if (is.null(extend) && is_gt_minus_inf(q)) q else class_bounds(x)[1]
  res <- santoku_cast_common(extra_break, unclass_breaks(breaks))
  breaks <- vctrs::vec_c(res[[1]], res[[2]])
  # Ensure that a new "singleton" break has the right TRUE,FALSE left-closed
  # pattern
  if (length(breaks) > 1 && breaks[1] == breaks[2]) {
    left[1] <- FALSE
  }
  breaks <- create_breaks(breaks, c(TRUE, left))

  breaks
}


#' Extend the right endpoint of a breaks object according to user parameters
#'
#' This always adds a new break, which is `Inf` if `extend` is `TRUE`
#' and equal to the maximum of `x` if `extend` is `NULL`.
#'
#' It fixes the `left` attribute if a new singleton break is going to be
#' created.
#'
#' @param breaks,x,extend Passed in from a `brk_` inner function
#'
#' @return A new breaks object
#' @noRd
extend_endpoint_right <- function (breaks, x, extend) {
  left <- attr(breaks, "left")
  q <- quiet_max(x)
  extra_break <- if (is.null(extend) && is_lt_inf(q)) q else class_bounds(x)[2]
  # necessary because min() and max() may unclass things
  res <- santoku_cast_common(unclass_breaks(breaks), extra_break)
  breaks <- vctrs::vec_c(res[[1]], res[[2]])
  lb <- length(breaks)
  # Ensure that a new "singleton" break has the right TRUE,FALSE left-closed
  # pattern
  if (lb > 1 && breaks[lb] == breaks[lb - 1]) {
    left[length(left)] <- TRUE
  }
  breaks <- create_breaks(breaks, c(left, FALSE))

  breaks
}


is_lt_inf <- function (x) {
  x <- tryCatch(strict_as_numeric(x),
                  error = function (...) return(TRUE)
                )
  x < Inf
}


is_gt_minus_inf <- function (x) {
  x <- tryCatch(strict_as_numeric(x),
                  error = function (...) return(TRUE)
                )
  x > -Inf
}


#' Return the infimum and supremum of a class
#'
#' The default tries to cast `c(-Inf, Inf)` to the
#' class. If this fails, it returns `c(min(x), max(x))`
#' and emits a warning.
#'
#' @param x Only used for its class
#'
#' @return A length-two object
#' @noRd
class_bounds <- function (x) {
  UseMethod("class_bounds")
}


#' @export
class_bounds.numeric <- function (x) c(-Inf, Inf)


#' @export
class_bounds.POSIXct <- function (x) {
  as.POSIXct(c(-Inf, Inf), origin = "1970-01-01")
}


#' @export
class_bounds.Date <- function (x) {
  as.Date(c(-Inf, Inf), origin = "1970-01-01")
}


#' @export
class_bounds.difftime <- function (x) {
  as.difftime(c(-Inf, Inf), units = units(x))
}


#' @export
class_bounds.units <- function (x) {
  loadNamespace("units")
  # note: the units() call is from namespace base, not units
  units::set_units(c(-Inf, Inf), units(x), mode = "standard")
}


#' @export
class_bounds.integer64 <- function (x) {
  loadNamespace("bit64")
  bit64::lim.integer64()
}


#' @export
class_bounds.zoo <- function (x) {
  loadNamespace("zoo")
  zoo::zoo(c(-Inf, Inf))
}


#' @export
class_bounds.default <- function (x) {
  tryCatch(
    vctrs::vec_cast(c(-Inf, Inf), x),
    error = function(...) {
      warning("Class '", paste(class(x), collapse = "', '"),
              "' has no natural endpoints corresponding to +/-Inf for `extend = TRUE`;")
      c(quiet_min(x), quiet_max(x))
    }
  )
}


#' Removes the "breaks" class, and all subclasses, from a break object
#'
#' @param breaks A breaks object
#'
#' @return The object, with any remaining (super)classes
#' @noRd
unclass_breaks <- function (breaks) {
  assert_that(is.breaks(breaks))

  class_pos <- inherits(breaks, "breaks", which = TRUE)
  superclasses <- class(breaks)[-seq_len(class_pos)]

  class(breaks) <- if (length(superclasses) == 0 ) {
                     NULL
                   } else {
                     superclasses
                   }

  # this helps vec_cast_common deal with unusual types of breaks
  attr(breaks, "left") <- NULL

  breaks
}


#' Return a sequence of width `width`
#'
#' @param width An object representing a width
#' @param start Element to start from
#' @param until Result must be just long enough to cover this element
#'
#' @return A sequence of breaks
#' @noRd
sequence_width <- function(width, start, until) {
  UseMethod("sequence_width")
}


#' @export
sequence_width.default <- function (width, start, until) {
  breaks <- seq(start, until, width)

  too_short <- if (sign(width) > 0) {
    breaks[length(breaks)] < until
  } else {
    breaks[length(breaks)] > until
  }

  # length(breaks) == 1L captures when start == max_x
  if (too_short || length(breaks) == 1L) {
    breaks <- c(breaks, breaks[length(breaks)] + width)
  }

  breaks
}


#' @export
sequence_width.Period <- function(width, start, until) {
  loadNamespace("lubridate")

  if (as.numeric(until - start) %% as.numeric(width) != 0 || until == start) {
    # extend to cover all data / ensure at least one interval
    until <- lubridate::add_with_rollback(until, width)
  }
  # alternative to seq, using Period arithmetic
  # We find the number n of widths that gets beyond seq_end
  # and add (width * 0:n) to start
  # normally this would be ceiling((seq_end - start)/width)
  # we calculate it roughly using a Duration
  n_intervals <- ceiling((until - start)/lubridate::as.duration(width))
  breaks <- lubridate::add_with_rollback(start, (seq(0, n_intervals) * width))

  last_break <- breaks[length(breaks)]
  too_short <- if (width > 0) {
    last_break < until
  } else {
    last_break > until
  }
  if (too_short) {
    breaks <- c(breaks, lubridate::add_with_rollback(last_break, width))
  }

  breaks
}

#' @name chop-doc
#' @param ... Passed to [chop()].
#' @return
#' `chop_*` functions return a [`factor`] of the same length as `x`.
#'
#' `brk_*` functions return a [`function`] to create `breaks`.
#'
#' `tab_*` functions return a contingency [table()].
NULL


#' Cut data into intervals
#'
#' `chop()` cuts `x` into intervals. It returns a [`factor`] of the same length as
#' `x`, representing which interval contains each element of `x`.
#' `kiru()` is an alias for `chop`.
#' `tab()` calls `chop()` and returns a contingency [table()] from the result.
#'
#' @param x A vector.
#' @param breaks A numeric vector of cut-points, or a function to create
#'   cut-points from `x`.
#' @param labels A character vector of labels or a function to create labels.
#' @param extend Logical. If `TRUE`, always extend breaks to `+/-Inf`. If `NULL`,
#'   extend breaks to `min(x)` and/or `max(x)` only if necessary. If `FALSE`, never
#'   extend.
#' @param left Logical. Left-closed or right-closed breaks?
#' @param close_end Logical. Close last break at right? (If `left` is `FALSE`,
#'   close first break at left?)
#' @param raw Logical. Use raw values in labels?
#' @param drop Logical. Drop unused levels from the result?
#'
#' @details
#'
#' `x` may be a numeric vector, or more generally, any vector which can be
#' compared with `<` and `==` (see [Ops][groupGeneric]). In particular [Date]
#' and [date-time][DateTimeClasses] objects are supported. Character vectors
#' are supported with a warning.
#'
#' ## Breaks
#'
#' `breaks` may be a vector or a function.
#'
#' If it is a vector, `breaks` gives the interval endpoints. Repeating a value
#' creates a "singleton" interval, which contains only that value.
#' For example `breaks = c(1, 3, 3, 5)` creates 3 intervals:
#' \code{[1, 3)}, \code{{3}} and \code{(3, 5]}.
#'
#' If `breaks` is a function, it is called with the `x`, `extend`, `left` and
#' `close_end` arguments, and should return an object of class `breaks`.
#' Use `brk_*` functions to create a variety of data-dependent breaks.
#'
#' Names of `breaks` may be used for labels. See "Labels" below.
#'
#' ## Options for breaks
#'
#' By default, left-closed intervals are created. If `left` is `FALSE`,
#' right-closed intervals are created.
#'
#' If `close_end` is `TRUE` the final break (or first break if `left` is `FALSE`)
#' will be closed at both ends. This guarantees that all values `x` with
#' `min(breaks) <= x <= max(breaks)` are included in the intervals.
#'
#' Before version 0.9.0, `close_end` was `FALSE` by default, and also behaved
#' differently with respect to extended breaks: see "Extending intervals" below.
#'
#' Using [mathematical set notation][lbl_intervals()]:
#'
#' * If `left` is `TRUE` and `close_end` is `TRUE`, breaks will look like
#'   \code{[b1, b2), [b2, b3) ... [b_(n-1), b_n]}.
#' * If `left` is `FALSE` and `close_end` is `TRUE`, breaks will look like
#'    \code{[b1, b2], (b2, b3] ... (b_(n-1), b_n]}.
#' * If `left` is `TRUE` and `close_end` is `FALSE`, all breaks will look like
#'   \code{... [b1, b2) ...}.
#' * If `left` is `FALSE` and `close_end` is `FALSE`, all breaks will look like
#'   \code{... (b1, b2] ...}.
#'
#' ## Extending intervals
#'
#' If `extend` is `TRUE`, intervals will be extended to \code{[-Inf,
#' min(breaks))} and \code{(max(breaks), Inf]}.
#'
#' If `extend` is `NULL` (the default), intervals will be extended to
#' \code{[min(x), min(breaks))} and \code{(max(breaks), max(x)]}, only if
#' necessary, i.e. only if elements of `x` would be outside the unextended
#' breaks.
#'
#' If `extend` is `FALSE`, intervals are never extended.
#'
#' Note that even when `extend = TRUE`, extended intervals will be
#' dropped from the factor levels if they contain no elements and `drop = TRUE`.
#'
#' `close_end` is only relevant if intervals are not extended;
#' extended intervals are always closed on the outside. This is a change from
#' previous behaviour. Up to version 0.8.0, `close_end` was applied to the
#' last user-specified interval, before any extended intervals were created.
#'
#' Since 1.1.0, infinity is represented as \eqn{\infty}{the infinity symbol} 
#' in breaks on unicode platforms. Set  `options(santoku.infinity = "Inf")` 
#' to get the old behaviour.
#'
#' ## Labels
#'
#' `labels` may be a character vector. It should have the same length as the
#' (possibly extended) number of intervals. Alternatively, `labels` may be a
#' `lbl_*` function such as [lbl_dash()].
#'
#' If `breaks` is a named vector, then names of `breaks` will be
#' used as labels for the interval starting at the corresponding element. This
#' overrides the `labels` argument (but unnamed breaks will still use `labels`).
#' This feature is `r lifecycle::badge("experimental")`.
#'
#' If `labels` is `NULL`, then integer codes will be returned instead of a
#' factor.
#'
#' If `raw` is `TRUE`, labels will show the actual interval endpoints, usually
#' numbers. If `raw` is `FALSE` then labels may show other objects, such
#' as quantiles for [chop_quantiles()] and friends, proportions of the range for
#' [chop_proportions()], or standard deviations for [chop_mean_sd()].
#'
#'  If `raw` is `NULL` then `lbl_*` functions will use their default (usually
#'  `FALSE`). Otherwise, the `raw` argument to `chop()` overrides `raw` arguments
#'  passed into `lbl_*` functions directly.
#'
#'
#' ## Miscellaneous
#'
#' `NA` values in `x`, and values which are outside the extended endpoints,
#' return `NA`.
#'
#' `kiru()` is a synonym for `chop()`. If you load `{tidyr}`, you can use it to
#' avoid confusion with `tidyr::chop()`.
#'
#' Note that `chop()`, like all of R, uses binary arithmetic. Thus, numbers may
#' not be exactly equal to what you think they should be. There is an example
#' below.
#'
#' @return
#' `chop()` returns a [`factor`] of the same length as `x`, representing the
#' intervals containing the value of `x`.
#'
#' `tab()` returns a contingency [table()].
#'
#' @export
#'
#' @family chopping functions
#'
#' @seealso [base::cut()], [`non-standard-types`] for chopping objects that
#'   aren't numbers.
#'
#' @examples
#'
#' chop(1:7, c(2, 4, 6))
#'
#' chop(1:7, c(2, 4, 6), extend = FALSE)
#'
#' # Repeat a number for a singleton break:
#' chop(1:7, c(2, 4, 4, 6))
#'
#' chop(1:7, c(2, 4, 6), left = FALSE)
#'
#' chop(1:7, c(2, 4, 6), close_end = FALSE)
#'
#' chop(1:7, brk_quantiles(c(0.25, 0.75)))
#'
#' # A single break is fine if `extend` is not `FALSE`:
#' chop(1:7, 4)
#'
#' # Floating point inaccuracy:
#' chop(0.3/3, c(0, 0.1, 0.1, 1), labels = c("< 0.1", "0.1", "> 0.1"))
#'
#' # -- Labels --
#'
#' chop(1:7, c(Lowest = 1, Low = 2, Mid = 4, High = 6))
#'
#' chop(1:7, c(2, 4, 6), labels = c("Lowest", "Low", "Mid", "High"))
#'
#' chop(1:7, c(2, 4, 6), labels = lbl_dash())
#'
#' # Mixing names and other labels:
#' chop(1:7, c("<2" = 1, 2, 4, ">=6" = 6), labels = lbl_dash())
#'
#' # -- Non-standard types --
#'
#' chop(as.Date("2001-01-01") + 1:7, as.Date("2001-01-04"))
#'
#' suppressWarnings(chop(LETTERS[1:7], "D"))
#'
#'
chop <- function (x, breaks,
          labels    = lbl_intervals(),
          extend    = NULL,
          left      = TRUE,
          close_end = TRUE,
          raw       = NULL,
          drop      = TRUE
        ) {
  assert_that(
          is.flag(extend) || is.null(extend),
          is.flag(left),
          is.flag(close_end),
          is.flag(drop),
          is.flag(raw) || is.null(raw)
        )

  if (! is.function(breaks)) breaks <- brk_default(breaks)
  breaks <- breaks(x, extend, left, close_end)
  assert_that(is.breaks(breaks), length(breaks) >= 2L)

  codes <- categorize(x, breaks)

  if (is.null(labels)) return(codes)

  lbls <- if (is.function(labels)) {
    if (is.null(raw)) labels(breaks) else labels(breaks, raw = raw)
  } else {
    labels
  }
  lbls <- add_break_names(lbls, breaks)
  stopifnot(length(lbls) == length(breaks) - 1)

  real_codes <- if (drop) unique(codes[! is.na(codes)]) else TRUE
  if (anyDuplicated(lbls[real_codes])) {
    stop("Duplicate labels found: ", paste(lbls, collapse = ", "))
  }

  result <- factor(codes, levels = seq.int(length(breaks) - 1L),
        labels = lbls)
  if (drop) result <- droplevels(result)

  return(result)
}


#' @rdname chop
#' @export
kiru <- chop


#' Chop data precisely (for programmers)
#'
#' `fillet()` calls [chop()] with `extend = FALSE` and `drop = FALSE`. This
#' ensures that you get only the `breaks` and `labels` you ask for. When
#' programming, consider using `fillet()` instead of `chop()`.
#'
#' @inheritParams chop
#'
#' @return `fillet()` returns a [`factor`] of the same length as `x`, representing
#'   the intervals containing the value of `x`.
#'
#' @family chopping functions
#'
#' @export
#'
#' @examples
#' fillet(1:10, c(2, 5, 8))
fillet <- function (
            x,
            breaks,
            labels    = lbl_intervals(),
            left      = TRUE,
            close_end = TRUE,
            raw       = NULL
          ) {
  chop(x, breaks, labels, left = left, close_end = close_end, extend = FALSE,
      raw = raw, drop = FALSE)
}

#' Chop by standard deviations
#'
#' Intervals are measured in standard deviations on either side of the
#' mean.
#'
#' In version 0.7.0, these functions changed to specifying `sds` as a vector.
#' To chop 1, 2 and 3 standard deviations around the mean, write
#' `chop_mean_sd(x, sds = 1:3)` instead of `chop_mean_sd(x, sd = 3)`.
#'
#' @param sds Positive numeric vector of standard deviations.
#' @param sd  `r lifecycle::badge("deprecated")`
#'
#' @inheritParams chop
#' @inherit chop-doc params return
#'
#' @family chopping functions
#'
#' @export
#' @order 1
#'
#' @examples
#' chop_mean_sd(1:10)
#'
#' chop(1:10, brk_mean_sd())
#'
#' @importFrom lifecycle deprecated
chop_mean_sd <- function (
                  x,
                  sds = 1:3,
                  ...,
                  raw = FALSE,
                  sd  = deprecated()
                ) {
  chop(x, brk_mean_sd(sds = sds, sd = sd), ..., raw = raw)
}


#' Chop using pretty breakpoints
#'
#' `chop_pretty()` uses [base::pretty()] to calculate breakpoints
#' which are 1, 2 or 5 times a power of 10. These look nice in graphs.
#'
#' [base::pretty()] tries to return `n+1` breakpoints, i.e. `n` intervals, but
#' note that this is not guaranteed. There are methods for Date and POSIXct
#' objects.
#'
#' For fine-grained control over [base::pretty()] parameters, use
#' `chop(x, brk_pretty(...))`.
#'
#' @inheritParams chop
#' @inherit chop-doc params return
#' @param n Positive integer passed to [base::pretty()]. How many intervals to chop into?
#' @param ... Passed to [chop()] by `chop_pretty()` and `tab_pretty()`; passed
#'   to [base::pretty()] by `brk_pretty()`.
#'
#' @export
#' @order 1
#'
#' @examples
#' chop_pretty(1:10)
#'
#' chop(1:10, brk_pretty(n = 5, high.u.bias = 0))
#'
chop_pretty <- function (x, n = 5, ...) {
  chop(x, brk_pretty(n = n), ...)
}


#' Chop using an existing function
#'
#' `chop_fn()` is a convenience wrapper: `chop_fn(x, foo, ...)`
#' is the same as `chop(x, foo(x, ...))`.
#'
#' @param fn A function which returns a numeric vector of breaks.
#' @param ... Further arguments to `fn`
#' @inheritParams chop
#' @inherit chop-doc params return
#'
#' @export
#' @order 1
#' @family chopping functions
#' @examples
#'
#' if (requireNamespace("scales")) {
#'   chop_fn(rlnorm(10), scales::breaks_log(5))
#'   # same as
#'   # x <- rlnorm(10)
#'   # chop(x, scales::breaks_log(5)(x))
#' }
#'
chop_fn <- function (
             x,
             fn,
             ...,
             extend = NULL,
             left = TRUE,
             close_end = TRUE,
             raw = NULL,
             drop = TRUE
) {
  chop(x, brk_fn(fn, ...), extend = extend, left = left, close_end = close_end,
         raw = raw, drop = drop)
}

# Generated by using Rcpp::compileAttributes() -> do not edit by hand
# Generator token: 10BE3573-1514-4C36-9D1C-5A225CD40393

categorize_impl <- function(x, breaks, left) {
    .Call(`_santoku_categorize_impl`, x, breaks, left)
}

#' @import assertthat
NULL

#' A versatile cutting tool for R: package overview and options
#'
#' santoku is a tool for cutting data into intervals. It provides
#' the function [chop()], which is similar to base R's [cut()] or `Hmisc::cut2()`.
#' `chop(x, breaks)` takes a vector `x` and returns a factor of the
#' same length, coding which interval each element of `x` falls into.
#'
#' Here are some advantages of santoku:
#'
#' * By default, `chop()` always covers the whole range of the data, so you
#'   won't get unexpected `NA` values.
#'
#' * Unlike `cut()` or `cut2()`, `chop()` can handle single values as well as
#'   intervals. For example, `chop(x, breaks = c(1, 2, 2, 3))` will create a
#'   separate factor level for values exactly equal to 2.
#'
#' * Flexible and easy labelling.
#'
#' * Convenience functions for creating quantile intervals, evenly-spaced
#'  intervals or equal-sized groups.
#'
#' * Convenience functions to quickly tabulate chopped data.
#'
#' * Can chop numbers, dates, date-times and other objects.
#'
#' These advantages make santoku especially useful for exploratory analysis,
#' where you may not know the range of your data in advance.
#'
#' To get started, read the vignette:
#'
#' ```r
#' vignette("santoku")
#' ```
#' 
#' For more details, start with the documentation for [chop()].
#'
#' # Options
#'
#' Santoku has two options:
#'
#' * `options("santoku.infinity")` sets the symbol for infinity in breaks. The default is
#'   `NULL`, in which case the infinity symbol is used on platforms that support it, otherwise
#'   `"Inf"` is used.
#'
#' * `options("santoku.warn_character")` warns if you try to chop a character vector. Set to
#'   `FALSE` to turn off this warning.
"_PACKAGE"

# The following block is used by usethis to automatically manage
# roxygen namespace tags. Modify with care!
## usethis namespace: start
#' @importFrom Rcpp sourceCpp
#' @useDynLib santoku, .registration = TRUE
## usethis namespace: end
NULL

#' @param breaks A numeric vector.
#' @name breaks-doc
#' @return A function which returns an object of class `breaks`.
NULL


#' Create a standard set of breaks
#'
#' @inherit breaks-doc params return
#' @export
#'
#' @examples
#'
#' chop(1:10, c(2, 5, 8))
#' chop(1:10, brk_default(c(2, 5, 8)))
#'
brk_default <- function (breaks) {
  assert_that(noNA(breaks))

  function (x, extend, left, close_end) {
    create_extended_breaks(breaks, x, extend, left, close_end)
  }
}


#' @rdname chop_spikes
#' @export
#' @order 2
brk_spikes <- function (breaks, n = NULL, prop = NULL) {
  assert_that(
    is.number(n) || is.number(prop),
    is.null(n) || is.null(prop),
    msg = "exactly one of `n` and `prop` must be a scalar numeric"
  )
  assert_that(
    # it's ok for one of these to be null
    n >= 0 || prop >= 0
  )

  if (! is.function(breaks)) breaks <- brk_default(breaks)

  function (x, extend, left, close_end) {
    breaks <- breaks(x, extend, left, close_end)
    break_elements <- unclass_breaks(breaks)
    left_vec <- attr(breaks, "left")

    spikes <- find_spikes(x, n, prop)
    # We sort spikes in decreasing order so that when we add elements,
    # earlier elements remain in place.
    spikes <- sort(spikes, decreasing = TRUE)

    for (spike in spikes) {
      # We could use match() here to go faster, or even put it outside the loop.
      match_location <- which(spike == break_elements)
      n_matches <- length(match_location)
      # If two break elements match the spike, it's already a singleton:
      # we don't need to do anything.
      if (n_matches >= 2L) next
      if (n_matches == 1L) {
        # We turn the single matching break into a singleton and make sure
        # that left is c(TRUE, FALSE)
        break_elements <- append(break_elements, spike, after = match_location)
        left_vec <- append(left_vec, FALSE, after = match_location)
        left_vec[match_location] <- TRUE
      } else {
        # We add a singleton break at `spike`
        insert_location <- quiet_max(which(spike > break_elements))
        if (insert_location <= 0) insert_location <- 0
        break_elements <- append(break_elements, rep(spike, 2),
                                 after = insert_location)
        left_vec <- append(left_vec, c(TRUE, FALSE), after = insert_location)
      }
    }

    create_breaks(break_elements, left = left_vec)
  }
}


#' Class representing a set of intervals
#'
#' @param x A breaks object
#' @param ... Unused
#'
#' @name breaks-class
NULL


#' @rdname breaks-class
#' @export
format.breaks <- function (x, ...) {
  if (length(x) < 2) return("Breaks object: no complete intervals")
  paste0("Breaks object: ", paste(lbl_intervals()(x), collapse = " "))
}


#' @rdname breaks-class
#' @export
print.breaks <- function (x, ...) cat(format(x, ...))


#' @rdname breaks-class
#' @export
is.breaks <- function (x, ...) inherits(x, "breaks")


on_failure(is.breaks) <- function (call, env) {
  paste0(deparse(call$x), " is not an object of class `breaks`")
}

#' Tips for chopping non-standard types
#'
#' Santoku can handle many non-standard types.
#'
#' * If objects can be compared using `<`, `==` etc. then they should
#'   be choppable.
#' * Objects which can't be converted to numeric are handled within R code,
#'   which may be slower.
#' * Character `x` and `breaks` are chopped with a warning.
#' * If `x` and `breaks` are not the same type, they should be able to
#'   be cast to the same type, usually using [vctrs::vec_cast_common()].
#' * Not all chopping operations make sense, for example, [chop_mean_sd()]
#'   on a character vector.
#' * For indexed objects such as [stats::ts()] objects, indices will be dropped
#'   from the result.
#' * If you get errors, try setting `extend = FALSE` (but also file a bug report).
#' * To request support for a type, open an issue on Github.
#'
#' @name non-standard-types
#' @seealso brk-width-for-Datetime
NULL

#' Label chopped intervals using the `glue` package
#'
#' Use `"{l}"` and `"{r}"` to show the left and right endpoints of the intervals.
#'
#' @inherit label-doc
#' @inherit first-last-doc params
#' @param label A glue string passed to [glue::glue()].
#' @param ... Further arguments passed to [glue::glue()].
#'
#' @details
#'
#' The following variables are available in the glue string:
#'
#' * `l` is a character vector of left endpoints of intervals.
#' * `r` is a character vector of right endpoints of intervals.
#' * `l_closed` is a logical vector. Elements are `TRUE` when the left
#'   endpoint is closed.
#' * `r_closed` is a logical vector, `TRUE` when the right endpoint is closed.
#'
#' Endpoints will be formatted by `fmt` before being passed to `glue()`.
#'
#' @family labelling functions
#'
#' @export
#'
#' @examples
#' tab(1:10, c(1, 3, 3, 7),
#'     labels = lbl_glue("{l} to {r}", single = "Exactly {l}"))
#'
#' tab(1:10 * 1000, c(1, 3, 5, 7) * 1000,
#'     labels = lbl_glue("{l}-{r}",
#'                       fmt = function(x) prettyNum(x, big.mark=',')))
#'
#' # reproducing lbl_intervals():
#' interval_left <- "{ifelse(l_closed, '[', '(')}"
#' interval_right <- "{ifelse(r_closed, ']', ')')}"
#' glue_string <- paste0(interval_left, "{l}", ", ", "{r}", interval_right)
#' tab(1:10, c(1, 3, 3, 7), labels = lbl_glue(glue_string, single = "{{{l}}}"))
#'
lbl_glue <- function (
              label,
              fmt    = NULL,
              single = NULL,
              first  = NULL,
              last   = NULL,
              raw    = FALSE,
              ...
            ) {
  assert_that(
    is.string(label),
    is.null(fmt) || is_format(fmt),
    is.string(first) || is.null(first),
    is.string(last) || is.null(last),
    is.flag(raw)
  )

  if (! isFALSE(raw)) {
    lifecycle::deprecate_warn("0.9.0", "lbl_glue(raw)", "chop(raw)")
  }

  RAW <- raw # avoid "recursive default argument reference"
  function (breaks, raw = RAW) {
    assert_that(is.breaks(breaks))

    len_breaks <- length(breaks)

    labels <- character(len_breaks - 1)

    elabels <- endpoint_labels(breaks, raw = raw, fmt = fmt)

    l <- elabels[-len_breaks]
    r <- elabels[-1]

    left <- attr(breaks, "left")
    # Breaks like [1, 2) [2, 3] have
    # left TRUE, TRUE, FALSE for breaks 1,2,3
    # The first two TRUEs say that the left brackets are closed
    # The last two TRUE & FALSE say that the right brackets are open
    # and closed respectively. So:
    l_closed <- left[-len_breaks]
    r_closed <- ! left[-1]

    # check ... for anything not in glue::glue args
    # effectively, we move any user-supplied arguments into
    # an environment specifically for glue
    # this is mostly to make the lbl_midpoints() hack
    # of passing in `m` work
    dots <- rlang::enexprs(...)
    glue_env <- new.env()
    not_glue_args <- setdiff(names(dots), names(formals(glue::glue)))
    for (nm in not_glue_args) {
      assign(deparse(dots[[nm]]),
               eval(dots[[nm]], parent.frame()),
               glue_env
             )
    }
    labels <- glue::glue(label, l = l, r = r, l_closed = l_closed,
                         r_closed = r_closed, ..., .envir = glue_env)

    if (! is.null(single)) {
      # which breaks are singletons?
      singletons <- singletons(breaks)

      labels[singletons] <- glue::glue(single,
                                         l = l[singletons],
                                         r = r[singletons],
                                         l_closed = l_closed[singletons],
                                         r_closed = r_closed[singletons],
                                         ...,
                                         .envir = glue_env
                                       )
    }

    if (! is.null(first)) {
      labels[1] <- glue::glue(first, l = l[1], r = r[1],
                                l_closed = l_closed[1],
                                r_closed = r_closed[1],
                                ...,
                                .envir = glue_env
                              )
    }

    if (! is.null(last)) {
      ll <- len_breaks - 1
      labels[ll] <- glue::glue(last, l = l[ll], r = r[ll],
                                 l_closed = l_closed[ll],
                                 r_closed = r_closed[ll],
                                 ...,
                                 .envir = glue_env
                               )
    }

    return(labels)
  }
}

#' Categorize `x` according to breaks
#'
#' @param x A vector of data
#' @param breaks A breaks object
#'
#' @return A set of vector codes
#' @noRd
categorize <- function (x, breaks) {
  # we first cast to the most informative common type. Then to numeric.

  left <- attr(breaks, "left")
  res <- santoku_cast_common(x, unclass_breaks(breaks))
  # vec_cast won't accept e.g. characters but it also won't convert e.g. Dates
  # as.numeric accepts both
  # We want to convert things to numeric objects, but NB, not all
  # numeric objects will work OK in categorize_impl
  x <- tryCatch(strict_as_numeric(res[[1]]),
                  error   = function (...) res[[1]]
                )
  breaks <- tryCatch(strict_as_numeric(res[[2]]),
                       error   = function (...) res[[2]]
                     )

  # we use is_bare_numeric here because e.g. large integer64 vectors will
  # fail in categorize_impl()
  codes <- if (rlang::is_bare_numeric(x) && rlang::is_bare_numeric(breaks)) {
    categorize_impl(x, breaks, left)
  } else {
    categorize_non_numeric(x, breaks, left)
  }

  codes
}


categorize_non_numeric <- function (x, breaks, left) {

  if (is.character(x) || is.character(breaks)) {
    if (getOption("santoku.warn_character", TRUE)) {
      warning_statement <- paste(
        "`x` or `breaks` is of type character, using lexical sorting.",
        "To turn off this warning, run:",
        "  options(santoku.warn_character = FALSE)",
        collapse = "\n")
      warning(warning_statement)
    }
  }

  codes <- rep(NA_integer_, length(x))

  for (j in seq_len(length(breaks) - 1)) {
    more_than_j <- x > breaks[j]
    less_than_j_plus_one <- x < breaks[j+1]
    equals_j <- x == breaks[j]
    equals_j_plus_one <- x == breaks[j+1]

    codes[more_than_j & less_than_j_plus_one] <- j
    if (left[j]) codes[equals_j] <- j
    if (! left[j+1]) codes[equals_j_plus_one] <- j
  }

  codes
}

test_that("brk_manual", {
  expect_silent(brk_res(brk_manual(1:3, c(TRUE, TRUE, TRUE))))
  expect_silent(brk_res(brk_manual(c(1, 2, 2, 3), c(TRUE, TRUE, FALSE, TRUE))))
  expect_silent(brk_res(brk_manual(c(-Inf, 1, Inf), rep(TRUE, 3))))

  # singletons must have FALSE, TRUE
  expect_error(brk_res(brk_manual(c(1, 2, 2, 3), c(TRUE, FALSE, FALSE, TRUE))))
  # out of order:
  expect_error(brk_res(brk_manual(c(0, 3, 1), c(TRUE, TRUE, TRUE))))
})


test_that("categorize works", {

  x <- seq(0.5, 3.5, 0.5)

  breaks <- brk_res(brk_manual(1:3, c(TRUE, TRUE, TRUE)))
  r <- categorize(x, breaks)
  expect_equivalent(r, c(NA, 1, 1, 2, 2, NA, NA))

  breaks <- brk_res(brk_manual(1:3, c(TRUE, TRUE, FALSE)))
  r <- categorize(x, breaks)
  expect_equivalent(r, c(NA, 1, 1, 2, 2, 2, NA))

  breaks <- brk_res(brk_manual(c(1, 2, 2, 3), c(TRUE, TRUE, FALSE, TRUE)))
  r <- categorize(x, breaks)
  expect_equivalent(r, c(NA, 1, 1, 2, 3, NA, NA))

  x <- c(Inf, 1, -Inf, NA, NaN)
  breaks <- brk_res(brk_manual(1:3, c(TRUE, TRUE, TRUE)))
  r <- categorize(x, breaks)
  expect_equivalent(r, c(NA, 1, NA, NA, NA))
})


test_that("categorize_impl/categorize_non_numeric equivalence", {
  replicate(100, {
    n <- 10
    x <- rnorm(n) * 10

    breaks_pop <- c(-Inf, -5:5, Inf)
    b <- sort(sample(breaks_pop, n, replace = FALSE))
    left <- sample(c(TRUE, FALSE), n, replace = TRUE)
    x[c(3,5,7)] <- sample(breaks_pop, 3)

    b[3] <- b[2]
    b[8] <- b[7]
    left[2:3] <- left[7:8] <- c(TRUE, FALSE)

    ci <- categorize_impl(x, b, left)
    cnn <- categorize_non_numeric(x, b, left)
    expect_equal(ci, cnn)
  })
})

#' @rdname chop
#' @export
#' @examples
#' tab(1:10, c(2, 5, 8))
#'
tab <- function (
         x,
         breaks,
         labels    = lbl_intervals(),
         extend    = NULL,
         left      = TRUE,
         close_end = TRUE,
         raw       = NULL,
         drop      = TRUE
       ) {
  default_table(
    chop(
      x         = x,
      breaks    = breaks,
      labels    = labels,
      extend    = extend,
      left      = left,
      close_end = close_end,
      raw       = raw,
      drop      = drop
    )
  )
}


#' @rdname chop_width
#' @export
#' @order 3
#' @examples
#' tab_width(1:10, 2, start = 0)
#'
tab_width <- function (
               x,
               width,
               start,
               ...,
               left = sign(width) > 0
             ) {
  default_table(
    chop_width(x = x, width = width, start = start, ..., left = left)
  )
}


#' @rdname chop_evenly
#' @export
#' @order 3
tab_evenly <- function (
                x,
                intervals,
                ...
              ) {
  default_table(
    chop_evenly(x = x, intervals = intervals, ...)
  )
}


#' @rdname chop_proportions
#' @export
#' @order 3
tab_proportions <- function (
                     x,
                     proportions,
                     ...,
                     raw = TRUE
                   ) {
  default_table(
    chop_proportions(x = x, proportions = proportions, ..., raw = raw)
  )
}


#' @rdname chop_n
#' @export
#' @order 3
#' @examples
#' tab_n(1:10, 5)
#'
#' # fewer elements in one group
#' tab_n(1:10, 4)
#'
tab_n <- function (
           x,
           n,
           ...,
           tail = "split"
         ) {
  default_table(chop_n(x = x, n = n, ..., tail = tail))
}


#' @rdname chop_mean_sd
#' @export
#' @order 3
#' @examples
#' tab_mean_sd(1:10)
#'
tab_mean_sd <- function (
                 x,
                 sds = 1:3,
                 ...,
                 raw = FALSE
               ) {
  default_table(chop_mean_sd(x = x, sds = sds, ..., raw = raw))
}


#' @rdname chop_pretty
#' @export
#' @order 3
#' @examples
#' tab_pretty(1:10)
#'
tab_pretty <- function (x, n = 5, ...) {
  default_table(chop_pretty(x = x, n = n, ...))
}


#' @rdname chop_quantiles
#' @export
#' @order 3
#' @examples
#' set.seed(42)
#' tab_quantiles(rnorm(100), probs = 1:3/4, raw = TRUE)
#'
tab_quantiles <- function (
                   x,
                   probs,
                   ...,
                   left      = is.numeric(x),
                   raw       = FALSE
                 ) {
  default_table(
    chop_quantiles(x = x, probs = probs, ..., left = left, raw = raw)
  )
}


#' @rdname chop_quantiles
#' @export
#' @order 3
tab_deciles <- function (x, ...) {
  default_table(chop_deciles(x = x, ...))
}


#' @rdname chop_equally
#' @export
#' @order 3
tab_equally <- function (
                 x,
                 groups,
                 ...,
                 left      = is.numeric(x),
                 raw       = TRUE
               ) {
  default_table(
    chop_equally(x = x, groups = groups, ..., left = left, raw = raw)
  )
}


#' @rdname chop_fn
#' @export
#' @order 3
tab_fn <- function (
            x,
            fn,
            ...,
            extend = NULL,
            left = TRUE,
            close_end = TRUE,
            raw = NULL,
            drop = TRUE
          ) {
  default_table(
    chop_fn(x = x, fn = fn, ..., extend = extend, left = left,
              close_end = close_end, raw = raw, drop = drop)
  )
}


#' @rdname chop_spikes
#' @export
#' @order 3
tab_spikes <- function (
                x,
                breaks,
                ...,
                n = NULL,
                prop = NULL
              ) {
  default_table(
    chop_spikes(x = x, breaks = breaks, ..., n = n, prop = prop)
  )
}

#' @rdname dissect
#' @export
#' @order 2
tab_dissect <- function (
                   x,
                   breaks,
                   ...,
                   n = NULL,
                   prop = NULL
               ) {
  default_table(
    dissect(x = x, breaks = breaks, ..., n = n, prop = prop)
  )
}


default_table <- function (x) {
  table(x, useNA = "ifany", dnn = NULL)
}

---
title: "Introduction to santoku"
output: 
  rmarkdown::html_vignette:
    toc: true
vignette: >
  %\VignetteIndexEntry{Introduction to santoku}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r, include = FALSE}
set.seed(23479)

knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)

options(digits = 4)
```


## Introduction

Santoku is a package for cutting data into intervals. It provides `chop()`, 
a replacement for base R's `cut()` function, as well as several convenience
functions to cut different kinds of intervals.

To install santoku, run:

``` r
install.packages("santoku")
```

## Basic usage

Use `chop()` like `cut()`, to cut numeric data into intervals between a set of
`breaks`.

```{r}
library(santoku)

x <- runif(10, 0, 10)
(chopped <- chop(x, breaks = 0:10))
data.frame(x, chopped)
```

`chop()` returns a factor.

If data is beyond the limits of `breaks`, they will be extended automatically:

```{r}
chopped <- chop(x, breaks = 3:7)
data.frame(x, chopped)
```

To chop a single number into a separate category, put the number twice in
`breaks`:

```{r}
x_fives <- x
x_fives[1:5] <- 5
chopped <- chop(x_fives, c(2, 5, 5, 8))
data.frame(x_fives, chopped)
```


To quickly produce a table of chopped data, use `tab()`:

```{r}
tab(1:10, c(2, 5, 8))
```

## Chopping by width and number of elements

To chop into fixed-width intervals, starting at the minimum value, use
`chop_width()`:

```{r}
chopped <- chop_width(x, 2)
data.frame(x, chopped)
```

To chop into a fixed number of intervals, each with the same width, 
use `chop_evenly()`:

```{r}
chopped <- chop_evenly(x, intervals = 3)
data.frame(x, chopped)
```

To chop into groups with a fixed number of elements, use `chop_n()`:

```{r}
chopped <- chop_n(x, 4)
table(chopped)
```


To chop into a fixed number of groups, each with the same number of elements,
use `chop_equally()`:

```{r}
chopped <- chop_equally(x, groups = 5)
table(chopped)
```

To chop data up by quantiles, use `chop_quantiles()`:

```{r}
chopped <- chop_quantiles(x, c(0.25, 0.5, 0.75))
data.frame(x, chopped)
```

To chop data up by proportions of the data range, use `chop_proportions()`:

```{r}
chopped <- chop_proportions(x, c(0.25, 0.5, 0.75))
data.frame(x, chopped)
```

You can think of these six functions as logically arranged in a table.

To chop into...                | Sizing intervals by...    |   
:------------------------------|:--------------------------|:----------------------
&nbsp;                         | number of elements:       | interval width:
a specific number of equal intervals... | `chop_equally()`   | `chop_evenly()`
intervals of one specific size...          | `chop_n()`         | `chop_width()`
intervals of different specific sizes...   | `chop_quantiles()` | `chop_proportions()`

: Different ways to chop by size

## Even more ways to chop

To chop data by standard deviations around the mean, use `chop_mean_sd()`:

```{r}
chopped <- chop_mean_sd(x)
data.frame(x, chopped)
```

To chop data into attractive intervals, use `chop_pretty()`. This 
selects intervals which are a multiple of 2, 5 or 10. It's useful for producing
bar plots.

```{r}
chopped <- chop_pretty(x)
data.frame(x, chopped)
```


## Isolating common values

In exploratory work, it's sometimes useful to find common values and
treat them differently. You can use `dissect()` to do this:

```{r}
x_spike <- rnorm(100)
x_spike[1:50] <- x_spike[1]

chopped <- dissect(x_spike, -3:3, prop = 0.1)
table(chopped)

```

`prop = 0.2` will put any unique value of `x` into its own separate
category if it makes up at least 20% of the data.

Note that unlike all the other `chop_*` functions, `dissect()`
doesn't always categorize `x` into ordered, connected intervals. 
To remind you of this, it is named differently. If you want to create
separate intervals on the left and right of common elements, use
`chop_spikes()`:

```{r}
chopped <- chop_spikes(x_spike, -3:3, prop = 0.1)
table(chopped)
```

Compare this to the table before. There are two intervals on either
side of the common value, instead of one interval surrounding it.


## Quick tables

`tab_n()`, `tab_width()`, and friends act similarly to
`tab()`, calling the related `chop_*` function and then `table()` on the result.

```{r}
tab_n(x, 4)
tab_width(x, 2)
tab_evenly(x, 5)
tab_mean_sd(x)
```


## Specifying labels

By default, santoku labels intervals using mathematical notation:

* `[0, 1]` means all numbers between 0 and 1 inclusive.
* `(0, 1)` means all numbers _strictly_ between 0 and 1, not including the
  endpoints.
* `[0, 1)` means all numbers between 0 and 1, including 0 but not 1.
* `(0, 1]` means all numbers between 0 and 1, including 1 but not 0.
* `{0}` means just the number 0.


To override these labels, provide names to the `breaks` argument:

```{r}
chopped <- chop(x, c(Lowest = 1, Low = 2, Higher = 5, Highest = 8))
data.frame(x, chopped)
```

Or, you can specify factor labels with the `labels` argument:

```{r}
chopped <- chop(x, c(2, 5, 8), labels = c("Lowest", "Low", "Higher", "Highest"))
data.frame(x, chopped)
```

You need as many labels as there are intervals - one fewer than `length(breaks)`
if your data doesn't extend beyond `breaks`, one more than `length(breaks)` if
it does.

To label intervals with a dash, use `lbl_dash()`:

```{r}
chopped <- chop(x, c(2, 5, 8), labels = lbl_dash())
data.frame(x, chopped)
```

To label integer data, use `lbl_discrete()`. It uses more informative right 
endpoints:

```{r}
chopped  <- chop(1:10, c(2, 5, 8), labels = lbl_discrete())
chopped2 <- chop(1:10, c(2, 5, 8), labels = lbl_dash())
data.frame(x = 1:10, lbl_discrete = chopped, lbl_dash = chopped2)
```

You can customize the first or last labels:

```{r}
chopped <- chop(x, c(2, 5, 8), labels = lbl_dash(first = "< 2", last = "8+"))
data.frame(x, chopped)
```


To label intervals in order use `lbl_seq()`:

```{r}
chopped <- chop(x, c(2, 5, 8), labels = lbl_seq())
data.frame(x, chopped)
```

You can use numerals or even roman numerals:

```{r}
chop(x, c(2, 5, 8), labels = lbl_seq("(1)"))
chop(x, c(2, 5, 8), labels = lbl_seq("i."))
```

Other labelling functions include:

* `lbl_endpoints()` - use left endpoints as labels
* `lbl_midpoints()` - use interval midpoints as labels
* `lbl_glue()` - specify labels flexibly with the `{glue}` package

## Specifying breaks

By default, `chop()` extends `breaks` if necessary. If you don't want
that, set `extend = FALSE`:

```{r}
chopped <- chop(x, c(3, 5, 7), extend = FALSE)
data.frame(x, chopped)
```

Data outside the range of `breaks` will become `NA`.

By default, intervals are closed on the left, i.e. they include their left
endpoints. If you want right-closed intervals, set `left = FALSE`:

```{r}
y <- 1:5
data.frame(
        y = y, 
        left_closed = chop(y, 1:5), 
        right_closed = chop(y, 1:5, left = FALSE)
      )
```

By default, the last interval is closed on both ends.
If you want to keep the last interval open at the end, 
set `close_end = FALSE`:

```{r}
data.frame(
  y = y,
  end_closed = chop(y, 1:5),
  end_open   = chop(y, 1:5, close_end = FALSE)
)

```


# Chopping dates, times and other vectors

You can chop many kinds of vectors with santoku, including Date objects...

```{r}
y2k <- as.Date("2000-01-01") + 0:10 * 7
data.frame(
  y2k = y2k,
  chopped = chop(y2k, as.Date(c("2000-02-01", "2000-03-01")))
)
```


... and POSIXct (date-time) objects:

```{r}
# hours of the 2020 Crew Dragon flight:
crew_dragon <- seq(as.POSIXct("2020-05-30 18:00", tz = "GMT"), 
                     length.out = 24, by = "hours")
liftoff <- as.POSIXct("2020-05-30 15:22", tz = "America/New_York")
dock    <- as.POSIXct("2020-05-31 10:16", tz = "America/New_York")

data.frame(
  crew_dragon = crew_dragon,
  chopped = chop(crew_dragon, c(liftoff, dock), 
                   labels = c("pre-flight", "flight", "docked"))
)

```

Note how santoku correctly handles the different timezones.

You can use `chop_width()` with objects from the `lubridate` package, 
to chop by irregular periods such as months:

```{r}
library(lubridate)
data.frame(
  y2k = y2k,
  chopped = chop_width(y2k, months(1))
)
```


`lbl_date()` produces nicely formatted dates:

```{r}
data.frame(
  y2k = y2k,
  chopped = chop_width(y2k, days(28), labels = lbl_date())
)
```

You can also chop vectors with units, using the `units` package:

```{r}
library(units)

x <- set_units(1:10 * 10, cm)
br <- set_units(1:3, ft)
data.frame(
  x = x,
  chopped = chop(x, br)
)
```

You should be able to chop anything that has a comparison operator. You can
even chop character data using lexical ordering. By default santoku emits a 
warning in this case, to avoid accidentally misinterpreting results:

```{r}
chop(letters[1:10], c("d", "f"))
```

If you find a type of data that you can't chop, please 
[file an issue](https://github.com/hughjonesd/santoku/issues).

---
title: "Introduction to santoku"
output: 
  rmarkdown::html_vignette:
    toc: true
vignette: >
  %\VignetteIndexEntry{Introduction to santoku}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r, include = FALSE}
set.seed(23479)

knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)

options(digits = 4)
```


## Introduction

Santoku is a package for cutting data into intervals. It provides `chop()`, 
a replacement for base R's `cut()` function, as well as several convenience
functions to cut different kinds of intervals.

To install santoku, run:

``` r
install.packages("santoku")
```

## Basic usage

Use `chop()` like `cut()`, to cut numeric data into intervals between a set of
`breaks`.

```{r}
library(santoku)

x <- runif(10, 0, 10)
(chopped <- chop(x, breaks = 0:10))
data.frame(x, chopped)
```

`chop()` returns a factor.

If data is beyond the limits of `breaks`, they will be extended automatically:

```{r}
chopped <- chop(x, breaks = 3:7)
data.frame(x, chopped)
```

To chop a single number into a separate category, put the number twice in
`breaks`:

```{r}
x_fives <- x
x_fives[1:5] <- 5
chopped <- chop(x_fives, c(2, 5, 5, 8))
data.frame(x_fives, chopped)
```


To quickly produce a table of chopped data, use `tab()`:

```{r}
tab(1:10, c(2, 5, 8))
```

## Chopping by width and number of elements

To chop into fixed-width intervals, starting at the minimum value, use
`chop_width()`:

```{r}
chopped <- chop_width(x, 2)
data.frame(x, chopped)
```

To chop into a fixed number of intervals, each with the same width, 
use `chop_evenly()`:

```{r}
chopped <- chop_evenly(x, intervals = 3)
data.frame(x, chopped)
```

To chop into groups with a fixed number of elements, use `chop_n()`:

```{r}
chopped <- chop_n(x, 4)
table(chopped)
```


To chop into a fixed number of groups, each with the same number of elements,
use `chop_equally()`:

```{r}
chopped <- chop_equally(x, groups = 5)
table(chopped)
```

To chop data up by quantiles, use `chop_quantiles()`:

```{r}
chopped <- chop_quantiles(x, c(0.25, 0.5, 0.75))
data.frame(x, chopped)
```

To chop data up by proportions of the data range, use `chop_proportions()`:

```{r}
chopped <- chop_proportions(x, c(0.25, 0.5, 0.75))
data.frame(x, chopped)
```

You can think of these six functions as logically arranged in a table.

To chop into...                | Sizing intervals by...    |   
:------------------------------|:--------------------------|:----------------------
&nbsp;                         | number of elements:       | interval width:
a specific number of equal intervals... | `chop_equally()`   | `chop_evenly()`
intervals of one specific size...          | `chop_n()`         | `chop_width()`
intervals of different specific sizes...   | `chop_quantiles()` | `chop_proportions()`

: Different ways to chop by size

## Even more ways to chop

To chop data by standard deviations around the mean, use `chop_mean_sd()`:

```{r}
chopped <- chop_mean_sd(x)
data.frame(x, chopped)
```

To chop data into attractive intervals, use `chop_pretty()`. This 
selects intervals which are a multiple of 2, 5 or 10. It's useful for producing
bar plots.

```{r}
chopped <- chop_pretty(x)
data.frame(x, chopped)
```


## Isolating common values

In exploratory work, it's sometimes useful to find common values and
treat them differently. You can use `dissect()` to do this:

```{r}
x_spike <- rnorm(100)
x_spike[1:50] <- x_spike[1]

chopped <- dissect(x_spike, -3:3, prop = 0.1)
table(chopped)

```

`prop = 0.2` will put any unique value of `x` into its own separate
category if it makes up at least 20% of the data.

Note that unlike all the other `chop_*` functions, `dissect()`
doesn't always categorize `x` into ordered, connected intervals. 
To remind you of this, it is named differently. If you want to create
separate intervals on the left and right of common elements, use
`chop_spikes()`:

```{r}
chopped <- chop_spikes(x_spike, -3:3, prop = 0.1)
table(chopped)
```

Compare this to the table before. There are two intervals on either
side of the common value, instead of one interval surrounding it.


## Quick tables

`tab_n()`, `tab_width()`, and friends act similarly to
`tab()`, calling the related `chop_*` function and then `table()` on the result.

```{r}
tab_n(x, 4)
tab_width(x, 2)
tab_evenly(x, 5)
tab_mean_sd(x)
```


## Specifying labels

By default, santoku labels intervals using mathematical notation:

* `[0, 1]` means all numbers between 0 and 1 inclusive.
* `(0, 1)` means all numbers _strictly_ between 0 and 1, not including the
  endpoints.
* `[0, 1)` means all numbers between 0 and 1, including 0 but not 1.
* `(0, 1]` means all numbers between 0 and 1, including 1 but not 0.
* `{0}` means just the number 0.


To override these labels, provide names to the `breaks` argument:

```{r}
chopped <- chop(x, c(Lowest = 1, Low = 2, Higher = 5, Highest = 8))
data.frame(x, chopped)
```

Or, you can specify factor labels with the `labels` argument:

```{r}
chopped <- chop(x, c(2, 5, 8), labels = c("Lowest", "Low", "Higher", "Highest"))
data.frame(x, chopped)
```

You need as many labels as there are intervals - one fewer than `length(breaks)`
if your data doesn't extend beyond `breaks`, one more than `length(breaks)` if
it does.

To label intervals with a dash, use `lbl_dash()`:

```{r}
chopped <- chop(x, c(2, 5, 8), labels = lbl_dash())
data.frame(x, chopped)
```

To label integer data, use `lbl_discrete()`. It uses more informative right 
endpoints:

```{r}
chopped  <- chop(1:10, c(2, 5, 8), labels = lbl_discrete())
chopped2 <- chop(1:10, c(2, 5, 8), labels = lbl_dash())
data.frame(x = 1:10, lbl_discrete = chopped, lbl_dash = chopped2)
```

You can customize the first or last labels:

```{r}
chopped <- chop(x, c(2, 5, 8), labels = lbl_dash(first = "< 2", last = "8+"))
data.frame(x, chopped)
```


To label intervals in order use `lbl_seq()`:

```{r}
chopped <- chop(x, c(2, 5, 8), labels = lbl_seq())
data.frame(x, chopped)
```

You can use numerals or even roman numerals:

```{r}
chop(x, c(2, 5, 8), labels = lbl_seq("(1)"))
chop(x, c(2, 5, 8), labels = lbl_seq("i."))
```

Other labelling functions include:

* `lbl_endpoints()` - use left endpoints as labels
* `lbl_midpoints()` - use interval midpoints as labels
* `lbl_glue()` - specify labels flexibly with the `{glue}` package

## Specifying breaks

By default, `chop()` extends `breaks` if necessary. If you don't want
that, set `extend = FALSE`:

```{r}
chopped <- chop(x, c(3, 5, 7), extend = FALSE)
data.frame(x, chopped)
```

Data outside the range of `breaks` will become `NA`.

By default, intervals are closed on the left, i.e. they include their left
endpoints. If you want right-closed intervals, set `left = FALSE`:

```{r}
y <- 1:5
data.frame(
        y = y, 
        left_closed = chop(y, 1:5), 
        right_closed = chop(y, 1:5, left = FALSE)
      )
```

By default, the last interval is closed on both ends.
If you want to keep the last interval open at the end, 
set `close_end = FALSE`:

```{r}
data.frame(
  y = y,
  end_closed = chop(y, 1:5),
  end_open   = chop(y, 1:5, close_end = FALSE)
)

```


# Chopping dates, times and other vectors

You can chop many kinds of vectors with santoku, including Date objects...

```{r}
y2k <- as.Date("2000-01-01") + 0:10 * 7
data.frame(
  y2k = y2k,
  chopped = chop(y2k, as.Date(c("2000-02-01", "2000-03-01")))
)
```


... and POSIXct (date-time) objects:

```{r}
# hours of the 2020 Crew Dragon flight:
crew_dragon <- seq(as.POSIXct("2020-05-30 18:00", tz = "GMT"), 
                     length.out = 24, by = "hours")
liftoff <- as.POSIXct("2020-05-30 15:22", tz = "America/New_York")
dock    <- as.POSIXct("2020-05-31 10:16", tz = "America/New_York")

data.frame(
  crew_dragon = crew_dragon,
  chopped = chop(crew_dragon, c(liftoff, dock), 
                   labels = c("pre-flight", "flight", "docked"))
)

```

Note how santoku correctly handles the different timezones.

You can use `chop_width()` with objects from the `lubridate` package, 
to chop by irregular periods such as months:

```{r}
library(lubridate)
data.frame(
  y2k = y2k,
  chopped = chop_width(y2k, months(1))
)
```

You can format labels using format strings from `strptime()`.
`lbl_discrete()` is useful here:

```{r}
data.frame(
  y2k = y2k,
  chopped = chop_width(y2k, months(1), labels = lbl_discrete(fmt = "%e %b"))
)
```

You can also chop vectors with units, using the `units` package:

```{r}
library(units)

x <- set_units(1:10 * 10, cm)
br <- set_units(1:3, ft)
data.frame(
  x = x,
  chopped = chop(x, br)
)
```

You should be able to chop anything that has a comparison operator. You can
even chop character data using lexical ordering. By default santoku emits a 
warning in this case, to avoid accidentally misinterpreting results:

```{r}
chop(letters[1:10], c("d", "f"))
```

If you find a type of data that you can't chop, please 
[file an issue](https://github.com/hughjonesd/santoku/issues).

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{percent}
\alias{percent}
\title{Simple percentage formatter}
\usage{
percent(x)
}
\arguments{
\item{x}{Numeric values.}
}
\value{
\code{x} formatted as a percent.
}
\description{
\code{percent()} formats \code{x} as a percentage.
For a wider range of formatters, consider the \href{https://cran.r-project.org/package=scales}{\code{scales} package}.
}
\examples{
percent(0.5)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/chop-by-width.R, R/breaks-by-width.R, R/tab.R
\name{chop_evenly}
\alias{chop_evenly}
\alias{brk_evenly}
\alias{tab_evenly}
\title{Chop into equal-width intervals}
\usage{
chop_evenly(x, intervals, ...)

brk_evenly(intervals)

tab_evenly(x, intervals, ...)
}
\arguments{
\item{x}{A vector.}

\item{intervals}{Integer: number of intervals to create.}

\item{...}{Passed to \code{\link[=chop]{chop()}}.}
}
\value{
\verb{chop_*} functions return a \code{\link{factor}} of the same length as \code{x}.

\verb{brk_*} functions return a \code{\link{function}} to create \code{breaks}.

\verb{tab_*} functions return a contingency \code{\link[=table]{table()}}.
}
\description{
\code{chop_evenly()} chops \code{x} into \code{intervals} intervals of equal width.
}
\examples{
chop_evenly(0:10, 5)

}
\seealso{
Other chopping functions: 
\code{\link{chop}()},
\code{\link{chop_equally}()},
\code{\link{chop_fn}()},
\code{\link{chop_mean_sd}()},
\code{\link{chop_n}()},
\code{\link{chop_proportions}()},
\code{\link{chop_quantiles}()},
\code{\link{chop_spikes}()},
\code{\link{chop_width}()},
\code{\link{fillet}()}
}
\concept{chopping functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/labels-datetime.R
\name{lbl_date}
\alias{lbl_date}
\alias{lbl_datetime}
\title{Label dates and datetimes}
\usage{
lbl_date(
  fmt = "\%e \%b \%Y",
  symbol = "-",
  unit = as.difftime(1, units = "days"),
  single = "{l}",
  first = NULL,
  last = NULL
)

lbl_datetime(
  fmt = "\%H:\%M:\%S \%b \%e \%Y",
  symbol = "-",
  unit = NULL,
  single = "{l}",
  first = NULL,
  last = NULL
)
}
\arguments{
\item{fmt}{String, list or function. A format for break endpoints.}

\item{symbol}{String: separator to use for full ranges.}

\item{unit}{Optional interval unit for non-overlapping labels. If not \code{NULL},
.  endpoints are adjusted in the style of \code{\link[=lbl_discrete]{lbl_discrete()}}.}

\item{single}{Glue string: label for singleton intervals. See \code{\link[=lbl_glue]{lbl_glue()}}
for details. If \code{NULL}, singleton intervals will be labelled the same way
as other intervals.}

\item{first}{Glue string: override label for the first category. Write e.g.
\code{first = "<{r}"} to create a label like \code{"<18"}. See \code{\link[=lbl_glue]{lbl_glue()}}
for details.}

\item{last}{String: override label for the last category. Write e.g.
\code{last = ">{l}"} to create a label like \code{">65"}. See \code{\link[=lbl_glue]{lbl_glue()}}
for details.}
}
\value{
A function that creates a vector of labels.
}
\description{
\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#experimental}{\figure{lifecycle-experimental.svg}{options: alt='[Experimental]'}}}{\strong{[Experimental]}}

\code{lbl_date()} and \code{lbl_datetime()} produce nice labels for dates
and datetimes. Where possible ranges are simplified, like
like "13-14 Jul 2026" or "11:15-12:15 1 Dec 2025".
}
\section{Formatting endpoints}{


If \code{fmt} is not \code{NULL} then it is used to format the endpoints.
\itemize{
\item If \code{fmt} is a string, then numeric endpoints will be formatted by
\code{sprintf(fmt, breaks)}; other endpoints, e.g. \link[base:Dates]{Date} objects,
will be formatted by \code{format(breaks, fmt)}.
\item If \code{fmt} is a list, then it will be used as arguments to \link{format}.
\item If \code{fmt} is a function, it should take a vector of numbers (or other objects
that can be used as breaks) and return a character vector. It may be helpful
to use functions from the \code{{scales}} package, e.g. \code{\link[scales:label_number]{scales::label_comma()}}.
}
}

\examples{
winter <- as.Date("2025-12-01") + 0:89
tab(winter, as.Date(c("2025-12-25", "2026-01-06")),
    labels = lbl_date())
new_year <- as.POSIXct("2025-12-31 23:00") + 0:120 * 60
round_midnight <- as.POSIXct(c("2025-12-31 23:59", "2026-01-01 00:05"))
tab(new_year, round_midnight,
    labels = lbl_datetime())
tab(new_year, round_midnight,
    labels = lbl_datetime(unit = as.difftime(1, units = "mins")))
}
\seealso{
Other labelling functions: 
\code{\link{lbl_dash}()},
\code{\link{lbl_discrete}()},
\code{\link{lbl_endpoints}()},
\code{\link{lbl_glue}()},
\code{\link{lbl_intervals}()},
\code{\link{lbl_manual}()},
\code{\link{lbl_midpoints}()},
\code{\link{lbl_seq}()}
}
\concept{labelling functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/chop-isolates.R, R/breaks.R, R/tab.R
\name{chop_spikes}
\alias{chop_spikes}
\alias{brk_spikes}
\alias{tab_spikes}
\title{Chop common values into singleton intervals}
\usage{
chop_spikes(x, breaks, ..., n = NULL, prop = NULL)

brk_spikes(breaks, n = NULL, prop = NULL)

tab_spikes(x, breaks, ..., n = NULL, prop = NULL)
}
\arguments{
\item{x}{A vector.}

\item{breaks}{A numeric vector of cut-points or a call to a \verb{brk_*} function.
The resulting \code{\link[=breaks-class]{breaks}} object will be modified to add
singleton breaks.}

\item{...}{Passed to \code{\link[=chop]{chop()}}.}

\item{n, prop}{Scalar. Provide either \code{n}, a number of values, or \code{prop},
a proportion of \code{length(x)}. Values of \code{x} which occur at least this
often will get their own singleton break.}
}
\value{
\verb{chop_*} functions return a \code{\link{factor}} of the same length as \code{x}.

\verb{brk_*} functions return a \code{\link{function}} to create \code{breaks}.

\verb{tab_*} functions return a contingency \code{\link[=table]{table()}}.
}
\description{
\code{chop_spikes()} lets you chop common values of \code{x} into their own
singleton intervals. This can help make unusual values visible.
}
\details{
This function is \ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#experimental}{\figure{lifecycle-experimental.svg}{options: alt='[Experimental]'}}}{\strong{[Experimental]}}.
}
\examples{
x <- c(1:4, rep(5, 5), 6:10)
chop_spikes(x, c(2, 7), n = 5)
chop_spikes(x, c(2, 7), prop = 0.25)
chop_spikes(x, brk_width(5), n = 5)

set.seed(42)
x <- runif(40, 0, 10)
x <- sample(x, 200, replace = TRUE)
tab_spikes(x, brk_width(2, 0), prop = 0.05)
}
\seealso{
\code{\link[=dissect]{dissect()}} for a different approach.

Other chopping functions: 
\code{\link{chop}()},
\code{\link{chop_equally}()},
\code{\link{chop_evenly}()},
\code{\link{chop_fn}()},
\code{\link{chop_mean_sd}()},
\code{\link{chop_n}()},
\code{\link{chop_proportions}()},
\code{\link{chop_quantiles}()},
\code{\link{chop_width}()},
\code{\link{fillet}()}
}
\concept{chopping functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/breaks-misc.R
\name{brk_manual}
\alias{brk_manual}
\title{Create a \code{breaks} object manually}
\usage{
brk_manual(breaks, left_vec)
}
\arguments{
\item{breaks}{A vector, which must be sorted.}

\item{left_vec}{A logical vector, the same length as \code{breaks}.
Specifies whether each break is left-closed or right-closed.}
}
\value{
A function which returns an object of class \code{breaks}.
}
\description{
Create a \code{breaks} object manually
}
\details{
All breaks must be closed on exactly one side, like \verb{..., x) [x, ...}
(left-closed) or \verb{..., x) [x, ...} (right-closed).

For example, if \code{breaks = 1:3} and \code{left = c(TRUE, FALSE, TRUE)}, then the
resulting intervals are \preformatted{
T        F       T
[ 1,  2 ] ( 2, 3 )
}

Singleton breaks are created by repeating a number in \code{breaks}. Singletons
must be closed on both sides, so if there is a repeated number
at indices \code{i}, \code{i+1}, \code{left[i]} \emph{must} be \code{TRUE} and \code{left[i+1]} must be
\code{FALSE}.

\code{brk_manual()} ignores \code{left} and \code{close_end} arguments passed in
from \code{\link[=chop]{chop()}}, since \code{left_vec} sets these manually.
\code{extend} and \code{drop} arguments are respected as usual.
}
\examples{
lbrks <- brk_manual(1:3, rep(TRUE, 3))
chop(1:3, lbrks, extend = FALSE)

rbrks <- brk_manual(1:3, rep(FALSE, 3))
chop(1:3, rbrks, extend = FALSE)

brks_singleton <- brk_manual(
      c(1,    2,    2,     3),
      c(TRUE, TRUE, FALSE, TRUE))

chop(1:3, brks_singleton, extend = FALSE)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/labels-single.R
\name{lbl_seq}
\alias{lbl_seq}
\title{Label chopped intervals in sequence}
\usage{
lbl_seq(start = "a")
}
\arguments{
\item{start}{String. A template for the sequence. See below.}
}
\value{
A function that creates a vector of labels.
}
\description{
\code{lbl_seq()} labels intervals sequentially, using numbers or letters.
}
\details{
\code{start} shows the first element of the sequence. It must contain exactly \emph{one}
character out of the set "a", "A", "i", "I" or "1". For later elements:
\itemize{
\item "a" will be replaced by "a", "b", "c", ...
\item "A" will be replaced by "A", "B", "C", ...
\item "i" will be replaced by lower-case Roman numerals "i", "ii", "iii", ...
\item "I" will be replaced by upper-case Roman numerals "I", "II", "III", ...
\item "1" will be replaced by numbers "1", "2", "3", ...
}

Other characters will be retained as-is.
}
\examples{
chop(1:10, c(2, 5, 8), lbl_seq())

chop(1:10, c(2, 5, 8), lbl_seq("i."))

chop(1:10, c(2, 5, 8), lbl_seq("(A)"))
}
\seealso{
Other labelling functions: 
\code{\link{lbl_dash}()},
\code{\link{lbl_discrete}()},
\code{\link{lbl_endpoints}()},
\code{\link{lbl_glue}()},
\code{\link{lbl_intervals}()},
\code{\link{lbl_manual}()},
\code{\link{lbl_midpoints}()}
}
\concept{labelling functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/labels-glue.R
\name{lbl_glue}
\alias{lbl_glue}
\title{Label chopped intervals using the \code{glue} package}
\usage{
lbl_glue(
  label,
  fmt = NULL,
  single = NULL,
  first = NULL,
  last = NULL,
  raw = FALSE,
  ...
)
}
\arguments{
\item{label}{A glue string passed to \code{\link[glue:glue]{glue::glue()}}.}

\item{fmt}{String, list or function. A format for break endpoints.}

\item{single}{Glue string: label for singleton intervals. See \code{\link[=lbl_glue]{lbl_glue()}}
for details. If \code{NULL}, singleton intervals will be labelled the same way
as other intervals.}

\item{first}{Glue string: override label for the first category. Write e.g.
\code{first = "<{r}"} to create a label like \code{"<18"}. See \code{\link[=lbl_glue]{lbl_glue()}}
for details.}

\item{last}{String: override label for the last category. Write e.g.
\code{last = ">{l}"} to create a label like \code{">65"}. See \code{\link[=lbl_glue]{lbl_glue()}}
for details.}

\item{raw}{\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#deprecated}{\figure{lifecycle-deprecated.svg}{options: alt='[Deprecated]'}}}{\strong{[Deprecated]}}. Use the \code{raw} argument to \code{\link[=chop]{chop()}}
instead.}

\item{...}{Further arguments passed to \code{\link[glue:glue]{glue::glue()}}.}
}
\value{
A function that creates a vector of labels.
}
\description{
Use \code{"{l}"} and \code{"{r}"} to show the left and right endpoints of the intervals.
}
\details{
The following variables are available in the glue string:
\itemize{
\item \code{l} is a character vector of left endpoints of intervals.
\item \code{r} is a character vector of right endpoints of intervals.
\item \code{l_closed} is a logical vector. Elements are \code{TRUE} when the left
endpoint is closed.
\item \code{r_closed} is a logical vector, \code{TRUE} when the right endpoint is closed.
}

Endpoints will be formatted by \code{fmt} before being passed to \code{glue()}.
}
\section{Formatting endpoints}{


If \code{fmt} is not \code{NULL} then it is used to format the endpoints.
\itemize{
\item If \code{fmt} is a string, then numeric endpoints will be formatted by
\code{sprintf(fmt, breaks)}; other endpoints, e.g. \link{Date} objects, will be
formatted by \code{format(breaks, fmt)}.
\item If \code{fmt} is a list, then it will be used as arguments to \link{format}.
\item If \code{fmt} is a function, it should take a vector of numbers (or other objects
that can be used as breaks) and return a character vector. It may be helpful
to use functions from the \code{{scales}} package, e.g. \code{\link[scales:label_number]{scales::label_comma()}}.
}
}

\examples{
tab(1:10, c(1, 3, 3, 7),
    labels = lbl_glue("{l} to {r}", single = "Exactly {l}"))

tab(1:10 * 1000, c(1, 3, 5, 7) * 1000,
    labels = lbl_glue("{l}-{r}",
                      fmt = function(x) prettyNum(x, big.mark=',')))

# reproducing lbl_intervals():
interval_left <- "{ifelse(l_closed, '[', '(')}"
interval_right <- "{ifelse(r_closed, ']', ')')}"
glue_string <- paste0(interval_left, "{l}", ", ", "{r}", interval_right)
tab(1:10, c(1, 3, 3, 7), labels = lbl_glue(glue_string, single = "{{{l}}}"))

}
\seealso{
Other labelling functions: 
\code{\link{lbl_dash}()},
\code{\link{lbl_discrete}()},
\code{\link{lbl_endpoints}()},
\code{\link{lbl_intervals}()},
\code{\link{lbl_manual}()},
\code{\link{lbl_midpoints}()},
\code{\link{lbl_seq}()}
}
\concept{labelling functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/santoku-cast.R
\name{santoku-cast}
\alias{santoku-cast}
\alias{santoku_cast_common.default}
\alias{santoku_cast_common.double}
\alias{santoku_cast_common.Date}
\alias{santoku_cast_common.POSIXct}
\alias{santoku_cast_common.ts}
\alias{santoku_cast_common.zoo}
\alias{santoku_cast_common.integer64}
\alias{santoku_cast_common.hexmode}
\alias{santoku_cast_common.octmode}
\title{Internal functions}
\usage{
\method{santoku_cast_common}{default}(x, y)

\method{santoku_cast_common}{double}(x, y)

\method{santoku_cast_common}{Date}(x, y)

\method{santoku_cast_common}{POSIXct}(x, y)

\method{santoku_cast_common}{ts}(x, y)

\method{santoku_cast_common}{zoo}(x, y)

\method{santoku_cast_common}{integer64}(x, y)

\method{santoku_cast_common}{hexmode}(x, y)

\method{santoku_cast_common}{octmode}(x, y)
}
\arguments{
\item{x, y}{Vectors to cast.}
}
\value{
A list.
}
\description{
Internal functions
}
\keyword{Do}
\keyword{These}
\keyword{are}
\keyword{functions.}
\keyword{internal}
\keyword{not}
\keyword{use.}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/chop-isolates.R, R/tab.R
\name{dissect}
\alias{dissect}
\alias{tab_dissect}
\title{Cut data into intervals, separating out common values}
\usage{
dissect(
  x,
  breaks,
  ...,
  n = NULL,
  prop = NULL,
  spike_labels = "{{{l}}}",
  exclude_spikes = FALSE
)

tab_dissect(x, breaks, ..., n = NULL, prop = NULL)
}
\arguments{
\item{x, breaks, ...}{Passed to \code{\link[=chop]{chop()}}.}

\item{n, prop}{Scalar. Provide either \code{n}, a number of values, or \code{prop},
a proportion of \code{length(x)}. Values of \code{x} which occur at least this
often will get their own singleton break.}

\item{spike_labels}{\link[glue:glue]{Glue} string for spike labels. Use \code{"{l}"}
for the spike value.}

\item{exclude_spikes}{Logical. Exclude spikes before chopping \code{x}? This
can affect the location of data-dependent breaks.}
}
\value{
\code{dissect()} returns the result of \code{\link[=chop]{chop()}}, but with common values put into
separate factor levels.

\code{tab_dissect()} returns a contingency \code{\link[=table]{table()}}.
}
\description{
Sometimes it's useful to separate out common elements of \code{x}.
\code{dissect()} chops \code{x}, but puts common elements of \code{x} ("spikes")
into separate categories.
}
\details{
Unlike \code{\link[=chop_spikes]{chop_spikes()}}, \code{dissect()} doesn't break up
intervals which contain a spike. As a result, unlike \verb{chop_*} functions,
\code{dissect()} does not chop \code{x} into disjoint intervals. See the examples.

If breaks are data-dependent, their labels may be misleading after common
elements have been removed. See the example below. To get round this,
set \code{exclude_spikes} to \code{TRUE}. Then breaks will be calculated after
removing spikes from the data.

Levels of the result are ordered by the minimum element in each level. As
a result, if \code{drop = FALSE}, empty levels will be placed last.

This function is \ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#experimental}{\figure{lifecycle-experimental.svg}{options: alt='[Experimental]'}}}{\strong{[Experimental]}}.
}
\examples{
x <- c(2, 3, 3, 3, 4)
dissect(x, c(2, 4), n = 3)
dissect(x, brk_width(2), prop = 0.5)

set.seed(42)
x <- runif(40, 0, 10)
x <- sample(x, 200, replace = TRUE)
# Compare:
table(dissect(x, brk_width(2, 0), prop = 0.05))
# Versus:
tab_spikes(x, brk_width(2, 0), prop = 0.05)

# Potentially confusing data-dependent breaks:
set.seed(42)
x <- rnorm(99)
x[1:9] <- x[1]
tab_quantiles(x, 1:2/3)
tab_dissect(x, brk_quantiles(1:2/3), n = 9)
# Calculate quantiles excluding spikes:
tab_dissect(x, brk_quantiles(1:2/3), n = 9, exclude_spikes = TRUE)
}
\seealso{
\code{\link[=chop_spikes]{chop_spikes()}} for a different approach.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/chop-by-width.R, R/breaks-by-width.R, R/tab.R
\name{chop_proportions}
\alias{chop_proportions}
\alias{brk_proportions}
\alias{tab_proportions}
\title{Chop into proportions of the range of x}
\usage{
chop_proportions(x, proportions, ..., raw = TRUE)

brk_proportions(proportions)

tab_proportions(x, proportions, ..., raw = TRUE)
}
\arguments{
\item{x}{A vector.}

\item{proportions}{Numeric vector between 0 and 1: proportions of x's range.
If \code{proportions} has names, these will be used for labels.}

\item{...}{Passed to \code{\link[=chop]{chop()}}.}

\item{raw}{Logical. Use raw values in labels?}
}
\value{
\verb{chop_*} functions return a \code{\link{factor}} of the same length as \code{x}.

\verb{brk_*} functions return a \code{\link{function}} to create \code{breaks}.

\verb{tab_*} functions return a contingency \code{\link[=table]{table()}}.
}
\description{
\code{chop_proportions()} chops \code{x} into \code{proportions} of its range, excluding
infinite values.
}
\details{
By default, labels show the raw numeric endpoints. To label intervals by
the proportions, use \code{raw = FALSE}.
}
\examples{
chop_proportions(0:10, c(0.2, 0.8))
chop_proportions(0:10, c(Low = 0, Mid = 0.2, High = 0.8))

}
\seealso{
Other chopping functions: 
\code{\link{chop}()},
\code{\link{chop_equally}()},
\code{\link{chop_evenly}()},
\code{\link{chop_fn}()},
\code{\link{chop_mean_sd}()},
\code{\link{chop_n}()},
\code{\link{chop_quantiles}()},
\code{\link{chop_spikes}()},
\code{\link{chop_width}()},
\code{\link{fillet}()}
}
\concept{chopping functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/non-standard-types-doc.R
\name{non-standard-types}
\alias{non-standard-types}
\title{Tips for chopping non-standard types}
\description{
Santoku can handle many non-standard types.
}
\details{
\itemize{
\item If objects can be compared using \code{<}, \code{==} etc. then they should
be choppable.
\item Objects which can't be converted to numeric are handled within R code,
which may be slower.
\item Character \code{x} and \code{breaks} are chopped with a warning.
\item If \code{x} and \code{breaks} are not the same type, they should be able to
be cast to the same type, usually using \code{\link[vctrs:vec_cast]{vctrs::vec_cast_common()}}.
\item Not all chopping operations make sense, for example, \code{\link[=chop_mean_sd]{chop_mean_sd()}}
on a character vector.
\item For indexed objects such as \code{\link[stats:ts]{stats::ts()}} objects, indices will be dropped
from the result.
\item If you get errors, try setting \code{extend = FALSE} (but also file a bug report).
\item To request support for a type, open an issue on Github.
}
}
\seealso{
brk-width-for-Datetime
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/chop-by-group-size.R, R/breaks-by-group-size.R,
%   R/tab.R
\name{chop_quantiles}
\alias{chop_quantiles}
\alias{chop_deciles}
\alias{brk_quantiles}
\alias{tab_quantiles}
\alias{tab_deciles}
\title{Chop by quantiles}
\usage{
chop_quantiles(
  x,
  probs,
  ...,
  labels = if (raw) lbl_intervals() else lbl_intervals(single = NULL),
  left = is.numeric(x),
  raw = FALSE,
  weights = NULL,
  recalc_probs = FALSE
)

chop_deciles(x, ...)

brk_quantiles(probs, ..., weights = NULL, recalc_probs = FALSE)

tab_quantiles(x, probs, ..., left = is.numeric(x), raw = FALSE)

tab_deciles(x, ...)
}
\arguments{
\item{x}{A vector.}

\item{probs}{A vector of probabilities for the quantiles. If \code{probs} has
names, these will be used for labels.}

\item{...}{For \code{chop_quantiles}, passed to \code{\link[=chop]{chop()}}. For \code{brk_quantiles()},
passed to \code{\link[stats:quantile]{stats::quantile()}} or \code{\link[Hmisc:wtd.stats]{Hmisc::wtd.quantile()}}.}

\item{labels}{A character vector of labels or a function to create labels.}

\item{left}{Logical. Left-closed or right-closed breaks?}

\item{raw}{Logical. Use raw values in labels?}

\item{weights}{\code{NULL} or numeric vector of same length as \code{x}. If not
\code{NULL}, \code{\link[Hmisc:wtd.stats]{Hmisc::wtd.quantile()}} is used to calculate weighted quantiles.}

\item{recalc_probs}{Logical. Recalculate probabilities of quantiles using
\code{\link[stats:ecdf]{ecdf(x)}}? See below.}
}
\value{
\verb{chop_*} functions return a \code{\link{factor}} of the same length as \code{x}.

\verb{brk_*} functions return a \code{\link{function}} to create \code{breaks}.

\verb{tab_*} functions return a contingency \code{\link[=table]{table()}}.
}
\description{
\code{chop_quantiles()} chops data by quantiles.
\code{chop_deciles()} is a convenience function which chops into deciles.
}
\details{
For non-numeric \code{x}, \code{left} is set to \code{FALSE} by default. This works better
for calculating "type 1" quantiles, since they round down. See
\code{\link[stats:quantile]{stats::quantile()}}.

By default, \code{chop_quantiles()} shows the requested probabilities in the
labels. To show the numeric quantiles themselves, set \code{raw = TRUE}.

When \code{x} contains duplicates, consecutive quantiles may be the same number. If
so, interval labels may be misleading, and if \code{recalc_probs = FALSE} a warning is
emitted. Set \code{recalc_probs = TRUE} to recalculate the probabilities of the quantiles
using the \link[stats:ecdf]{empirical cumulative distribution function} of \code{x}.
Doing so may give you different labels from what you expect, and will
remove any names from \code{probs}, but it never changes the actual
quantiles used for breaks. At present, \code{recalc_probs = TRUE} is incompatible
with non-null \code{weights}. See the example below.
}
\examples{
chop_quantiles(1:10, 1:3/4)

chop_quantiles(1:10, c(Q1 = 0, Q2 = 0.25, Q3 = 0.5, Q4 = 0.75))

chop(1:10, brk_quantiles(1:3/4))

chop_deciles(1:10)

# to label by the quantiles themselves:
chop_quantiles(1:10, 1:3/4, raw = TRUE)

# duplicate quantiles:
x <- c(1, 1, 1, 2, 3)
quantile(x, 1:5/5)
tab_quantiles(x, 1:5/5)
tab_quantiles(x, 1:5/5, recalc_probs = TRUE)
set.seed(42)
tab_quantiles(rnorm(100), probs = 1:3/4, raw = TRUE)

}
\seealso{
Other chopping functions: 
\code{\link{chop}()},
\code{\link{chop_equally}()},
\code{\link{chop_evenly}()},
\code{\link{chop_fn}()},
\code{\link{chop_mean_sd}()},
\code{\link{chop_n}()},
\code{\link{chop_proportions}()},
\code{\link{chop_spikes}()},
\code{\link{chop_width}()},
\code{\link{fillet}()}
}
\concept{chopping functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/chop-by-group-size.R, R/breaks-by-group-size.R,
%   R/tab.R
\name{chop_n}
\alias{chop_n}
\alias{brk_n}
\alias{tab_n}
\title{Chop into fixed-sized groups}
\usage{
chop_n(x, n, ..., tail = "split")

brk_n(n, tail = "split")

tab_n(x, n, ..., tail = "split")
}
\arguments{
\item{x}{A vector.}

\item{n}{Integer. Number of elements in each interval.}

\item{...}{Passed to \code{\link[=chop]{chop()}}.}

\item{tail}{String. What to do if the final interval has fewer than \code{n} elements?
\code{"split"} to keep it separate. \code{"merge"} to merge it with the neighbouring
interval.}
}
\value{
\verb{chop_*} functions return a \code{\link{factor}} of the same length as \code{x}.

\verb{brk_*} functions return a \code{\link{function}} to create \code{breaks}.

\verb{tab_*} functions return a contingency \code{\link[=table]{table()}}.
}
\description{
\code{chop_n()} creates intervals containing a fixed number of elements.
}
\details{
The algorithm guarantees that intervals contain no more than \code{n} elements, so
long as there are no duplicates in \code{x} and \code{tail = "split"}. It also
guarantees that intervals contain no fewer than \code{n} elements, except possibly
the last interval (or first interval if \code{left} is \code{FALSE}).

To ensure that all intervals contain at least \code{n} elements (so long as there
are at least \code{n} elements in \code{x}!) set \code{tail = "merge"}.

If \code{tail = "split"} and there are intervals containing duplicates with more
than \code{n} elements, a warning is given.
}
\examples{
chop_n(1:10, 5)

chop_n(1:5, 2)
chop_n(1:5, 2, tail = "merge")

# too many duplicates
x <- rep(1:2, each = 3)
chop_n(x, 2)

tab_n(1:10, 5)

# fewer elements in one group
tab_n(1:10, 4)

}
\seealso{
Other chopping functions: 
\code{\link{chop}()},
\code{\link{chop_equally}()},
\code{\link{chop_evenly}()},
\code{\link{chop_fn}()},
\code{\link{chop_mean_sd}()},
\code{\link{chop_proportions}()},
\code{\link{chop_quantiles}()},
\code{\link{chop_spikes}()},
\code{\link{chop_width}()},
\code{\link{fillet}()}
}
\concept{chopping functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{exactly}
\alias{exactly}
\title{Define singleton intervals explicitly}
\usage{
exactly(x)
}
\arguments{
\item{x}{A numeric vector.}
}
\value{
The same as \code{rep(x, each = 2)}.
}
\description{
\code{exactly()} duplicates its input.
It lets you define singleton intervals like this: \code{chop(x, c(1, exactly(2), 3))}.
This is the same as \code{chop(x, c(1, 2, 2, 3))} but conveys your intent more
clearly.
}
\examples{
chop(1:10, c(2, exactly(5), 8))

# same:
chop(1:10, c(2, 5, 5, 8))
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/chop.R
\name{fillet}
\alias{fillet}
\title{Chop data precisely (for programmers)}
\usage{
fillet(
  x,
  breaks,
  labels = lbl_intervals(),
  left = TRUE,
  close_end = TRUE,
  raw = NULL
)
}
\arguments{
\item{x}{A vector.}

\item{breaks}{A numeric vector of cut-points, or a function to create
cut-points from \code{x}.}

\item{labels}{A character vector of labels or a function to create labels.}

\item{left}{Logical. Left-closed or right-closed breaks?}

\item{close_end}{Logical. Close last break at right? (If \code{left} is \code{FALSE},
close first break at left?)}

\item{raw}{Logical. Use raw values in labels?}
}
\value{
\code{fillet()} returns a \code{\link{factor}} of the same length as \code{x}, representing
the intervals containing the value of \code{x}.
}
\description{
\code{fillet()} calls \code{\link[=chop]{chop()}} with \code{extend = FALSE} and \code{drop = FALSE}. This
ensures that you get only the \code{breaks} and \code{labels} you ask for. When
programming, consider using \code{fillet()} instead of \code{chop()}.
}
\examples{
fillet(1:10, c(2, 5, 8))
}
\seealso{
Other chopping functions: 
\code{\link{chop}()},
\code{\link{chop_equally}()},
\code{\link{chop_evenly}()},
\code{\link{chop_fn}()},
\code{\link{chop_mean_sd}()},
\code{\link{chop_n}()},
\code{\link{chop_proportions}()},
\code{\link{chop_quantiles}()},
\code{\link{chop_spikes}()},
\code{\link{chop_width}()}
}
\concept{chopping functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/labels-single.R
\name{lbl_manual}
\alias{lbl_manual}
\title{Defunct: label chopped intervals in a user-defined sequence}
\usage{
lbl_manual(sequence, fmt = "\%s")
}
\arguments{
\item{sequence}{A character vector of labels.}

\item{fmt}{String, list or function. A format for break endpoints.}
}
\value{
A function that creates a vector of labels.
}
\description{
\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#defunct}{\figure{lifecycle-defunct.svg}{options: alt='[Defunct]'}}}{\strong{[Defunct]}}
}
\details{
\code{lbl_manual()} is defunct since santoku 1.0.0. It is little used and is not
closely related to the rest of the package. It also risks mislabelling
intervals, e.g. if intervals are extended. Use of \code{lbl_manual()} will give
an error.
}
\section{Formatting endpoints}{


If \code{fmt} is not \code{NULL} then it is used to format the endpoints.
\itemize{
\item If \code{fmt} is a string, then numeric endpoints will be formatted by
\code{sprintf(fmt, breaks)}; other endpoints, e.g. \link{Date} objects, will be
formatted by \code{format(breaks, fmt)}.
\item If \code{fmt} is a list, then it will be used as arguments to \link{format}.
\item If \code{fmt} is a function, it should take a vector of numbers (or other objects
that can be used as breaks) and return a character vector. It may be helpful
to use functions from the \code{{scales}} package, e.g. \code{\link[scales:label_number]{scales::label_comma()}}.
}
}

\examples{
\dontrun{
chop(1:10, c(2, 5, 8), lbl_manual(c("w", "x", "y", "z")))
# ->
chop(1:10, c(2, 5, 8), labels = c("w", "x", "y", "z"))
}
}
\seealso{
Other labelling functions: 
\code{\link{lbl_dash}()},
\code{\link{lbl_discrete}()},
\code{\link{lbl_endpoints}()},
\code{\link{lbl_glue}()},
\code{\link{lbl_intervals}()},
\code{\link{lbl_midpoints}()},
\code{\link{lbl_seq}()}
}
\concept{labelling functions}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/chop-misc.R, R/breaks-misc.R, R/tab.R
\name{chop_mean_sd}
\alias{chop_mean_sd}
\alias{brk_mean_sd}
\alias{tab_mean_sd}
\title{Chop by standard deviations}
\usage{
chop_mean_sd(x, sds = 1:3, ..., raw = FALSE, sd = deprecated())

brk_mean_sd(sds = 1:3, sd = deprecated())

tab_mean_sd(x, sds = 1:3, ..., raw = FALSE)
}
\arguments{
\item{x}{A vector.}

\item{sds}{Positive numeric vector of standard deviations.}

\item{...}{Passed to \code{\link[=chop]{chop()}}.}

\item{raw}{Logical. Use raw values in labels?}

\item{sd}{\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#deprecated}{\figure{lifecycle-deprecated.svg}{options: alt='[Deprecated]'}}}{\strong{[Deprecated]}}}
}
\value{
\verb{chop_*} functions return a \code{\link{factor}} of the same length as \code{x}.

\verb{brk_*} functions return a \code{\link{function}} to create \code{breaks}.

\verb{tab_*} functions return a contingency \code{\link[=table]{table()}}.
}
\description{
Intervals are measured in standard deviations on either side of the
mean.
}
\details{
In version 0.7.0, these functions changed to specifying \code{sds} as a vector.
To chop 1, 2 and 3 standard deviations around the mean, write
\code{chop_mean_sd(x, sds = 1:3)} instead of \code{chop_mean_sd(x, sd = 3)}.
}
\examples{
chop_mean_sd(1:10)

chop(1:10, brk_mean_sd())

tab_mean_sd(1:10)

}
\seealso{
Other chopping functions: 
\code{\link{chop}()},
\code{\link{chop_equally}()},
\code{\link{chop_evenly}()},
\code{\link{chop_fn}()},
\code{\link{chop_n}()},
\code{\link{chop_proportions}()},
\code{\link{chop_quantiles}()},
\code{\link{chop_spikes}()},
\code{\link{chop_width}()},
\code{\link{fillet}()}
}
\concept{chopping functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/labels-single.R
\name{lbl_midpoints}
\alias{lbl_midpoints}
\title{Label chopped intervals by their midpoints}
\usage{
lbl_midpoints(
  fmt = NULL,
  single = NULL,
  first = NULL,
  last = NULL,
  raw = FALSE
)
}
\arguments{
\item{fmt}{String, list or function. A format for break endpoints.}

\item{single}{Glue string: label for singleton intervals. See \code{\link[=lbl_glue]{lbl_glue()}}
for details. If \code{NULL}, singleton intervals will be labelled the same way
as other intervals.}

\item{first}{Glue string: override label for the first category. Write e.g.
\code{first = "<{r}"} to create a label like \code{"<18"}. See \code{\link[=lbl_glue]{lbl_glue()}}
for details.}

\item{last}{String: override label for the last category. Write e.g.
\code{last = ">{l}"} to create a label like \code{">65"}. See \code{\link[=lbl_glue]{lbl_glue()}}
for details.}

\item{raw}{\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#deprecated}{\figure{lifecycle-deprecated.svg}{options: alt='[Deprecated]'}}}{\strong{[Deprecated]}}. Use the \code{raw} argument to \code{\link[=chop]{chop()}}
instead.}
}
\value{
A function that creates a vector of labels.
}
\description{
This uses the midpoint of each interval for
its label.
}
\section{Formatting endpoints}{


If \code{fmt} is not \code{NULL} then it is used to format the endpoints.
\itemize{
\item If \code{fmt} is a string, then numeric endpoints will be formatted by
\code{sprintf(fmt, breaks)}; other endpoints, e.g. \link{Date} objects, will be
formatted by \code{format(breaks, fmt)}.
\item If \code{fmt} is a list, then it will be used as arguments to \link{format}.
\item If \code{fmt} is a function, it should take a vector of numbers (or other objects
that can be used as breaks) and return a character vector. It may be helpful
to use functions from the \code{{scales}} package, e.g. \code{\link[scales:label_number]{scales::label_comma()}}.
}
}

\examples{
chop(1:10, c(2, 5, 8), lbl_midpoints())
}
\seealso{
Other labelling functions: 
\code{\link{lbl_dash}()},
\code{\link{lbl_discrete}()},
\code{\link{lbl_endpoints}()},
\code{\link{lbl_glue}()},
\code{\link{lbl_intervals}()},
\code{\link{lbl_manual}()},
\code{\link{lbl_seq}()}
}
\concept{labelling functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/chop-by-width.R, R/breaks-by-width.R, R/tab.R
\name{chop_width}
\alias{chop_width}
\alias{brk_width}
\alias{brk_width.default}
\alias{tab_width}
\title{Chop into fixed-width intervals}
\usage{
chop_width(x, width, start, ..., left = sign(width) > 0)

brk_width(width, start)

\method{brk_width}{default}(width, start)

tab_width(x, width, start, ..., left = sign(width) > 0)
}
\arguments{
\item{x}{A vector.}

\item{width}{Width of intervals.}

\item{start}{Starting point for intervals. By default the smallest
finite \code{x} (largest if \code{width} is negative).}

\item{...}{Passed to \code{\link[=chop]{chop()}}.}

\item{left}{Logical. Left-closed or right-closed breaks?}
}
\value{
\verb{chop_*} functions return a \code{\link{factor}} of the same length as \code{x}.

\verb{brk_*} functions return a \code{\link{function}} to create \code{breaks}.

\verb{tab_*} functions return a contingency \code{\link[=table]{table()}}.
}
\description{
\code{chop_width()} chops \code{x} into intervals of fixed \code{width}.
}
\details{
If \code{width} is negative, \code{chop_width()} sets \code{left = FALSE} and intervals will
go downwards from \code{start}.
}
\examples{
chop_width(1:10, 2)

chop_width(1:10, 2, start = 0)

chop_width(1:9, -2)

chop(1:10, brk_width(2, 0))

tab_width(1:10, 2, start = 0)

}
\seealso{
\link{brk_width-for-datetime}

Other chopping functions: 
\code{\link{chop}()},
\code{\link{chop_equally}()},
\code{\link{chop_evenly}()},
\code{\link{chop_fn}()},
\code{\link{chop_mean_sd}()},
\code{\link{chop_n}()},
\code{\link{chop_proportions}()},
\code{\link{chop_quantiles}()},
\code{\link{chop_spikes}()},
\code{\link{fillet}()}
}
\concept{chopping functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/breaks.R
\name{brk_default}
\alias{brk_default}
\title{Create a standard set of breaks}
\usage{
brk_default(breaks)
}
\arguments{
\item{breaks}{A numeric vector.}
}
\value{
A function which returns an object of class \code{breaks}.
}
\description{
Create a standard set of breaks
}
\examples{

chop(1:10, c(2, 5, 8))
chop(1:10, brk_default(c(2, 5, 8)))

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/labels-single.R
\name{lbl_endpoints}
\alias{lbl_endpoints}
\alias{lbl_endpoint}
\title{Label chopped intervals by their left or right endpoints}
\usage{
lbl_endpoints(
  left = TRUE,
  fmt = NULL,
  single = NULL,
  first = NULL,
  last = NULL,
  raw = FALSE
)

lbl_endpoint(fmt = NULL, raw = FALSE, left = TRUE)
}
\arguments{
\item{left}{Flag. Use left endpoint or right endpoint?}

\item{fmt}{String, list or function. A format for break endpoints.}

\item{single}{Glue string: label for singleton intervals. See \code{\link[=lbl_glue]{lbl_glue()}}
for details. If \code{NULL}, singleton intervals will be labelled the same way
as other intervals.}

\item{first}{Glue string: override label for the first category. Write e.g.
\code{first = "<{r}"} to create a label like \code{"<18"}. See \code{\link[=lbl_glue]{lbl_glue()}}
for details.}

\item{last}{String: override label for the last category. Write e.g.
\code{last = ">{l}"} to create a label like \code{">65"}. See \code{\link[=lbl_glue]{lbl_glue()}}
for details.}

\item{raw}{\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#deprecated}{\figure{lifecycle-deprecated.svg}{options: alt='[Deprecated]'}}}{\strong{[Deprecated]}}. Use the \code{raw} argument to \code{\link[=chop]{chop()}}
instead.}
}
\value{
A function that creates a vector of labels.
}
\description{
This is useful when the left endpoint unambiguously indicates the
interval. In other cases it may give errors due to duplicate labels.
}
\details{
\code{lbl_endpoint()} is \ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#defunct}{\figure{lifecycle-defunct.svg}{options: alt='[Defunct]'}}}{\strong{[Defunct]}} and gives an
error since santoku 1.0.0.
}
\section{Formatting endpoints}{


If \code{fmt} is not \code{NULL} then it is used to format the endpoints.
\itemize{
\item If \code{fmt} is a string, then numeric endpoints will be formatted by
\code{sprintf(fmt, breaks)}; other endpoints, e.g. \link{Date} objects, will be
formatted by \code{format(breaks, fmt)}.
\item If \code{fmt} is a list, then it will be used as arguments to \link{format}.
\item If \code{fmt} is a function, it should take a vector of numbers (or other objects
that can be used as breaks) and return a character vector. It may be helpful
to use functions from the \code{{scales}} package, e.g. \code{\link[scales:label_number]{scales::label_comma()}}.
}
}

\examples{
chop(1:10, c(2, 5, 8), lbl_endpoints(left = TRUE))
chop(1:10, c(2, 5, 8), lbl_endpoints(left = FALSE))
if (requireNamespace("lubridate")) {
  tab_width(
          as.Date("2000-01-01") + 0:365,
         months(1),
         labels = lbl_endpoints(fmt = "\%b")
       )
}

\dontrun{
  # This gives breaks `[1, 2) [2, 3) {3}` which lead to
  # duplicate labels `"2", "3", "3"`:
  chop(1:3, 1:3, lbl_endpoints(left = FALSE))
}
}
\seealso{
Other labelling functions: 
\code{\link{lbl_dash}()},
\code{\link{lbl_discrete}()},
\code{\link{lbl_glue}()},
\code{\link{lbl_intervals}()},
\code{\link{lbl_manual}()},
\code{\link{lbl_midpoints}()},
\code{\link{lbl_seq}()}
}
\concept{labelling functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/chop-misc.R, R/breaks-misc.R, R/tab.R
\name{chop_fn}
\alias{chop_fn}
\alias{brk_fn}
\alias{tab_fn}
\title{Chop using an existing function}
\usage{
chop_fn(
  x,
  fn,
  ...,
  extend = NULL,
  left = TRUE,
  close_end = TRUE,
  raw = NULL,
  drop = TRUE
)

brk_fn(fn, ...)

tab_fn(
  x,
  fn,
  ...,
  extend = NULL,
  left = TRUE,
  close_end = TRUE,
  raw = NULL,
  drop = TRUE
)
}
\arguments{
\item{x}{A vector.}

\item{fn}{A function which returns a numeric vector of breaks.}

\item{...}{Further arguments to \code{fn}}

\item{extend}{Logical. If \code{TRUE}, always extend breaks to \verb{+/-Inf}. If \code{NULL},
extend breaks to \code{min(x)} and/or \code{max(x)} only if necessary. If \code{FALSE}, never
extend.}

\item{left}{Logical. Left-closed or right-closed breaks?}

\item{close_end}{Logical. Close last break at right? (If \code{left} is \code{FALSE},
close first break at left?)}

\item{raw}{Logical. Use raw values in labels?}

\item{drop}{Logical. Drop unused levels from the result?}
}
\value{
\verb{chop_*} functions return a \code{\link{factor}} of the same length as \code{x}.

\verb{brk_*} functions return a \code{\link{function}} to create \code{breaks}.

\verb{tab_*} functions return a contingency \code{\link[=table]{table()}}.
}
\description{
\code{chop_fn()} is a convenience wrapper: \code{chop_fn(x, foo, ...)}
is the same as \code{chop(x, foo(x, ...))}.
}
\examples{

if (requireNamespace("scales")) {
  chop_fn(rlnorm(10), scales::breaks_log(5))
  # same as
  # x <- rlnorm(10)
  # chop(x, scales::breaks_log(5)(x))
}

}
\seealso{
Other chopping functions: 
\code{\link{chop}()},
\code{\link{chop_equally}()},
\code{\link{chop_evenly}()},
\code{\link{chop_mean_sd}()},
\code{\link{chop_n}()},
\code{\link{chop_proportions}()},
\code{\link{chop_quantiles}()},
\code{\link{chop_spikes}()},
\code{\link{chop_width}()},
\code{\link{fillet}()}
}
\concept{chopping functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/santoku-package.R
\docType{package}
\name{santoku-package}
\alias{santoku}
\alias{santoku-package}
\title{A versatile cutting tool for R: package overview and options}
\description{
santoku is a tool for cutting data into intervals. It provides
the function \code{\link[=chop]{chop()}}, which is similar to base R's \code{\link[=cut]{cut()}} or \code{Hmisc::cut2()}.
\code{chop(x, breaks)} takes a vector \code{x} and returns a factor of the
same length, coding which interval each element of \code{x} falls into.
}
\details{
Here are some advantages of santoku:
\itemize{
\item By default, \code{chop()} always covers the whole range of the data, so you
won't get unexpected \code{NA} values.
\item Unlike \code{cut()} or \code{cut2()}, \code{chop()} can handle single values as well as
intervals. For example, \code{chop(x, breaks = c(1, 2, 2, 3))} will create a
separate factor level for values exactly equal to 2.
\item Flexible and easy labelling.
\item Convenience functions for creating quantile intervals, evenly-spaced
intervals or equal-sized groups.
\item Convenience functions to quickly tabulate chopped data.
\item Can chop numbers, dates, date-times and other objects.
}

These advantages make santoku especially useful for exploratory analysis,
where you may not know the range of your data in advance.

To get started, read the vignette:

\if{html}{\out{<div class="sourceCode r">}}\preformatted{vignette("santoku")
}\if{html}{\out{</div>}}

For more details, start with the documentation for \code{\link[=chop]{chop()}}.
}
\section{Options}{
Santoku has two options:
\itemize{
\item \code{options("santoku.infinity")} sets the symbol for infinity in breaks. The default is
\code{NULL}, in which case the infinity symbol is used on platforms that support it, otherwise
\code{"Inf"} is used.
\item \code{options("santoku.warn_character")} warns if you try to chop a character vector. Set to
\code{FALSE} to turn off this warning.
}
}

\seealso{
Useful links:
\itemize{
  \item \url{https://github.com/hughjonesd/santoku}
  \item \url{https://hughjonesd.github.io/santoku/}
  \item Report bugs at \url{https://github.com/hughjonesd/santoku/issues}
}

}
\author{
\strong{Maintainer}: David Hugh-Jones \email{davidhughjones@gmail.com}

Other contributors:
\itemize{
  \item Daniel Possenriede \email{possenriede@gmail.com} [contributor]
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/chop-by-group-size.R, R/breaks-by-group-size.R,
%   R/tab.R
\name{chop_equally}
\alias{chop_equally}
\alias{brk_equally}
\alias{tab_equally}
\title{Chop equal-sized groups}
\usage{
chop_equally(
  x,
  groups,
  ...,
  labels = lbl_intervals(),
  left = is.numeric(x),
  raw = TRUE
)

brk_equally(groups)

tab_equally(x, groups, ..., left = is.numeric(x), raw = TRUE)
}
\arguments{
\item{x}{A vector.}

\item{groups}{Number of groups.}

\item{...}{Passed to \code{\link[=chop]{chop()}}.}

\item{labels}{A character vector of labels or a function to create labels.}

\item{left}{Logical. Left-closed or right-closed breaks?}

\item{raw}{Logical. Use raw values in labels?}
}
\value{
\verb{chop_*} functions return a \code{\link{factor}} of the same length as \code{x}.

\verb{brk_*} functions return a \code{\link{function}} to create \code{breaks}.

\verb{tab_*} functions return a contingency \code{\link[=table]{table()}}.
}
\description{
\code{chop_equally()} chops \code{x} into groups with an equal number of elements.
}
\details{
\code{chop_equally()} uses \code{\link[=brk_quantiles]{brk_quantiles()}} under the hood. If \code{x} has duplicate
elements, you may get fewer \code{groups} than requested. If so, a warning will
be emitted. See the examples.
}
\examples{
chop_equally(1:10, 5)

# You can't always guarantee equal-sized groups:
dupes <- c(1, 1, 1, 2, 3, 4, 4, 4)
quantile(dupes, 0:4/4)
chop_equally(dupes, 4)
# Or as many groups as you ask for:
chop_equally(c(1, 1, 2, 2), 3)
}
\seealso{
Other chopping functions: 
\code{\link{chop}()},
\code{\link{chop_evenly}()},
\code{\link{chop_fn}()},
\code{\link{chop_mean_sd}()},
\code{\link{chop_n}()},
\code{\link{chop_proportions}()},
\code{\link{chop_quantiles}()},
\code{\link{chop_spikes}()},
\code{\link{chop_width}()},
\code{\link{fillet}()}
}
\concept{chopping functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/chop-misc.R, R/breaks-misc.R, R/tab.R
\name{chop_pretty}
\alias{chop_pretty}
\alias{brk_pretty}
\alias{tab_pretty}
\title{Chop using pretty breakpoints}
\usage{
chop_pretty(x, n = 5, ...)

brk_pretty(n = 5, ...)

tab_pretty(x, n = 5, ...)
}
\arguments{
\item{x}{A vector.}

\item{n}{Positive integer passed to \code{\link[base:pretty]{base::pretty()}}. How many intervals to chop into?}

\item{...}{Passed to \code{\link[=chop]{chop()}} by \code{chop_pretty()} and \code{tab_pretty()}; passed
to \code{\link[base:pretty]{base::pretty()}} by \code{brk_pretty()}.}
}
\value{
\verb{chop_*} functions return a \code{\link{factor}} of the same length as \code{x}.

\verb{brk_*} functions return a \code{\link{function}} to create \code{breaks}.

\verb{tab_*} functions return a contingency \code{\link[=table]{table()}}.
}
\description{
\code{chop_pretty()} uses \code{\link[base:pretty]{base::pretty()}} to calculate breakpoints
which are 1, 2 or 5 times a power of 10. These look nice in graphs.
}
\details{
\code{\link[base:pretty]{base::pretty()}} tries to return \code{n+1} breakpoints, i.e. \code{n} intervals, but
note that this is not guaranteed. There are methods for Date and POSIXct
objects.

For fine-grained control over \code{\link[base:pretty]{base::pretty()}} parameters, use
\code{chop(x, brk_pretty(...))}.
}
\examples{
chop_pretty(1:10)

chop(1:10, brk_pretty(n = 5, high.u.bias = 0))

tab_pretty(1:10)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/breaks.R
\name{breaks-class}
\alias{breaks-class}
\alias{format.breaks}
\alias{print.breaks}
\alias{is.breaks}
\title{Class representing a set of intervals}
\usage{
\method{format}{breaks}(x, ...)

\method{print}{breaks}(x, ...)

is.breaks(x, ...)
}
\arguments{
\item{x}{A breaks object}

\item{...}{Unused}
}
\description{
Class representing a set of intervals
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/breaks-by-width.R
\name{brk_width-for-datetime}
\alias{brk_width-for-datetime}
\alias{brk_width.Duration}
\title{Equal-width intervals for dates or datetimes}
\usage{
\method{brk_width}{Duration}(width, start)
}
\arguments{
\item{width}{A scalar \link{difftime}, \link[lubridate:Period-class]{Period} or
\link[lubridate:Duration-class]{Duration} object.}

\item{start}{A scalar of class \link{Date} or \link[=DateTimeClasses]{POSIXct}.
Can be omitted.}
}
\description{
\code{brk_width()} can be used with time interval classes from base R or the
\code{lubridate} package.
}
\details{
If \code{width} is a Period, \code{\link[lubridate:mplus]{lubridate::add_with_rollback()}}
is used to calculate the widths. This can be useful for e.g. calendar months.
}
\examples{

if (requireNamespace("lubridate")) {
  year2001 <- as.Date("2001-01-01") + 0:364
  tab_width(year2001, months(1),
        labels = lbl_discrete(" to ", fmt = "\%e \%b \%y"))
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/labels.R
\name{lbl_discrete}
\alias{lbl_discrete}
\title{Label discrete data}
\usage{
lbl_discrete(
  symbol = em_dash(),
  unit = 1L,
  fmt = NULL,
  single = NULL,
  first = NULL,
  last = NULL
)
}
\arguments{
\item{symbol}{String: symbol to use for the dash.}

\item{unit}{Minimum difference between distinct values of data.
For integers, 1.}

\item{fmt}{String, list or function. A format for break endpoints.}

\item{single}{Glue string: label for singleton intervals. See \code{\link[=lbl_glue]{lbl_glue()}}
for details. If \code{NULL}, singleton intervals will be labelled the same way
as other intervals.}

\item{first}{Glue string: override label for the first category. Write e.g.
\code{first = "<{r}"} to create a label like \code{"<18"}. See \code{\link[=lbl_glue]{lbl_glue()}}
for details.}

\item{last}{String: override label for the last category. Write e.g.
\code{last = ">{l}"} to create a label like \code{">65"}. See \code{\link[=lbl_glue]{lbl_glue()}}
for details.}
}
\value{
A function that creates a vector of labels.
}
\description{
\code{lbl_discrete()} creates labels for discrete data, such as integers.
For example, breaks
\code{c(1, 3, 4, 6, 7)} are labelled: \verb{"1-2", "3", "4-5", "6-7"}.
}
\details{
No check is done that the data are discrete-valued. If they are not, then
these labels may be misleading. Here, discrete-valued means that if
\code{x < y}, then \code{x <= y - unit}.

Be aware that Date objects may have non-integer values. See \link{Date}.
}
\section{Formatting endpoints}{


If \code{fmt} is not \code{NULL} then it is used to format the endpoints.
\itemize{
\item If \code{fmt} is a string, then numeric endpoints will be formatted by
\code{sprintf(fmt, breaks)}; other endpoints, e.g. \link{Date} objects, will be
formatted by \code{format(breaks, fmt)}.
\item If \code{fmt} is a list, then it will be used as arguments to \link{format}.
\item If \code{fmt} is a function, it should take a vector of numbers (or other objects
that can be used as breaks) and return a character vector. It may be helpful
to use functions from the \code{{scales}} package, e.g. \code{\link[scales:label_number]{scales::label_comma()}}.
}
}

\examples{
tab(1:7, c(1, 3, 5), lbl_discrete())

tab(1:7, c(3, 5), lbl_discrete(first = "<= {r}"))

tab(1:7 * 1000, c(1, 3, 5) * 1000, lbl_discrete(unit = 1000))

# Misleading labels for non-integer data
chop(2.5, c(1, 3, 5), lbl_discrete())

}
\seealso{
Other labelling functions: 
\code{\link{lbl_dash}()},
\code{\link{lbl_endpoints}()},
\code{\link{lbl_glue}()},
\code{\link{lbl_intervals}()},
\code{\link{lbl_manual}()},
\code{\link{lbl_midpoints}()},
\code{\link{lbl_seq}()}
}
\concept{labelling functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/chop.R, R/tab.R
\name{chop}
\alias{chop}
\alias{kiru}
\alias{tab}
\title{Cut data into intervals}
\usage{
chop(
  x,
  breaks,
  labels = lbl_intervals(),
  extend = NULL,
  left = TRUE,
  close_end = TRUE,
  raw = NULL,
  drop = TRUE
)

kiru(
  x,
  breaks,
  labels = lbl_intervals(),
  extend = NULL,
  left = TRUE,
  close_end = TRUE,
  raw = NULL,
  drop = TRUE
)

tab(
  x,
  breaks,
  labels = lbl_intervals(),
  extend = NULL,
  left = TRUE,
  close_end = TRUE,
  raw = NULL,
  drop = TRUE
)
}
\arguments{
\item{x}{A vector.}

\item{breaks}{A numeric vector of cut-points, or a function to create
cut-points from \code{x}.}

\item{labels}{A character vector of labels or a function to create labels.}

\item{extend}{Logical. If \code{TRUE}, always extend breaks to \verb{+/-Inf}. If \code{NULL},
extend breaks to \code{min(x)} and/or \code{max(x)} only if necessary. If \code{FALSE}, never
extend.}

\item{left}{Logical. Left-closed or right-closed breaks?}

\item{close_end}{Logical. Close last break at right? (If \code{left} is \code{FALSE},
close first break at left?)}

\item{raw}{Logical. Use raw values in labels?}

\item{drop}{Logical. Drop unused levels from the result?}
}
\value{
\code{chop()} returns a \code{\link{factor}} of the same length as \code{x}, representing the
intervals containing the value of \code{x}.

\code{tab()} returns a contingency \code{\link[=table]{table()}}.
}
\description{
\code{chop()} cuts \code{x} into intervals. It returns a \code{\link{factor}} of the same length as
\code{x}, representing which interval contains each element of \code{x}.
\code{kiru()} is an alias for \code{chop}.
\code{tab()} calls \code{chop()} and returns a contingency \code{\link[=table]{table()}} from the result.
}
\details{
\code{x} may be a numeric vector, or more generally, any vector which can be
compared with \code{<} and \code{==} (see \link[=groupGeneric]{Ops}). In particular \link{Date}
and \link[=DateTimeClasses]{date-time} objects are supported. Character vectors
are supported with a warning.
\subsection{Breaks}{

\code{breaks} may be a vector or a function.

If it is a vector, \code{breaks} gives the interval endpoints. Repeating a value
creates a "singleton" interval, which contains only that value.
For example \code{breaks = c(1, 3, 3, 5)} creates 3 intervals:
\code{[1, 3)}, \code{{3}} and \code{(3, 5]}.

If \code{breaks} is a function, it is called with the \code{x}, \code{extend}, \code{left} and
\code{close_end} arguments, and should return an object of class \code{breaks}.
Use \verb{brk_*} functions to create a variety of data-dependent breaks.

Names of \code{breaks} may be used for labels. See "Labels" below.
}

\subsection{Options for breaks}{

By default, left-closed intervals are created. If \code{left} is \code{FALSE},
right-closed intervals are created.

If \code{close_end} is \code{TRUE} the final break (or first break if \code{left} is \code{FALSE})
will be closed at both ends. This guarantees that all values \code{x} with
\verb{min(breaks) <= x <= max(breaks)} are included in the intervals.

Before version 0.9.0, \code{close_end} was \code{FALSE} by default, and also behaved
differently with respect to extended breaks: see "Extending intervals" below.

Using \link[=lbl_intervals]{mathematical set notation}:
\itemize{
\item If \code{left} is \code{TRUE} and \code{close_end} is \code{TRUE}, breaks will look like
\code{[b1, b2), [b2, b3) ... [b_(n-1), b_n]}.
\item If \code{left} is \code{FALSE} and \code{close_end} is \code{TRUE}, breaks will look like
\code{[b1, b2], (b2, b3] ... (b_(n-1), b_n]}.
\item If \code{left} is \code{TRUE} and \code{close_end} is \code{FALSE}, all breaks will look like
\code{... [b1, b2) ...}.
\item If \code{left} is \code{FALSE} and \code{close_end} is \code{FALSE}, all breaks will look like
\code{... (b1, b2] ...}.
}
}

\subsection{Extending intervals}{

If \code{extend} is \code{TRUE}, intervals will be extended to \code{[-Inf,
min(breaks))} and \code{(max(breaks), Inf]}.

If \code{extend} is \code{NULL} (the default), intervals will be extended to
\code{[min(x), min(breaks))} and \code{(max(breaks), max(x)]}, only if
necessary, i.e. only if elements of \code{x} would be outside the unextended
breaks.

If \code{extend} is \code{FALSE}, intervals are never extended.

Note that even when \code{extend = TRUE}, extended intervals will be
dropped from the factor levels if they contain no elements and \code{drop = TRUE}.

\code{close_end} is only relevant if intervals are not extended;
extended intervals are always closed on the outside. This is a change from
previous behaviour. Up to version 0.8.0, \code{close_end} was applied to the
last user-specified interval, before any extended intervals were created.

Since 1.1.0, infinity is represented as \eqn{\infty}{the infinity symbol}
in breaks on unicode platforms. Set  \code{options(santoku.infinity = "Inf")}
to get the old behaviour.
}

\subsection{Labels}{

\code{labels} may be a character vector. It should have the same length as the
(possibly extended) number of intervals. Alternatively, \code{labels} may be a
\verb{lbl_*} function such as \code{\link[=lbl_dash]{lbl_dash()}}.

If \code{breaks} is a named vector, then names of \code{breaks} will be
used as labels for the interval starting at the corresponding element. This
overrides the \code{labels} argument (but unnamed breaks will still use \code{labels}).
This feature is \ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#experimental}{\figure{lifecycle-experimental.svg}{options: alt='[Experimental]'}}}{\strong{[Experimental]}}.

If \code{labels} is \code{NULL}, then integer codes will be returned instead of a
factor.

If \code{raw} is \code{TRUE}, labels will show the actual interval endpoints, usually
numbers. If \code{raw} is \code{FALSE} then labels may show other objects, such
as quantiles for \code{\link[=chop_quantiles]{chop_quantiles()}} and friends, proportions of the range for
\code{\link[=chop_proportions]{chop_proportions()}}, or standard deviations for \code{\link[=chop_mean_sd]{chop_mean_sd()}}.

If \code{raw} is \code{NULL} then \verb{lbl_*} functions will use their default (usually
\code{FALSE}). Otherwise, the \code{raw} argument to \code{chop()} overrides \code{raw} arguments
passed into \verb{lbl_*} functions directly.
}

\subsection{Miscellaneous}{

\code{NA} values in \code{x}, and values which are outside the extended endpoints,
return \code{NA}.

\code{kiru()} is a synonym for \code{chop()}. If you load \code{{tidyr}}, you can use it to
avoid confusion with \code{tidyr::chop()}.

Note that \code{chop()}, like all of R, uses binary arithmetic. Thus, numbers may
not be exactly equal to what you think they should be. There is an example
below.
}
}
\examples{

chop(1:7, c(2, 4, 6))

chop(1:7, c(2, 4, 6), extend = FALSE)

# Repeat a number for a singleton break:
chop(1:7, c(2, 4, 4, 6))

chop(1:7, c(2, 4, 6), left = FALSE)

chop(1:7, c(2, 4, 6), close_end = FALSE)

chop(1:7, brk_quantiles(c(0.25, 0.75)))

# A single break is fine if `extend` is not `FALSE`:
chop(1:7, 4)

# Floating point inaccuracy:
chop(0.3/3, c(0, 0.1, 0.1, 1), labels = c("< 0.1", "0.1", "> 0.1"))

# -- Labels --

chop(1:7, c(Lowest = 1, Low = 2, Mid = 4, High = 6))

chop(1:7, c(2, 4, 6), labels = c("Lowest", "Low", "Mid", "High"))

chop(1:7, c(2, 4, 6), labels = lbl_dash())

# Mixing names and other labels:
chop(1:7, c("<2" = 1, 2, 4, ">=6" = 6), labels = lbl_dash())

# -- Non-standard types --

chop(as.Date("2001-01-01") + 1:7, as.Date("2001-01-04"))

suppressWarnings(chop(LETTERS[1:7], "D"))


tab(1:10, c(2, 5, 8))

}
\seealso{
\code{\link[base:cut]{base::cut()}}, \code{\link{non-standard-types}} for chopping objects that
aren't numbers.

Other chopping functions: 
\code{\link{chop_equally}()},
\code{\link{chop_evenly}()},
\code{\link{chop_fn}()},
\code{\link{chop_mean_sd}()},
\code{\link{chop_n}()},
\code{\link{chop_proportions}()},
\code{\link{chop_quantiles}()},
\code{\link{chop_spikes}()},
\code{\link{chop_width}()},
\code{\link{fillet}()}
}
\concept{chopping functions}
